// $Header: https://oberon.oapd.inaf.it/shins/trunk/shins/src/calunit_ctrl/calunitCtrl.cpp 3965 2022-05-30 18:43:35Z depa $

#include "shins_util.h"
#include "calunitCtrl.h"

namespace su = shins::util;


// ----------------------
// OVERLOADED CONSTRUCTOR
// ----------------------
calunitCtrl::calunitCtrl(int argc, char *argv[],
                         const std::string &_strAliasFileName,
                         const shins::util::mapPosType & mainMapNamedPos,
                         const std::map<std::string, std::string> &_mapCfgFiles,
                         boost::shared_ptr<osfsm::OSFsm> &_ptrStateMachine)
:
    m_mainMapNamedPos(mainMapNamedPos),
    __OBJ_NAME__(typeid(this).name())
{
    U6_LLOG(__FUNCTION__);
    I_LLOG(__FUNCTION__ << " - constructor");
    // U1_LLOG(__OBJ_NAME__);
    // m_ptrLogger = _ptrLogger;
    // I_LLOG(__FUNCTION__);
    isActive = false;
    // initializes TAN application: necessary to build a local property tree.
    // Pay attention: Nice::Application derived objets are sigleton.
    // Since motionCtrl module works in the same way, thus uses
    // an Nice::Application derived objet, setups to calibration and motion cannot be
    // simultaneous. One of the two Application must be destroied first.
    // char * cargv[16];
    // for(int i = 0; i < argc && i < 16; i++)
    // {
    //     cargv[i] = new char[argv[i].size() + 4];
    //     memcpy(cargv[i], &(*argv[i].begin()), argv[i].size());
    //     *(cargv[i] + argv[i].size()) = 0;
    // }
    // m_app = new applicationWrapper(argc, cargv);

    // for(int i = 0; i < argc && i < 16; i++)
    // {
    //     delete cargv[i];
    // }
    m_app = new applicationWrapper(argc, argv);
    // Create device node at client side
    // U4_LLOG(__FUNCTION__ << " Create device node at client side ...");

    std::vector<std::string> vecAlias;

    m_ptrStateMachine = _ptrStateMachine;

    // U4_LLOG(__FUNCTION__ << " Looking for aliases from calunit_ctrl module ...");
    vecAlias = shins::util::getAliasesFromFile(_strAliasFileName, "calunit_ctrl");
    if(vecAlias.size() == 0)
    {
        throw std::runtime_error("Issue with aliases: no alias found in " + _strAliasFileName);
    }

    // U4_LLOG(__FUNCTION__ << " Creating Nice endpoints from aliases ...");
    // U4_LLOG(__FUNCTION__ << " Creating callback fuctions ...");
    m_vecAlias = vecAlias;
    m_mapCfgFiles = _mapCfgFiles;
    m_strAliasFilename = _strAliasFileName;

    m_callbackFnSM = boost::bind(&niceCallback::callback_motion_complete_SM, _1, boost::ref(m_ptrStateMachine), boost::ref(m_mtxStateMachine));

    // bind to function, parameter of callback used by multiDevice
//    m_callbackFnIndi = boost::bind(&niceCallback::callback_motion_complete_indi, m_ptrCallbackObjIndi, _1, _2,
//        boost::ref(m_ptrLbtoIF));
    m_ptrMonSeq = boost::make_shared<IceUtil::Monitor<IceUtil::Mutex> >();
//    m_callbackFnCount = boost::bind(&niceCallback::callback_motion_complete_count, m_ptrCallbackObjCount, _1,
//        boost::ref(m_moving_devs), boost::ref(m_mtxCount), boost::ref(m_ptrMonSeq));
    m_callbackFnCount = boost::bind(&niceCallback::callback_motion_complete_count, _1, boost::ref(m_moving_devs), boost::ref(m_mtxCount), boost::ref(m_ptrMonSeq));

    // U4_LLOG(__FUNCTION__ << " Connecting to remote configuration tree ...");
    // moveHome();

    // U4_LLOG(__FUNCTION__ << " Constructor done.");
    U8_LLOG("");
}

calunitCtrl::~calunitCtrl()
{
    U6_LLOG(__FUNCTION__);
    delete m_app;
    U8_LLOG("");
}
void calunitCtrl::connect()
{
    Nice::JointPtr nodeX = Nice::Application::config()->node("STAGEX");
    Nice::Endpoint ep;
    std::string strEP;
    std::vector<std::string> vecAliasX;
    std::vector<std::string>::const_iterator it;
    for(it = m_vecAlias.begin(); it < m_vecAlias.end(); ++it)
    {
        strEP = "[ALIAS=" + (*it) + "]"; // endpoint name
        ep = Nice::Endpoint::fromString(strEP); // endpoint
        std::string upper_it = boost::to_upper_copy(*it);
        if((it->find(boost::to_lower_copy(suc::CAL_MIRROR_DEP)) != std::string::npos)
            || (it->find(boost::to_lower_copy(suc::CAL_FIBER_DEP)) != std::string::npos)
            || (it->find(boost::to_lower_copy(suc::CAL_FILTER_DEP)) != std::string::npos))
        {
            try
            {

                U9_LLOG(upper_it << " Exists? ");
                if(nodeX->exist(upper_it))
                {
                    U9_LLOG("Chopping " << upper_it);
                    Nice::JointPtr nodeY = nodeX->node(upper_it, false);
                    nodeX->chop(nodeY);
                }
            }
            catch(...)
            {
                U9_LLOG("catch " << upper_it);
            }
            nodeX->setAtNode<Nice::Endpoint>(upper_it + ".CON", ep);
            vecAliasX.push_back(*it);
            U9_LLOG("Alias stageX name : " << *it);
            //aliases are lower case
        }
        else
        {
            E_LLOG(std::string(__FUNCTION__) << ":Wrong alias file for calunit_ctrl " + *it +". Throwing");
            throw std::runtime_error(std::string(__FUNCTION__) + "Wrong alias file for calunit_ctrl " + *it);
        }
    }
    if(vecAliasX.size() != 0)
    {
        Nice::JointPtr nodeX2 = Nice::Application::config()->node("STAGEX");
        //multiStageX shifts to uppercase aliases. Since then lower case disappears
        m_ptrMultiStageX = boost::make_shared<multiStageX>(nodeX2, vecAliasX, m_mapCfgFiles);
        U9_LLOG("nodeX2 -> " << nodeX2._ptr << suc::tilde << "nodex -> " << nodeX._ptr);
    }
    if(m_ptrMultiStageX)
    {
        m_ptrMultiStageX->connect();
        try
        {
            m_ptrMultiStageX->standby();
            m_ptrMultiStageX->online();
        }
        catch(...)
        {
            E_LLOG("Erorr changing state to MultiStageX");
        }
        m_ptrMultiStageX->set_callback_fn(m_callbackFnSM, "calib_dev-motion_complete_SM");
        m_ptrMultiStageX->set_callback_fn_count(m_callbackFnCount, "motion_complete_count");
    }
    I_LLOG(__FUNCTION__ << suc::tilde << "Completed");
};

void calunitCtrl::SetPrimaryBridges()
{
//    std::string strEP;
//    Nice::Endpoint ep;
//    std::vector<std::string>::const_iterator it;
//    Nice::JointPtr nodeX2 = Nice::Application::config()->node("STAGEX");
//    for(it = m_vecAlias.begin(); it < m_vecAlias.end(); ++it)
//    {
//        strEP = "[ALIAS=" + (*it) + "]"; // endpoint name
//        ep = Nice::Endpoint::fromString(strEP); // endpoint
//        nodeX2->setAtNode<Nice::Endpoint>(*it + ".CON", ep);
//    }
//
//    m_ptrMultiStageX = boost::make_shared<multiStageX>(nodeX2, m_vecAlias, m_mapCfgFiles);
    //_ptrLogger->cloneWithPrefix("SHINS"));
//    m_ptrMultiStageX->connect();
//    m_ptrMultiStageX->standby();
//    m_ptrMultiStageX->online();
    // Setting callback function for multiStageX object
    m_ptrMultiStageX->set_callback_fn(m_callbackFnSM, "calib_dev-motion_complete_SM");
//    m_ptrMultiStageX->set_callback_fn_count(m_callbackFnCount, "motion_complete_count");

        //m_ptrMultiStageX->connect();
    //m_ptrMultiStageX->set_callback_fn(m_callbackFnSM, "calib_dev-motion_complete_SM");
}

// TAN Application related methods
// int calunitCtrl::run(int argc, char* argv []);

// std::vector<std::string> calunitCtrl::getAliasesFromFile(const std::string& _strFileName)
// {
//   std::ifstream ifsAlias(_strFileName.c_str(), std::ifstream::in);

//   std::vector<std::string> vecAlias;
//   std::vector<std::string> tmpVecSplit;
//   BOOST_AUTO(itFind, tmpVecSplit.begin());

//   std::string line = "";
//   std::string strAlias = "";
//   std::string strTokens = "[+#,=:\"";

//   while (std::getline(ifsAlias, line))
//     {
//       if(boost::algorithm::starts_with(line,"calunit_ctrl") &&//
//           !(boost::algorithm::contains(line, "LABTEST")) &&//
//           (boost::algorithm::contains(line, "ALIAS="))) // excluding alias endpoint for server
//         {
//           tmpVecSplit = shins::util::split_with_boost(line, strTokens);
//           itFind = std::find(tmpVecSplit.begin(), tmpVecSplit.end(), "ALIAS");
//           if(itFind < tmpVecSplit.end())
//             {
//               ++itFind;
//               vecAlias.push_back(*itFind);
//               tmpVecSplit.clear();
//               U3_LLOG(vecAlias.back());
//               continue;
//             }
//           else
//             {
//               E_LLOG("Did not find alias in line");
//               throw std::runtime_error("Did not find alias in line " + line);
//             }
//         }
//     }
//   ifsAlias.close();
//   return vecAlias;
// }

// std::vector<std::string> calunitCtrl::getAliasesFromFile(const std::string& _strFileName)
// {
//   std::vector<std::string> vecAlias;
//   std::ifstream ifsAlias(_strFileName.c_str(), std::ifstream::in);
//   std::string line = "";
//   std::string strAlias = "";

//   std::size_t posAlias = 0, posComma = 0, posEqual = 0;

//   while (std::getline(ifsAlias, line))
//     {
//       if(line.find("calunit_ctrl") != std::string::npos )
//         {
//           posAlias = line.find("[ALIAS=");
//           if(posAlias != std::string::npos)
//             {
//               posComma = line.find(",", posAlias);
//               posEqual = line.find("=", posAlias);
//               strAlias = line.substr(posEqual+1, posComma-posEqual-1);
//               boost::trim_right(strAlias);
//               boost::trim_left(strAlias);
//               vecAlias.push_back(strAlias);
//             }
//         }
//     }
//   ifsAlias.close();
//   return vecAlias;
// }


bool calunitCtrl::isDevice(const std::string& _strDevAlias)
{
    return m_ptrMultiStageX->isDevice(_strDevAlias);
}

double calunitCtrl::ConvertToDefaultUnit(const std::string & alias, double destination, const std::string & unit)
{
    std::map<std::string, StageXDevice>::const_iterator itDevice = GetMultiStageX().find(alias);
    static std::string log_msg;
    log_msg.clear();
    if(itDevice != GetMultiStageX().end())
    {
        std::string final_unit = unit;
        if(unit == suc::default_unit)final_unit = suc::steps;//TBD unify default unit
        std::map<std::string, std::tr1::tuple<double, double> >::const_iterator itUnit = itDevice->second.units.find(final_unit);
        if(itUnit != itDevice->second.units.end())
        {
            return destination/std::tr1::get<0>(itUnit->second) + std::tr1::get<1>(itUnit->second);
        }
        else
        {   log_msg = std::string(__FUNCTION__) + ": No " + unit + " unit in " + alias + ".";
            E_LLOG(log_msg);
            throw std::invalid_argument(log_msg);
        }
    }
    else
    {
        log_msg = std::string(__FUNCTION__) + ": No " + alias + " device in CALUNIT_CTRL. Throwing!";
        E_LLOG(log_msg);
        throw std::invalid_argument(log_msg);
    }
}


void calunitCtrl::moveCalFiberDepToSafe()
{
    std::vector<std::string> vecDevAlias(1,suc::CAL_FIBER_DEP);
    m_ptrMultiStageX->moveToPositiveLimit(vecDevAlias);
}


void calunitCtrl::SetSecondaryCallbacks(const std::string & alias,
                                     void (* MoveToHomeCB)(const Nice::Completion & c),
                                     void (* MoveToNamedCB)(const Nice::Completion & c),
                                     void (* MoveAbsoluteCB)(const Nice::Completion & c),
                                     void (* MoveRelativeCB)(const Nice::Completion & c),
                                     void (* MoveToLimitCB)(const Nice::Completion & c))
{
    m_ptrMultiStageX->SetSecondaryCallbacks(alias, MoveToHomeCB, MoveToNamedCB, MoveAbsoluteCB, MoveRelativeCB, MoveToLimitCB);
}


Ice::Int calunitCtrl::getPosition(std::map<std::string, Ice::Double> &_mapPosition, const std::string &_unit)
{
//    U6_LLOG(__FUNCTION__);
    Ice::Int status = EXIT_SUCCESS;
    try
    {
        status = m_ptrMultiStageX->getPosition(_mapPosition, _unit);
    }
    catch(...)
    {
        U6_LLOG(__FUNCTION__);
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
//    U8_LLOG("");
    return status;
}


int calunitCtrl::getNdevices()
{
    U6_LLOG(__FUNCTION__);
    U8_LLOG("");
    return (m_ptrMultiStageX->getNdevices());
}


void calunitCtrl::moveHome()
{
    U6_LLOG(__FUNCTION__);
    if(m_ptrMultiStageX->getNdevices() > 0)
    {
        m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));
        m_ptrMultiStageX->moveToReference();
    }
    else
    {
        E_LLOG(std::string(__FUNCTION__) << ":No devices in calunit_ctlr! Throwing");
        U8_LLOG("");
        throw std::runtime_error("No devices in calunit_ctlr! ");
    }
    U8_LLOG("");
}


void calunitCtrl::moveDeviceHome(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    std::vector<std::string> tmpVecStageX;

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
    }
    try
    {
        m_ptrStateMachine->process_event(osfsm::ready_busy(tmpVecStageX.size()));
        m_ptrMultiStageX->moveDeviceToReference(tmpVecStageX);
        m_ptrMultiStageX->moveDeviceToReferenceWait(tmpVecStageX, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void calunitCtrl::changeState(const Basda::ServiceState &_state, const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    switch (_state)
    {
    case Basda::ONLINE:
    {
        m_ptrMultiStageX->online(_vecDevAlias);
        break;
    }
    case Basda::STANDBY:
    {
        m_ptrMultiStageX->standby(_vecDevAlias);
        break;
    }
    case Basda::SHUTDOWN:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::OFFLINE:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::WORKING:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::ERROR:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    default:
        W_LLOG("State " << _state << " not handled")
        ;
    }
    U8_LLOG("");
}



void calunitCtrl::moveToReferenceAll()
{
    U6_LLOG(__FUNCTION__);
    static Nice::Time nice_time = Nice::Time::seconds(80);
    try
    {
        moveToReferenceAllStart();
        moveToReferenceAllWait(nice_time);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void calunitCtrl::moveToReferenceAllStart()
{
    U6_LLOG(__FUNCTION__);
//    m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));
    // moving to "home" reference switch all devices
    // the number of moving wheels is multiplied by 2 since they move to reference twice
    m_moving_devs = m_ptrMultiStageX->getNdevices();

    if(m_ptrMultiStageX->getNdevices() != 0)
    {
        try
        {
            m_ptrMultiStageX->moveToReferenceStart();
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }

    U8_LLOG("");
}


void calunitCtrl::moveToReferenceAllWait(const Nice::Time &_time)
{
    // m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));
    // moving to "home" reference switch all devices
    // the number of moving wheels is multiplied yby 2 since they move to reference twice
    // m_moving_devs = m_ptrMultiStageX->getNdevices();

    U6_LLOG(__FUNCTION__);
    if(m_ptrMultiStageX->getNdevices() != 0)
    {
        try
        {
            m_ptrMultiStageX->moveToReferenceWait(_time);
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }
//    m_ptrStateMachine->decrease_running_devs();
//    m_ptrStateMachine->decrease_running_devs();
//    m_ptrStateMachine->decrease_running_devs();
//    m_ptrStateMachine->process_event(osfsm::busy_ready);

    U8_LLOG("");
}
void calunitCtrl::moveToReference(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        moveToReferenceStart(_vecDevAlias);
        moveToReferenceWait(_vecDevAlias, Nice::Time::seconds(60));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void calunitCtrl::initPrepare(const std::vector<std::string> & _vecDevAlias)
{
    m_init.clear();
    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(*it))
        {
            m_init.InitStageX.push_back(*it);
            continue;
        }
    }
}


void calunitCtrl::moveToReferenceStart(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    initPrepare(_vecDevAlias);
    m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));
    // moving to "home" reference switch all devices
    // the number of moving wheels is multiplied by 2 since they move to reference twice
    m_moving_devs = m_ptrMultiStageX->getNdevices();

    if(m_ptrMultiStageX->getNdevices() != 0)
    {
        try
        {
            m_ptrMultiStageX->moveDeviceToReferenceStart(_vecDevAlias);
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }

    U8_LLOG("");
}


void calunitCtrl::moveToReferenceWait(const std::vector<std::string> &_vecDevAlias, const Nice::Time &_time)
{
    // m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));
    // moving to "home" reference switch all devices
    // the number of moving wheels is multiplied yby 2 since they move to reference twice
    // m_moving_devs = m_ptrMultiStageX->getNdevices();

    U6_LLOG(__FUNCTION__);
    if(m_ptrMultiStageX->getNdevices() != 0)
    {
        try
        {
            m_ptrMultiStageX->moveDeviceToReferenceWait(_vecDevAlias, _time);
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }

    U8_LLOG("");
}


void calunitCtrl::moveToNamedPosition(const std::map<std::string, std::string> &_mapNamedPos)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        moveToNamedPositionStart(_mapNamedPos);
        moveToNamedPositionWait(_mapNamedPos, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}

void calunitCtrl::setupPrepare(const std::map<std::string, std::string>& _mapNamedPos)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, std::tr1::tuple<int, int> > NamedPosMap;
    std::string svcNameXY = "";
    int idxPos;
    std::map<std::string, std::string>::const_iterator it;
    shins::util::EShinsMovements detected;
    double absolute, relative = 0;
    std::string unit;
    m_setup.clear();
    for(it = _mapNamedPos.begin(); it != _mapNamedPos.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            switch(detected)
            {
            case shins::util::SHINS_MOV_NAMED:
                NamedPosMap = m_mainMapNamedPos[it->first];
                if(NamedPosMap.find(it->second) != NamedPosMap.end())
                    idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                else
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    throw(std::invalid_argument(msg.str()));
                }
                m_setup.NamedMapStageX.insert(std::pair<std::string, int>(it->first, idxPos));
                NamedPosMap.clear();
                break;
            case shins::util::SHINS_MOV_ABSOLUTE:
                U9_LLOG("Detected Absolute Position : " << it->first << " -> " << it->second << "; value = " << absolute << "; unit = " << unit);
                m_setup.AbsMapStageX.insert(std::pair<std::string, double>(it->first, absolute));
                m_setup.AbsMapStageXUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_RELATIVE:
                U9_LLOG("Detected Relative Position : " << it->first << " -> " << it->second << "; value = " << relative << "; unit = " << unit);
                m_setup.RelMapStageX.insert(std::pair<std::string, double>(it->first, relative));
                m_setup.RelMapStageXUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_UNDEFINED:
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    throw(std::invalid_argument(msg.str()));
                }
            }
            continue;
        }
    }
    U8_LLOG("");
}

void calunitCtrl::moveToNamedPositionStart(const std::map<std::string, std::string> &_mapNamedPos)
{
    U6_LLOG(__FUNCTION__);
    setupPrepare(_mapNamedPos);

    // swich state machine to "busy"
    // U8_LLOG(tmpMapStageX.size());
    // I_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));
    m_moving_devs = m_setup.NamedMapStageX.size() + m_setup.AbsMapStageX.size() + m_setup.RelMapStageX.size();
//    m_ptrStateMachine->process_event(osfsm::ready_busy(m_moving_devs));

    // I_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));
    // I_LLOG(*(m_ptrStateMachine->event_(osfsm::readyDone)));
    // IceUtil::Monitor<IceUtil::Mutex>::Lock lock(*(m_ptrMonSeq));
    if(m_moving_devs > 0)
    {
        try
        {
            m_ptrMultiStageX->moveToNamedPositionStart(m_setup.NamedMapStageX);
            m_ptrMultiStageX->moveToAbsolutePositionStart(m_setup.AbsMapStageX, m_setup.AbsMapStageXUnits);
            m_ptrMultiStageX->moveToRelativePositionStart(m_setup.RelMapStageX, m_setup.RelMapStageXUnits);
        }
        catch(Basda::Svc::MoccaMotionException exMotion)
        {
            E_LLOG(__FUNCTION__ << " ~ Error : " << exMotion.what());
            U8_LLOG("");
            throw exMotion;
        }
        catch(Basda::ServiceException exService)
        {
            E_LLOG(__FUNCTION__ << " ~ Error : " <<  exService.what());
            U8_LLOG("");
            throw exService;
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }
    U8_LLOG("");
}


void calunitCtrl::moveToNamedPositionWait(const std::map<std::string, std::string> &_mapNamedPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__);
    su::Timer timer(_time);
    try
    {
        m_ptrMultiStageX->moveToNamedPositionWait(m_setup.NamedMapStageX, timer.residual());
        m_ptrMultiStageX->moveToAbsolutePositionWait(m_setup.AbsMapStageX, timer.residual());
        m_ptrMultiStageX->moveToRelativePositionWait(m_setup.RelMapStageX, timer.residual());
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void calunitCtrl::moveToAbsolutePosition(const std::map<std::string, double> &_mapPos, const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        moveToAbsolutePositionStart(_mapPos, _mapUnit);
        moveToAbsolutePositionWait(_mapPos, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void calunitCtrl::moveToAbsolutePositionStart(const std::map<std::string, double> &_mapPos,
    const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, double> tmpMapStageX;
    int movingDevices = 0;

    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            // std::cout << "stageX" << std::endl;
            tmpMapStageX.insert(*it);
            continue;
        }

    }
    movingDevices = tmpMapStageX.size();
    m_ptrStateMachine->process_event(osfsm::ready_busy(movingDevices));

    if(tmpMapStageX.size() > 0)
    {
        try
        {
            m_ptrMultiStageX->moveToAbsolutePositionStart(tmpMapStageX, _mapUnit);
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }
    U8_LLOG("");
}


void calunitCtrl::moveToAbsolutePositionWait(const std::map<std::string, double> &_mapPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, double> tmpMapStageX;
//    int movingDevices = 0;

    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            // std::cout << "stageX" << std::endl;
            tmpMapStageX.insert(*it);
            continue;
        }
    }

    if(tmpMapStageX.size() > 0)
    {
        try
        {
            m_ptrMultiStageX->moveToAbsolutePositionWait(tmpMapStageX, _time);
        }
        catch(...)
        {
            U8_LLOG("Error Waiting");
            throw;
        }

    }
    U8_LLOG("");
}


void calunitCtrl::moveToRelativePosition(const std::map<std::string, double> &_mapPos, const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    E_LLOG(__FUNCTION__ << " not implemented");
    U8_LLOG("");
    throw(std::runtime_error("Function not implemented yet"));
}


void calunitCtrl::moveToRelativePositionStart(const std::map<std::string, double> &_mapPos, const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    E_LLOG(__FUNCTION__ << " not implemented");
    U8_LLOG("");
    throw(std::runtime_error("Function not implemented yet"));
}


void calunitCtrl::moveToRelativePositionWait(const std::map<std::string, double> &_mapPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__);
    E_LLOG(__FUNCTION__ << " not implemented");
    U8_LLOG("");
    throw(std::runtime_error("Function not implemented yet"));
}


void calunitCtrl::setupLamps(const std::map<std::string, std::string> &_mapSetup)
{
    U6_LLOG(__FUNCTION__);
    E_LLOG(__FUNCTION__ << " not implemented");
    U8_LLOG("");
    throw(std::runtime_error("Function not implemented yet"));
}


void calunitCtrl::emergencyAbort()
{
    U6_LLOG(__FUNCTION__);
    try
    {
        m_ptrMultiStageX->emergencyAbort();
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }
    U8_LLOG("");
}


int calunitCtrl::stop(const std::vector<std::string> &_vecDevAlias)
{
    int status = EXIT_SUCCESS;
    U6_LLOG(__FUNCTION__);
    try
    {
        m_ptrMultiStageX->stop(_vecDevAlias);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }
    U8_LLOG("");
    return status;
}


int calunitCtrl::abort(const std::vector<std::string> &_vecDevAlias)
{
    int status = EXIT_SUCCESS;
    U6_LLOG(__FUNCTION__);
    try
    {
        m_ptrMultiStageX->abort(_vecDevAlias);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }
    U8_LLOG("");
    return status;
}

int calunitCtrl::getJsonStatus(std::string & jsonStatus)
{
    int status = EXIT_SUCCESS;

    std::vector<std::string> vecAlias;
    vecAlias.push_back("CAL_FIBER_DEP");
    vecAlias.push_back("CAL_MIRROR_DEP");
    vecAlias.push_back("CAL_FILTER_DEP");

    std::map<std::string, std::string> status_motion;
    status_motion = m_ptrMultiStageX->getNamedPosition2(vecAlias);

    //    struct MotorStatus{
    //               int steps;
    //               string namedPos;
    //               bool connected;
    //               bool moving;
    //               bool atHome;
    //               bool inSwNegativeLimit;
    //               bool inSwPositiveLimit;
    //               bool inHwNegativeLimit;
    //               bool inWhPositiveLimit;
    //    };
    std::map<std::string, StageXDevice>::const_iterator itStageX;
    const std::map<std::string, StageXDevice> & MapStageX = m_ptrMultiStageX->GetMultiStageX();
    int counter = 0;
    for(itStageX = MapStageX.begin(); itStageX != MapStageX.end(); itStageX++)
    {
        counter++;
        Nice::SeqBitSet switch_status;
        bool isSimulated = false;
        std::string service_name;
        try{service_name = itStageX->second.iPtr->readServiceName();}catch(...){E_LLOG("readServiceName");}
//        try{switch_status = itStageX->second.iPtr->getPositionSwitchStatus(); isSimulated = switch_status[0].getDepth() < 32;}catch(...){}
        try
        {
            switch_status = itStageX->second.iPtr->getPositionSwitchStatus();
            isSimulated = switch_status[0].getDepth() < 32;
        }
        // catch(Nice::ChatException err)
        // {
        //     usleep(1000000);
        //     continue;
        // }
        catch(...)
        {
            E_LLOG(service_name << " ~ getPositionSwitchStatus");
            usleep(1000000);
            continue;
        }
        std::string named_pos = itStageX->second.iPtr->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
//        try{U9_LLOG("getPositionSwitchStatus : " << switch_status << " ~ " << switch_status[0].getDepth());}catch(...){E_LLOG("getPositionSwitchStatus");}
        bool isAtHome = false;
        try{isAtHome = itStageX->second.iPtr->isAtHome();}catch(...){E_LLOG(service_name << " ~ isAtHome");}
        bool inLimit = false;
        try{inLimit = itStageX->second.iPtr->isAtLimit();}catch(...){E_LLOG(service_name << " ~ isAtLimit");}
//         if(itStageX->second.iPtr->isAtLimit())
//         {
//             try{U9_LLOG("isAtLimit2 : " << itStageX->second.iPtr->get<int>("VAR.LIMIT"));}catch(...){E_LLOG("isAtLimit2");}
//             std::string limit_result = itStageX->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             itStageX->second.iPtr->config()->setAtNode<std::string>("VAR.LIMIT", "-2");
//             limit_result = itStageX->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             limit_result = itStageX->second.iPtr->config()->leaf("VAR.LIMIT")->get<std::string>();
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
// //                val.limit = m_node->getAtNode<double> ("LIMIT")
//             double limit_double = itStageX->second.iPtr->config()->getAtNode<double>("LIMIT");
//             U9_LLOG("PropertyName : " << "LIMIT" << "; limit_double : " << limit_double);
//         }
        bool isMoving = false;
        try{isMoving = itStageX->second.iPtr->isMoving();}catch(...){E_LLOG(service_name << " ~ isMoving");}
//        try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
//        try{U9_LLOG("getNamedPosition : " << itStageX->second.iPtr->getNamedPosition(1));}catch(...){E_LLOG("getNamedPosition");}
//        try{U9_LLOG("readPosition : " << itStageX->second.iPtr->readPosition());}catch(...){E_LLOG("readPosition");}
//        try{U9_LLOG("readDeviceEncoderPosition : " << itStageX->second.iPtr->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
//        try{U9_LLOG("readAbsoluteEncoderPosition : " << itStageX->second.iPtr->readAbsoluteEncoderPosition());}catch(...){E_LLOG("readAbsoluteEncoderPosition");}
//        try{U9_LLOG("readIncrementalEncoderPosition : " << itStageX->second.iPtr->readIncrementalEncoderPosition());}catch(...){E_LLOG("readIncrementalEncoderPosition");}
        bool isReachable = false;
        try{isReachable = itStageX->second.iPtr->isReachable();}catch(...){E_LLOG(service_name << " ~ isReachable");}
//        try{U9_LLOG("getPosition : " << itStageX->second.iPtr->getPosition());}catch(...){E_LLOG("getPosition");}
        int deviceEncoderPosition;
        try{deviceEncoderPosition = itStageX->second.iPtr->getDeviceEncoderPosition();}catch(...){E_LLOG(service_name << " ~ getDeviceEncoderPosition");}
//        try{U9_LLOG("getAbsoluteEncoderPosition : " << itStageX->second.iPtr->getAbsoluteEncoderPosition());}catch(...){E_LLOG("getAbsoluteEncoderPosition");}
//        try{U9_LLOG("getIncrementalEncoderPosition : " << itStageX->second.iPtr->getIncrementalEncoderPosition());}catch(...){E_LLOG("getIncrementalEncoderPosition");}
//        try{U9_LLOG("getVelocity : " << itStageX->second.iPtr->getVelocity());}catch(...){E_LLOG("getVelocity");}

        std::stringstream json_alias;
        json_alias << "\"" << itStageX->first << "\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        if(counter < MapStageX.size())jsonStatus += ", ";
    }

    return status;
}
