// $Header: https://oberon.oapd.inaf.it/shins/trunk/shins/src/motion_ctrl/motionCtrl.cpp 3987 2022-07-06 16:38:50Z depa $

#include "shins_util.h"
#include "motionCtrl.hpp"

namespace su = shins::util;

void MotionLogBoostDiagnosticInformation(int line, std::string device)
{
    std::string what = boost::current_exception_diagnostic_information();
    std::string toErase = "std::exception::what:";
    size_t pos = what.find(toErase);
    if(pos != std::string::npos)what.erase(pos, toErase.length());
    what.erase(std::remove(what.begin(), what.end(), '\n'), what.end());
    std::stringstream log_msg;
    log_msg << "Line: " << line << " ~ " << device << " ~ " << what;
    E_LLOG(log_msg.str());
}
// this constructor
motionCtrl::motionCtrl(int argc, char *argv[],
                       const std::string &_strAliasFileName,
                       const shins::util::mapPosType & mainMapNamedPos,
                       const std::map<std::string, std::string> &_mapCfgFiles,
                       boost::shared_ptr<osfsm::OSFsm> &_ptrStateMachine)
:
    m_mainMapNamedPos(mainMapNamedPos),
    __OBJ_NAME__(typeid(this).name())
{
    U6_LLOG(__FUNCTION__ << " Constructor");
    isActive = false;
    std::string strEP;
    std::string strDevAlias;
    std::string logMsg;
//    std::stringstream log_message("Created Endpoints for : ");
    // vecAlias are defined in alias.cfg file
    std::vector<std::string> vecAlias, vecAliasWheel, vecAliasX, vecAliasXY;
    // Create device node at client side
    Nice::JointPtr nodeWheel = Nice::Application::config()->node("WHEEL");
    Nice::JointPtr nodeXY = Nice::Application::config()->node("STAGEXY");
    Nice::JointPtr nodeX = Nice::Application::config()->node("STAGEX");

//    Nice::JointPtr nodeDev = Nice::Application::config()->node("DEV");

    Nice::Endpoint ep, epDev;
    // Basda::MoccaWheelServiceWorkerInterfacePtr tmpWheelPtr;
    U2_LLOG("Init TaN application wrapper for multiple devices control ...");

    m_app = new applicationWrapper(argc, argv); // initializes TAN application
    // LOG_SYSLOG;

    m_ptrStateMachine = _ptrStateMachine;
    // U8_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));

    vecAlias = shins::util::getAliasesFromFile(_strAliasFileName, "motion_ctrl");
    if(vecAlias.size() == 0)
    {
        logMsg = "Issue with aliases: no alias found in " + _strAliasFileName;
        E_LLOG(logMsg);
        U8_LLOG("");
        throw std::runtime_error(logMsg);
    }

//    std::map<std::string, std::map<std::string, int> >::iterator itNamedPos;
//    for(itNamedPos = m_mainMapNamedPos.begin(); itNamedPos != m_mainMapNamedPos.end(); itNamedPos++)
//    {
//        std::map<std::string, int>::iterator itPoz;
//        std::stringstream log_msg;
//        for(itPoz = (itNamedPos->second).begin(); itPoz != (itNamedPos->second).end(); itPoz++)
//        {
//            log_msg << itPoz->first << " ~ " << itPoz->second << " * ";
//        }
//        U9_LLOG(itNamedPos->first  << " * " << log_msg.str());
// APODIZER_W * HOLE ~ 3 * HOLE2 ~ 4 * HOLE3 ~ 5 * HOLE4 ~ 6 * HOLE5 ~ 7 * SP1_APO ~ 0 * SP2a_APO ~ 1 * SP2b_APO ~ 2 *
// CAL_FIBER_DEP * DEFOCUS_IN ~ 1 * FOCUS_IN ~ 0 * NO_FIBERS ~ 2 *       etc etc
//    }
    // Creating Nice endpoints from aliases
    std::vector<std::string>::iterator it;
    for(it = vecAlias.begin(); it < vecAlias.end(); ++it)
    {
        strEP = "[ALIAS=" + (*it) + "]"; // endpoint name
        if(!(boost::algorithm::contains((*it), "-dev")))
        {
            ep = Nice::Endpoint::fromString(strEP); // endpoint
        }
        else
        {
            W_LLOG("Do Contains '-dev': " << *it);
            epDev = Nice::Endpoint::fromString(strEP);
        }
//        log_message << *it << " " << strEP << " ~ ";
//        U9_LLOG("Created endpoint " << *it  << " " << strEP);

        // Creating endpoints in local property tree using aliases
        // looking for wheels first
        if((*it).find("_w") != std::string::npos)
        {
            if(!(boost::algorithm::contains((*it), "DEV")))
            {
                nodeWheel->setAtNode<Nice::Endpoint>(boost::to_upper_copy(*it) + ".CON", ep);
                vecAliasWheel.push_back(boost::to_upper_copy(*it));
            }
            else
            {
                continue;
                // try
                //   {
                // D_LLOG(__PRETTY_FUNCTION__ << ":" << __LINE__ << ":" << "DEV";
                // nodeDev->setAtNode<Nice::Endpoint>(boost::to_upper_copy(*it), ep);
                // //Nice::JointPtr transfNode = nodeDev->node("CFG.TRANS.DEGREE"); //"MOTION_CTRL.WHEELS."+boost::to_upper_copy(*it)+".DEV.1.CFG.TRANS.DEGREE");
                // // Nice::Point transfFactor = transfNode->
                // //Nice::Point transfNode = nodeWheel->getAtNode<Nice::Point>("."+boost::to_upper_copy(*it)+".DEV.1.CFG.TRANS.DEG");
                // U2_LLOG(":" << __LINE__ << ":" << "printing trnafer node X value";
                // U2_LLOG(":" << __LINE__ << ":" << nodeDev->get<std::string>();
                //   }
                // catch(std::exception e)
                //   {
                //     m_error << "ERROR: " << __PRETTY_FUNCTION__ << ":" << __LINE__ << ":" << e.what();
                //   }
            }
        }
        else
        {
            // looking for X stages
            if((it->find("_dep") != std::string::npos) || (it->find("shutter") != std::string::npos)
                || (it->find("_lens") != std::string::npos))
            {
                // U8_LLOG("Found deployer")
                nodeX->setAtNode<Nice::Endpoint>(boost::to_upper_copy(*it) + ".CON", ep);
                vecAliasX.push_back(*it);
            }
            else
            {
                // looking for XY stage
                if(it->find("_tt") != std::string::npos)
                {
                    nodeXY->setAtNode<Nice::Endpoint>(boost::to_upper_copy(*it) + ".CON", ep);
                    vecAliasXY.push_back(*it);
                }
                else
                {
                    U8_LLOG("");
                    throw std::runtime_error("Wrong alias file for motion_ctrl " + *it);
                }
            }
        }
    }
//    U9_LLOG(log_message.str());
//    inbeam_dep [ALIAS=inbeam_dep] ~ inbeam_tt [ALIAS=inbeam_tt] ~ pupil_lens_dep [... etc etc
    // ***** code sent by briegel
    // ep = Nice::Endpoint::fromString(std::string("[ALIAS=sci_w1]"));
    // node->setAtNode<Nice::Endpoint>("SCU2_W1.CON", ep);

    // ep = Nice::Endpoint::fromString(std::string("[ALIAS=sci_w2]"));
    // node->setAtNode<Nice::Endpoint>("SCU2_W2.CON", ep);
    // *****

    // +++++++
    // CREATING CALLBACK FUNCTIONS
    // +++++++

//    m_callbackFnSM = boost::bind(&niceCallback::callback_motion_complete_SM, m_ptrCallbackObjSM, _1,
//        boost::ref(m_ptrStateMachine), boost::ref(m_mtxStateMachine));
    m_callbackFnSM = boost::bind(&niceCallback::callback_motion_complete_SM, _1,
        boost::ref(m_ptrStateMachine), boost::ref(m_mtxStateMachine));
    //std::cout << __FUNCTION__ << &m_mtxStateMachine << std::endl;

    // bind to function, parameter of callback used by multiDevice
    // m_callbackFnIndi = boost::bind(&niceCallback::callback_motion_complete_indi,
    //                                m_ptrCallbackObjIndi,
    //                                _1, _2,
    //                                boost::ref(m_ptrLbtoIF));

    m_ptrMonSeq = boost::make_shared<IceUtil::Monitor<IceUtil::Mutex> >();
//    m_callbackFnCount = boost::bind(&niceCallback::callback_motion_complete_count, m_ptrCallbackObjCount, _1,
//        boost::ref(m_moving_devs), boost::ref(m_mtxCount), boost::ref(m_ptrMonSeq));
    m_callbackFnCount = boost::bind(&niceCallback::callback_motion_complete_count, _1,
        boost::ref(m_moving_devs), boost::ref(m_mtxCount), boost::ref(m_ptrMonSeq));

    //void callback_device_move_to_reference(const Nice::Completion& _c, const Basda::MoccaWheelServiceWorkerInterfacePtr& _device);

    // m_callbackFnIndi = boost::bind(&niceCallback::callback_motion_complete_indi,
    //                                m_ptrCallbackObjIndi,
    //                                _1,
    //                                boost::ref(m_ptrLbtoIF));
    // U2_LLOG(__FUNCTION__ << " Connecting to remote configuration tree ...")

    // Connecting to device
    if(vecAliasWheel.size() != 0)
    {
        // U2_LLOG(__FUNCTION__ << "... for wheels ...")
        // std::cout << "I have wheels!!" << std::endl;
        Nice::JointPtr nodeWheel2 = Nice::Application::config()->node("WHEEL");
        // Nice::JointPtr nodeWheel2Dummy = Nice::Application::config()->node("WHEEL-DUMMY");
        // Create multiWheel object
        m_ptrMultiWheel = boost::make_shared<multiWheel>(nodeWheel2, vecAliasWheel, _mapCfgFiles);        //,
        // _ptrLogger->cloneWithPrefix(_ptrLogger->getPrefix()));
        // m_ptrMultiWheel = boost::make_shared<multiWheel>(nodeWheel2, nodeWheel2Dummy, vecAliasWheel);

        m_ptrMultiWheel->connect();
        m_ptrMultiWheel->standby();
        m_ptrMultiWheel->online();

        // Setting callback function for multiWheel object
        m_ptrMultiWheel->set_callback_fn(m_callbackFnSM, "motion_complete_SM");
        m_ptrMultiWheel->set_callback_fn_state(m_callbackFnIndi, "motion_state_change"); // on callback foreach wheel
        m_ptrMultiWheel->set_callback_fn_count(m_callbackFnCount, "motion_complete_count");
        U2_LLOG(":callbacks set on rotation stages");
    }
    if(vecAliasXY.size() != 0)
    {
        // U2_LLOG(__FUNCTION__ << "... for XY stages ...")
        Nice::JointPtr nodeXY2 = Nice::Application::config()->node("STAGEXY");

        // Create multiStageXY object
        m_ptrMultiStageXY = boost::make_shared<multiStageXY>(nodeXY2, vecAliasXY, _mapCfgFiles); //,
        // _ptrLogger->cloneWithPrefix(_ptrLogger->getPrefix()));
        m_ptrMultiStageXY->connect();
        m_ptrMultiStageXY->standby();
        m_ptrMultiStageXY->online();

        // Setting callback function for multiStageXY object
        m_ptrMultiStageXY->set_callback_fn(m_callbackFnCount, "motion_complete_count");
        U2_LLOG("callbacks set on XY stages");
    }
    if(vecAliasX.size() != 0)
    {
        // U2_LLOG(__FUNCTION__ << "... for X stages ...")
        Nice::JointPtr nodeX2 = Nice::Application::config()->node("STAGEX");

        // Create multiStageX object
        m_ptrMultiStageX = boost::make_shared<multiStageX>(nodeX2, vecAliasX, _mapCfgFiles);        //,
        // _ptrLogger->cloneWithPrefix(_ptrLogger->getPrefix()));
        m_ptrMultiStageX->connect();
        m_ptrMultiStageX->standby();
        m_ptrMultiStageX->online();
        // Setting callback function for multiStageX object
        m_ptrMultiStageX->set_callback_fn(m_callbackFnSM, "motion_complete_SM");
        m_ptrMultiStageX->set_callback_fn_count(m_callbackFnCount, "motion_complete_count");
        U2_LLOG("callbacks set on X stages");
    }
    // m_sharedIndiDrv = std::make_shared<indiWheels>(devName, driverVersion, protocolVersion,
    //                                                m_ptrMultiWheel);
    // Initialiazing devices
    // moveHome();
    U8_LLOG(__FUNCTION__ << " : Done");
}


motionCtrl::~motionCtrl()
{
    //
    delete m_app;
    // for(auto it = m_vecSharedWheelApp.begin(); it < m_vecSharedWheelApp.end(); ++it)
    //   {
    //     (*it)->exit(0);
    //   }
}


double motionCtrl::ConvertToDefaultUnit(const std::string & alias, double destination, const std::string & unit)
{
    if(m_ptrMultiWheel->isDevice(alias))
    {
        std::map<std::string, StageXDevice>::const_iterator itDevice = GetMultiStageX().find(alias);
        if(itDevice != GetMultiStageX().end())
        {
            std::string final_unit = unit;
            if(unit == suc::default_unit)final_unit = suc::steps;
            std::map<std::string, std::tr1::tuple<double, double> >::const_iterator itUnit = itDevice->second.units.find(final_unit);
            if(itUnit != itDevice->second.units.end())
            {
                return destination/std::tr1::get<0>(itUnit->second) + std::tr1::get<1>(itUnit->second);
            }
            else
            {
                E_LLOG(std::string(__FUNCTION__) << ":No " << unit << " unit in " << alias << ". Throwing!");
                throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + unit + " unit in " + alias + ".");
            }
        }
        else
        {
            E_LLOG(std::string(__FUNCTION__) << ":No " << alias << " device in CALUNIT_CTRL. Throwing!");
            throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + alias + " device in CALUNIT_CTRL.");
        }
    }
    else if(m_ptrMultiStageXY->isDevice(alias))
    {
        E_LLOG("ERROR: " << __FUNCTION__ << " TBD");
        throw std::invalid_argument(std::string("ERROR: ") + __FUNCTION__ + " TBD");
        std::map<std::string, StageXYDevice>::const_iterator itDevice = GetMultiStageXY().find(alias);
        if(itDevice != GetMultiStageXY().end())
        {
            std::string final_unit = unit;
            if(unit == suc::default_unit)final_unit = suc::steps;
//            std::map<std::string, std::tr1::tuple<double, double> >::iterator itUnit = itDevice->second.units.find(final_unit);
//            if(itUnit != itDevice->second.units.end())
            {
                return destination;///std::tr1::get<0>(itUnit->second) + std::tr1::get<1>(itUnit->second);
            }
//            else
            {
                E_LLOG(std::string(__FUNCTION__) << ":No " << unit << " unit in " << alias << ". Throwing!");
                throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + unit + " unit in " + alias + ".");
            }
        }
        else
        {
            E_LLOG(std::string(__FUNCTION__) << ":No " << alias << " device in CALUNIT_CTRL. Throwing!");
            throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + alias + " device in CALUNIT_CTRL.");
        }
    }
    if(m_ptrMultiWheel->isDevice(alias))
    {
        std::map<std::string, WheelDevice>::const_iterator itDevice = GetMultiWheel().find(alias);
        if(itDevice != GetMultiWheel().end())
        {
            std::string final_unit = unit;
            if(unit == suc::default_unit)final_unit = suc::steps;
            std::map<std::string, std::tr1::tuple<double, double> >::const_iterator itUnit = itDevice->second.units.find(final_unit);
            if(itUnit != itDevice->second.units.end())
            {
                return destination/std::tr1::get<0>(itUnit->second) + std::tr1::get<1>(itUnit->second);
            }
            else
            {
                E_LLOG(std::string(__FUNCTION__) << ":No " << unit << " unit in " << alias << ". Throwing!");
                throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + unit + " unit in " + alias + ".");
            }
        }
        else
        {
            E_LLOG(std::string(__FUNCTION__) << ":No " << alias << " device in CALUNIT_CTRL. Throwing!");
            throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + alias + " device in CALUNIT_CTRL.");
        }
    }
    else
    {
        E_LLOG(std::string(__FUNCTION__) << ":No " << alias << " device in MOTION_CTRL. Throwing!");
        throw std::invalid_argument(std::string(__FUNCTION__) + ": No " + alias + " device in MOTION_CTRL.");
    }
}


void motionCtrl::SetSecondaryCallbacks(const std::string & alias,
                                     void (* MoveToHomeCB)(const Nice::Completion & c),
                                     void (* MoveToNamedCB)(const Nice::Completion & c),
                                     void (* MoveAbsoluteCB)(const Nice::Completion & c),
                                     void (* MoveRelativeCB)(const Nice::Completion & c),
                                     void (* MoveToLimitCB)(const Nice::Completion & c))
{
    if(m_ptrMultiWheel->isDevice(alias))
        m_ptrMultiWheel->SetSecondaryCallbacks(alias, MoveToHomeCB, MoveToNamedCB, MoveAbsoluteCB, MoveRelativeCB, MoveToLimitCB);
    else if(m_ptrMultiStageX->isDevice(alias))
        m_ptrMultiStageX->SetSecondaryCallbacks(alias, MoveToHomeCB, MoveToNamedCB, MoveAbsoluteCB, MoveRelativeCB, MoveToLimitCB);
    else if(m_ptrMultiStageXY->isDevice(alias))
        m_ptrMultiStageXY->SetSecondaryCallbacks(alias, MoveToHomeCB, MoveToNamedCB, MoveAbsoluteCB, MoveRelativeCB, MoveToLimitCB);
}


bool motionCtrl::isDevice(const std::string& _strDevAlias)
{
    return (m_ptrMultiWheel->isDevice(_strDevAlias) ||
        m_ptrMultiStageX->isDevice(_strDevAlias) ||
        m_ptrMultiStageXY->isDevice(_strDevAlias));
}


Ice::Int motionCtrl::getPosition(std::map<std::string, Ice::Double> &_mapPosition, const std::string &_unit)
{
//    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    status = m_ptrMultiWheel->getPosition(_mapPosition, _unit);
    status = m_ptrMultiStageX->getPosition(_mapPosition, _unit);
    status = m_ptrMultiStageXY->getPosition(_mapPosition, _unit);
//    U8_LLOG("");
    return status;
}


std::map<std::string, std::string> motionCtrl::getNamedPosition(const std::vector<std::string>& _vecAlias)
{
//    U6_LLOG(__FUNCTION__);
    std::map<std::string, std::string> mapWheels, mapStageX, mapStageXY, result;
//    std::vector<std::string>::const_iterator it;
//    for(it = _vecAlias.begin(); it < _vecAlias.end(); it++){ result[*it] = "";};
    std::map<std::string, std::string>::iterator it;
    U9_LLOG(__FUNCTION__ << " ~ MultiWheel");
    mapWheels = m_ptrMultiWheel->getNamedPosition2(_vecAlias);
    for(it = mapWheels.begin(); it != mapWheels.end(); it++)
    {
        U9_LLOG(it->first << " ~ " << it->second);
    }
    U9_LLOG(__FUNCTION__ << " ~ MultiStageX");
    mapStageX = m_ptrMultiStageX->getNamedPosition2(_vecAlias);
    for(it = mapStageX.begin(); it != mapStageX.end(); it++)
    {
        U9_LLOG(it->first << " ~ " << it->second);
    }
    U9_LLOG(__FUNCTION__ << " ~ MultiStageXY");
    mapStageXY = m_ptrMultiStageXY->getNamedPosition2(_vecAlias);
    for(it = mapStageXY.begin(); it != mapStageXY.end(); it++)
    {
        U9_LLOG(it->first << " ~ " << it->second);
    }
    result.insert(mapWheels.begin(), mapWheels.end());
    result.insert(mapStageX.begin(), mapStageX.end());
    result.insert(mapStageXY.begin(), mapStageXY.end());
//    U8_LLOG("");
    return result;
}


// return number of devices controlled
int motionCtrl::getNdevices()
{
    U6_LLOG(__FUNCTION__);
    int result = (m_ptrMultiWheel->getNdevices() + m_ptrMultiStageX->getNdevices() + m_ptrMultiStageXY->getNdevices());
    U8_LLOG("");
    return result;
}


void motionCtrl::moveHome()
{
    U6_LLOG(__FUNCTION__);

    m_ptrStateMachine->process_event(
        osfsm::ready_busy(
            m_ptrMultiWheel->getNdevices() + m_ptrMultiStageX->getNdevices() + m_ptrMultiStageXY->getNdevices()));
    // U8_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));

    if(m_ptrMultiWheel->getNdevices() != 0)
    {
        m_ptrMultiWheel->moveToReference();
    }
    if(m_ptrMultiStageX->getNdevices() != 0)
    {
        m_ptrMultiStageX->moveToReference();
    }
    if(m_ptrMultiStageXY->getNdevices() != 0)
    {
        m_ptrMultiStageXY->moveToReference();
    }
    U8_LLOG(__FUNCTION__);
}


void motionCtrl::changeState(const Basda::ServiceState &_state, const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    switch (_state)
    {
    case Basda::ONLINE:
    {
        m_ptrMultiWheel->online(_vecDevAlias);
        m_ptrMultiStageX->online(_vecDevAlias);
        m_ptrMultiStageXY->online(_vecDevAlias);
        break;
    }
    case Basda::STANDBY:
    {
        m_ptrMultiWheel->online(_vecDevAlias);
        m_ptrMultiStageX->online(_vecDevAlias);
        m_ptrMultiStageXY->online(_vecDevAlias);
        break;
    }
    case Basda::SHUTDOWN:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::OFFLINE:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::WORKING:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    case Basda::ERROR:
    {
        W_LLOG("State " << _state << " not handled");
        break;
    }
    default:
        W_LLOG("State " << _state << " not handled")
        ;
    }
    U8_LLOG("");
}


// Performs initialization of motion stages
// move wheels to reference
// move stages X to referece
// move stage XY to referece
// wait wheels to reference
// wait stages X to referece
// wait stage XY to referece
// move wheels to reference
// wait wheels to reference

void motionCtrl::initializeAll()
{
    U6_LLOG(__FUNCTION__);
    try
    {
        initializeAllStart();
        initializeAllWait(Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::initializeAllStart()
{
    U6_LLOG(__FUNCTION__);
    // moving to "home" reference switch all devices
    // the number of moving wheels is multiplied yby 2 since they move to reference twice
//    m_ptrStateMachine->process_event(osfsm::ready_busy(m_ptrMultiStageX->getNdevices()));

    m_moving_devs = 2 * m_ptrMultiWheel->getNdevices() + m_ptrMultiStageX->getNdevices() + m_ptrMultiStageXY->getNdevices();

    try
    {
        if(m_ptrMultiWheel->getNdevices() != 0)
        {
            // wheels must be moved to home reference switch
            // twice to reach in spech repeatability. The second initialization
            // is implemented in initializeWait
            m_ptrMultiWheel->moveToReferenceStart();
            U9_LLOG("Moving rotation stages to reference ...");

        }
        if(m_ptrMultiStageX->getNdevices() != 0)
        {
            m_ptrMultiStageX->moveToReferenceStart();
            U9_LLOG("Moving X stages to reference ...");

        }
        if(m_ptrMultiStageXY->getNdevices() != 0)
        {
            m_ptrMultiStageXY->moveToReferenceStart();
            U9_LLOG("Moving rotation XY to reference ...");

        }
    }
    catch(std::exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw e;
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::initializeAllWait(const Nice::Time& _time)
{
    U6_LLOG(__FUNCTION__);
    su::Timer timer(_time);
    try
    {
        if(m_ptrMultiWheel->getNdevices() != 0)
        {
            m_ptrMultiWheel->moveToReferenceWait(timer.residual());
        }
        if(m_ptrMultiStageX->getNdevices() != 0)
        {
            m_ptrMultiStageX->moveToReferenceWait(timer.residual());
        }
        if(m_ptrMultiStageXY->getNdevices() != 0)
        {
            m_ptrMultiStageXY->moveToReferenceWait(timer.residual());
        }
    }
    catch(std::exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw e;
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::initialize(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        initializeStart(_vecDevAlias);
        initializeWait(Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::initPrepare(const std::vector<std::string> & _vecDevAlias)
{
    m_init.clear();
    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiWheel->isDevice(*it))
        {
            m_init.InitWheels.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(*it))
        {
            m_init.InitStageX.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageXY->isDevice(*it))
        {
            m_init.InitStageXY.push_back(*it);
            continue;
        }
    }
}

    // Moves all or specified devices to start positions
void motionCtrl::initializeStart(const std::vector<std::string> &_vecDevAlias)
{

    // LOG_SYSLOG;
    U6_LLOG(__FUNCTION__ << " Initialiazing motion devices");
    initPrepare(_vecDevAlias);
    std::string logMsg;

    m_ptrStateMachine->process_event(
        osfsm::ready_busy(
            m_ptrMultiWheel->getNdevices() + m_ptrMultiStageX->getNdevices() + m_ptrMultiStageXY->getNdevices()));


    try
    {
        if((m_ptrMultiStageX->getNdevices() > 0) && !(m_init.InitStageX.empty()))
            m_ptrMultiStageX->moveDeviceToReferenceStart(m_init.InitStageX);
    }
    catch(Nice::Exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw(e);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    try
    {
        if((m_ptrMultiWheel->getNdevices() > 0) && !(m_init.InitWheels.empty()))
            m_ptrMultiWheel->moveDeviceToReferenceStart(m_init.InitWheels);
    }
    catch(Nice::Exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw(e);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    try
    {
        if((m_ptrMultiStageXY->getNdevices() > 0) && !(m_init.InitStageXY.empty()))
            m_ptrMultiStageXY->moveDeviceToReferenceStart(m_init.InitStageXY);
    }
    catch(Nice::Exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw(e);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::initializeWait(const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__ << " - Waiting device initialization");
    su::Timer timer(_time);

    // filter input vector
    std::vector<std::string>::const_iterator it;

    try
    {
        if((m_ptrMultiWheel->getNdevices() > 0) && !(m_init.InitWheels.empty()))
        {
            m_ptrMultiWheel->moveDeviceToReferenceWait(m_init.InitWheels, timer.residual());
        }

        if((m_ptrMultiStageX->getNdevices() > 0) && !(m_init.InitStageX.empty()))
        {
            m_ptrMultiStageX->moveDeviceToReferenceWait(m_init.InitStageX, timer.residual());
        }
        if((m_ptrMultiStageXY->getNdevices() > 0) && !(m_init.InitStageXY.empty()))
        {
            m_ptrMultiStageXY->moveDeviceToReferenceWait(m_init.InitStageXY, timer.residual());
        }
//        if((m_ptrMultiWheel->getNdevices() > 0) && !(m_init.InitWheels.empty()))
//        {
//            m_ptrMultiWheel->moveDeviceToReference(m_init.InitWheels);
//        }
    }
    catch(std::exception e)
    {
        E_LLOG(e.what());
        U8_LLOG("");
        throw e;
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::moveDeviceHome(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);

    std::vector<std::string> tmpVecWheel;
    std::vector<std::string> tmpVecStageX;
    std::vector<std::string> tmpVecStageXY;
    std::string logMsg = "";
    boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
    boost::chrono::steady_clock::duration deltat;

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiWheel->isDevice(*it))
        {
            tmpVecWheel.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageXY->isDevice(*it))
        {
            tmpVecStageXY.push_back(*it);
            continue;
        }
    }
    // swich state machine to "busy"
    m_ptrStateMachine->process_event(osfsm::ready_busy(tmpVecWheel.size()));
    try
    {
        m_ptrMultiWheel->moveDeviceToReference(tmpVecWheel);
        m_ptrMultiStageX->moveDeviceToReference(tmpVecStageX);
        m_ptrMultiStageXY->moveDeviceToReference(tmpVecStageXY);

        su::Timer timer(Nice::Time::seconds(80));
        m_ptrMultiWheel->moveDeviceToReferenceWait(tmpVecWheel, timer.residual());
        m_ptrMultiStageX->moveDeviceToReferenceWait(tmpVecStageX, timer.residual());
        m_ptrMultiStageXY->moveDeviceToReferenceWait(tmpVecStageXY, timer.residual());
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    if((deltat = shins::util::timer(m_moving_devs, delay)) > delay)
    {
        logMsg = "TIMED OUT in motion to named position!";
        E_LLOG(logMsg);
        U8_LLOG("");
        throw std::runtime_error("motion_ctrl TIMED OUT in motion to named position.  Elapsed time = " + shins::util::convertTypeToString(deltat));
    }
    U8_LLOG("");
}


int motionCtrl::moveDeviceToNegativeLimit(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    std::vector<std::string> tmpVecStageX;
    std::string logMsg = "";

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
        else
        {
            E_LLOG("Error: tryed to move " << *it << " to negative limit. Only Motion Stage X allowed for this operation");
            continue;
        }
    }
    // swich state machine to "busy"
//    m_ptrStateMachine->process_event(osfsm::ready_busy(tmpVecWheel.size()));


    try
    {
        m_ptrMultiStageX->moveToNegativeLimitStart(tmpVecStageX);
        m_ptrMultiStageX->moveToNegativeLimitWait(tmpVecStageX, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }


    U8_LLOG("");
    return EXIT_SUCCESS;
}


int motionCtrl::moveDeviceToPositiveLimit(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    std::vector<std::string> tmpVecStageX;
    std::string logMsg = "";

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
        else
        {
            E_LLOG("Error: tryed to move " << *it << " to positive limit. Only Motion Stage X allowed for this operation");
            continue;
        }
    }
    // swich state machine to "busy"
//    m_ptrStateMachine->process_event(osfsm::ready_busy(tmpVecWheel.size()));

    try
    {
        m_ptrMultiStageX->moveToPositiveLimitStart(tmpVecStageX);
        m_ptrMultiStageX->moveToPositiveLimitWait(tmpVecStageX, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }

    U8_LLOG("");
    return EXIT_SUCCESS;
}
void motionCtrl::setup(const std::map<std::string, std::string> &_mapNamedPos)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        setupStart(_mapNamedPos);
        setupWait(_mapNamedPos, Nice::Time::seconds(180));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::setupPrepare(const std::map<std::string, std::string>& _mapNamedPos, motionSetup & setup)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, std::tr1::tuple<int, int> > NamedPosMap;
    std::string svcNameXY = "";
    int idxPos;
    std::map<std::string, std::string>::const_iterator it;
    shins::util::EShinsMovements detected;
    double absolute, relative = 0;
    std::string unit;
    setup.clear();
    for(it = _mapNamedPos.begin(); it != _mapNamedPos.end(); ++it)
    {
        if(m_ptrMultiWheel->isDevice(it->first))
        {
            detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            switch(detected)
            {
            case shins::util::SHINS_MOV_NAMED:
                NamedPosMap = m_mainMapNamedPos[it->first];
                if(NamedPosMap.find(it->second) != NamedPosMap.end())
                    idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                else
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
                setup.NamedMapWheel.insert(std::pair<std::string, int>(it->first, idxPos));
                break;
            case shins::util::SHINS_MOV_ABSOLUTE:
                U9_LLOG("Detected Absolute Position : " << it->first << " -> " << it->second << "; value = " << absolute << "; unit = " << unit);
                setup.AbsMapWheel.insert(std::pair<std::string, double>(it->first, absolute));
                setup.AbsMapWheelUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_RELATIVE:
                U9_LLOG("Detected Relative Position : " << it->first << " -> " << it->second << "; value = " << relative << "; unit = " << unit);
                setup.RelMapWheel.insert(std::pair<std::string, double>(it->first, relative));
                setup.RelMapWheelUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_UNDEFINED:
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    U9_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
            }
            continue;
        }
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            switch(detected)
            {
            case shins::util::SHINS_MOV_NAMED:
                NamedPosMap = m_mainMapNamedPos[it->first];
                if(NamedPosMap.find(it->second) != NamedPosMap.end())
                    idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                else
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
                setup.NamedMapStageX.insert(std::pair<std::string, int>(it->first, idxPos));
                NamedPosMap.clear();
                break;
            case shins::util::SHINS_MOV_ABSOLUTE:
                U9_LLOG("Detected Absolute Position : " << it->first << "-> " << it->second << "; value = " << absolute);
                setup.AbsMapStageX.insert(std::pair<std::string, double>(it->first, absolute));
                setup.AbsMapStageXUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_RELATIVE:
                U9_LLOG("Detected Relative Position : " << it->first << "-> " << it->second << "; value = " << relative);
                setup.RelMapStageX.insert(std::pair<std::string, double>(it->first, relative));
                setup.RelMapStageXUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_UNDEFINED:
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
            }
            continue;
        }
        if(boost::algorithm::starts_with(it->first, "INBEAM_TT"))
        {
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            std::string first_end = shins::util::right(it->first, 2);
            if(first_end == "_X" || first_end == "_Y")
            {
                detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
                switch(detected)
                {
                case shins::util::SHINS_MOV_NAMED:
                case shins::util::SHINS_MOV_UNDEFINED:
                    {
                        std::stringstream msg;
                        msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                        E_LLOG(msg.str());
                        U8_LLOG("");
                        throw(std::invalid_argument(msg.str()));
                    }
                    break;
                case shins::util::SHINS_MOV_ABSOLUTE:
                    setup.AbsMapStageXY.insert(std::pair<std::string, double>(it->first, absolute));
                    setup.AbsMapStageXYUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                    break;
                case shins::util::SHINS_MOV_RELATIVE:
                    setup.AbsMapStageXY.insert(std::pair<std::string, double>(it->first, absolute));
                    setup.AbsMapStageXYUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                    break;
                }
            }
            else
            {
                if(m_ptrMultiStageXY->isDevice(it->first))
                {
                    NamedPosMap = m_mainMapNamedPos[it->first];
                    if(NamedPosMap.find(it->second) != NamedPosMap.end())
                        idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                    else
                    {
                        std::stringstream msg;
                        msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                        E_LLOG(msg.str());
                        U8_LLOG("");
                        throw(std::invalid_argument(msg.str()));
                    }
                    setup.NamedMapStageXY.insert(std::pair<std::string, int>(it->first, idxPos));
                }
            }
        }
    }
    U8_LLOG("");
}


void motionCtrl::setupStart(const std::map<std::string, std::string> &_mapNamedPos)
{
    U6_LLOG(__FUNCTION__);
    setupPrepare(_mapNamedPos, m_setup);
    //m_mtxSeq.lock();
    // swich state machine to "busy"
    // U2_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));
    m_moving_devs = m_setup.NamedMapWheel.size() + m_setup.NamedMapStageX.size();
    U9_LLOG(__FUNCTION__ << " ~ calling process event to SM");
    m_ptrStateMachine->process_event(osfsm::ready_busy(m_setup.NamedMapWheel.size() + m_setup.NamedMapStageX.size()));    //+
    U9_LLOG(__FUNCTION__ << " ~ Finished calling process event to SM");
    // U2_LLOG("State Machine state: " << *(m_ptrStateMachine->current_state()));
    // U2_LLOG(*(m_ptrStateMachine->event_(osfsm::readyDone)));
    // IceUtil::Monitor<IceUtil::Mutex>::Lock lock(*(m_ptrMonSeq));
    try
    {
        m_ptrMultiWheel->moveToNamedPositionStart(m_setup.NamedMapWheel);
        m_ptrMultiWheel->moveToAbsolutePositionStart(m_setup.AbsMapWheel, m_setup.AbsMapWheelUnits);
        m_ptrMultiWheel->moveToRelativePositionStart(m_setup.RelMapWheel, m_setup.RelMapWheelUnits);
        m_ptrMultiStageX->moveToNamedPositionStart(m_setup.NamedMapStageX);
        m_ptrMultiStageX->moveToAbsolutePositionStart(m_setup.AbsMapStageX, m_setup.AbsMapStageXUnits);
        m_ptrMultiStageX->moveToRelativePositionStart(m_setup.RelMapStageX, m_setup.RelMapStageXUnits);
        m_ptrMultiStageXY->moveToNamedPositionStart(m_setup.NamedMapStageXY);
        m_ptrMultiStageXY->moveToAbsolutePositionStart(m_setup.AbsMapStageXY, m_setup.AbsMapStageXYUnits);
        m_ptrMultiStageXY->moveToRelativePositionStart(m_setup.RelMapStageXY, m_setup.RelMapStageXYUnits);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    
    U8_LLOG("");
}


void motionCtrl::setupWait(const std::map<std::string, std::string> &_mapNamedPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__);
    setupPrepare(_mapNamedPos, m_setupWait);
    su::Timer timer(_time);
//    boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
//    boost::chrono::steady_clock::duration deltat;

    try
    {
        m_ptrMultiWheel->moveToNamedPositionWait(m_setupWait.NamedMapWheel, timer.residual());
        m_ptrMultiWheel->moveToAbsolutePositionWait(m_setupWait.AbsMapWheel, timer.residual());
        m_ptrMultiWheel->moveToRelativePositionWait(m_setupWait.RelMapWheel, timer.residual());
        m_ptrMultiStageX->moveToNamedPositionWait(m_setupWait.NamedMapStageX, timer.residual());
        m_ptrMultiStageX->moveToAbsolutePositionWait(m_setupWait.AbsMapStageX, timer.residual());
        m_ptrMultiStageX->moveToRelativePositionWait(m_setupWait.RelMapStageX, timer.residual());
        m_ptrMultiStageXY->moveToNamedPositionWait(m_setupWait.NamedMapStageXY, timer.residual());
        m_ptrMultiStageXY->moveToAbsolutePositionWait(m_setupWait.AbsMapStageXY, timer.residual());
        m_ptrMultiStageXY->moveToRelativePositionWait(m_setupWait.RelMapStageXY, timer.residual());
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }

    U8_LLOG("motionCtrl::" << __FUNCTION__);
}


// moving to arbitrary position in steps. NOTE that here the input map can
// contain the position of INBEAT_TT expressed with X and Y coordinates
void motionCtrl::moveToAbsolutePosition(const std::map<std::string, double> &_mapPos, const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        moveToAbsolutePositionStart(_mapPos, _mapUnit);
        moveToAbsolutePositionWait(_mapPos, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::moveToAbsolutePositionStart(const std::map<std::string, double> &_mapPos,
    const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, double> AbsMapWheel;
    std::map<std::string, double> AbsMapStageX;
    std::map<std::string, double> AbsMapStageXY;
    std::string strAlias = "";
    int movingDevices = 0;
    // boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
//    boost::chrono::steady_clock::duration deltat;
    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        D_LLOG(__FUNCTION__ << ":" << __LINE__ << ":" << it->first);

        if(m_ptrMultiWheel->isDevice(it->first))
        {
            AbsMapWheel.insert(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            AbsMapStageX.insert(*it);
            continue;
        }
        if(boost::algorithm::starts_with(it->first, "INBEAM_TT_"))
        {
            strAlias = "INBEAM_TT";
            // if the service INBEAM_TT is active then add the two coodinates to
            // the temporary map for motion.
            if(m_ptrMultiStageXY->isDevice(strAlias))
            {
                // std::cout << "stageXY" << std::endl;
                // to move XY stage we need X and Y positions
                AbsMapStageXY.insert(*it);
                continue;
            }
        }
    }
    // switch state machine to "busy"
    movingDevices = AbsMapWheel.size() + AbsMapStageX.size() + (AbsMapStageXY.size() / 2); // the XY stage equals to 1 device
    m_ptrStateMachine->process_event(osfsm::ready_busy(movingDevices));
    m_moving_devs = AbsMapWheel.size() + AbsMapStageX.size() + (AbsMapStageXY.size() / 2);
    try
    {
        if(AbsMapWheel.size() > 0)
        {
            m_ptrMultiWheel->moveToAbsolutePositionStart(AbsMapWheel, _mapUnit);
        }
        if(AbsMapStageX.size() > 0)
        {
            m_ptrMultiStageX->moveToAbsolutePositionStart(AbsMapStageX, _mapUnit);
        }
        if(AbsMapStageXY.size() > 0)
        {
            m_ptrMultiStageXY->moveToAbsolutePositionStart(AbsMapStageXY, _mapUnit);
        }
    }
    catch(Basda::Svc::MoccaMotionException basdaExecp)
    {
        E_LLOG(basdaExecp.what());
        U8_LLOG("");
        throw(basdaExecp);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }

    U8_LLOG("");
}


void motionCtrl::moveToAbsolutePositionWait(const std::map<std::string, double> &_mapPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, double> AbsMapWheel;
    std::map<std::string, double> AbsMapStageX;
    std::map<std::string, double> AbsMapStageXY;
    std::string strAlias = "";
    su::Timer timer(_time);
    // int movingDevices = 0;
    // boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
//    boost::chrono::steady_clock::duration deltat;

    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        D_LLOG("_mapPos.first : " << it->first);

        if(m_ptrMultiWheel->isDevice(it->first))
        {
            // std::cout << "wheel" << std::endl;
            AbsMapWheel.insert(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            // std::cout << "stageX" << std::endl;
            AbsMapStageX.insert(*it);
            continue;
        }
        if(boost::algorithm::starts_with(it->first, "INBEAM_TT_"))
        {
            strAlias = "INBEAM_TT";
            // if the service INBEAM_TT is active then add the two coodinates to
            // the temporary map for motion.
            if(m_ptrMultiStageXY->isDevice(strAlias))
            {
                // std::cout << "stageXY" << std::endl;
                // to move XY stage we need X and Y positions
                AbsMapStageXY.insert(*it);
                continue;
            }
        }
    }
    try
    {
        if(AbsMapWheel.size() > 0)
            m_ptrMultiWheel->moveToAbsolutePositionWait(AbsMapWheel, timer.residual());
        if(AbsMapStageX.size() > 0)
            m_ptrMultiStageX->moveToAbsolutePositionWait(AbsMapStageX, timer.residual());
        if(AbsMapStageXY.size() > 0)
            m_ptrMultiStageXY->moveToAbsolutePositionWait(AbsMapStageXY, timer.residual());
    }
    catch(Basda::Svc::MoccaMotionException basdaExecp)
    {
        E_LLOG(basdaExecp.what());
        U8_LLOG("");
        throw(basdaExecp);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG(__FUNCTION__ << " Done");
}


void motionCtrl::moveToRelativePosition(const std::map<std::string, double> &_mapPos, const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        moveToRelativePositionStart(_mapPos, _mapUnit);
        moveToRelativePositionWait(_mapPos, Nice::Time::seconds(80));
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::moveToRelativePositionStart(const std::map<std::string, double> &_mapPos,
    const std::map<std::string, std::string> &_mapUnit)
{
    U6_LLOG(__FUNCTION__);

    std::map<std::string, double> RelMapWheel;
    std::map<std::string, double> RelMapStageX;
    std::map<std::string, double> RelMapStageXY;
    std::string strAlias = "";
    int movingDevices = 0;
    // boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
//  boost::chrono::steady_clock::duration deltat;

    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        D_LLOG("_mapPos.first :" << it->first);

        if(m_ptrMultiWheel->isDevice(it->first))
        {
            RelMapWheel.insert(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            RelMapStageX.insert(*it);
            continue;
        }
        if(boost::algorithm::starts_with(it->first, "INBEAM_TT_"))
        {
            strAlias = "INBEAM_TT";
            // if the service INBEAM_TT is active then add the two coordinates to
            // the temporary map for motion.
            if(m_ptrMultiStageXY->isDevice(strAlias))
            {
                // std::cout << "stageXY" << std::endl;
                // to move XY stage we need X and Y positions
                RelMapStageXY.insert(*it);
                continue;
            }
        }
    }
    // switch state machine to "busy"
    movingDevices = RelMapWheel.size() + RelMapStageX.size() + (RelMapStageXY.size() / 2); // the XY stage equals to 1 device
    m_ptrStateMachine->process_event(osfsm::ready_busy(movingDevices));
    m_moving_devs = RelMapWheel.size() + RelMapStageX.size() + (RelMapStageXY.size() / 2);
    try
    {
        if(RelMapWheel.size() > 0)
        {
            m_ptrMultiWheel->moveToRelativePositionStart(RelMapWheel, _mapUnit);
        }
        if(RelMapStageX.size() > 0)
        {
            m_ptrMultiStageX->moveToRelativePositionStart(RelMapStageX, _mapUnit);
        }
        if(RelMapStageXY.size() > 0)
        {
            m_ptrMultiStageXY->moveToRelativePositionStart(RelMapStageXY, _mapUnit);
        }
    }
    catch(Basda::Svc::MoccaMotionException basdaExecp)
    {
        E_LLOG(basdaExecp.what());
        U8_LLOG("");
        throw(basdaExecp);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void motionCtrl::moveToRelativePositionWait(const std::map<std::string, double> &_mapPos, const Nice::Time &_time)
{
    U6_LLOG(__FUNCTION__ << " : Implements moveToRelativePositionWait");
    std::map<std::string, double> RelMapWheel;
    std::map<std::string, double> RelMapStageX;
    std::map<std::string, double> RelMapStageXY;
    std::string strAlias = "";
    su::Timer timer(_time);
    // int movingDevices = 0;
    // boost::chrono::steady_clock::duration delay = boost::chrono::seconds(120);
//  boost::chrono::steady_clock::duration deltat;

    std::map<std::string, double>::const_iterator it;
    for(it = _mapPos.begin(); it != _mapPos.end(); ++it)
    {
        D_LLOG("_mapPos.first : " << it->first);

        if(m_ptrMultiWheel->isDevice(it->first))
        {
            // std::cout << "wheel" << std::endl;
            RelMapWheel.insert(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(it->first))
        {
            // std::cout << "stageX" << std::endl;
            RelMapStageX.insert(*it);
            continue;
        }
        if(boost::algorithm::starts_with(it->first, "INBEAM_TT_"))
        {
            strAlias = "INBEAM_TT";
            // if the service INBEAM_TT is active then add the two coodinates to
            // the temporary map for motion.
            if(m_ptrMultiStageXY->isDevice(strAlias))
            {
                // std::cout << "stageXY" << std::endl;
                // to move XY stage we need X and Y positions
                RelMapStageXY.insert(*it);
                continue;
            }
        }
    }
    try
    {
        if(RelMapWheel.size() > 0)
            m_ptrMultiWheel->moveToRelativePositionWait(RelMapWheel, timer.residual());
        if(RelMapStageX.size() > 0)
            m_ptrMultiStageX->moveToRelativePositionWait(RelMapStageX, timer.residual());
        if(RelMapStageXY.size() > 0)
            m_ptrMultiStageXY->moveToRelativePositionWait(RelMapStageXY, timer.residual());
    }
    catch(Basda::Svc::MoccaMotionException basdaExecp)
    {
        E_LLOG(basdaExecp.what());
        U8_LLOG("");
        throw(basdaExecp);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG(__FUNCTION__);
}

// int motionCtrl::run(int argc, char* argv [])
// {
//   for(auto it = m_.begin(); it < m_vecSharedWheelApp.end(); ++it)
//     {
//       (*it)->run(argc, argv);
//     }
// }

void motionCtrl::emergencyAbort()
{
    U6_LLOG(__FUNCTION__);
    if(m_ptrMultiStageX->getNdevices() > 0)
        m_ptrMultiStageX->emergencyAbort();
    if(m_ptrMultiWheel->getNdevices() > 0)
        m_ptrMultiWheel->emergencyAbort();
    if(m_ptrMultiStageXY->getNdevices() > 0)
        m_ptrMultiStageXY->emergencyAbort();
    U8_LLOG(__FUNCTION__ << " Emergency Abort!");
}


int motionCtrl::stop(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__ << " Stopping selected devices");
    std::vector<std::string> tmpVecWheel;
    std::vector<std::string> tmpVecStageX;
    std::vector<std::string> tmpVecStageXY;
    int status = EXIT_SUCCESS;

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiWheel->isDevice(*it))
        {
            tmpVecWheel.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageXY->isDevice(*it))
        {
            tmpVecStageXY.push_back(*it);
            continue;
        }
    }

    try
    {
        if(tmpVecStageX.size())
            m_ptrMultiStageX->stop(tmpVecStageX);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }

    try
    {
        if(tmpVecWheel.size())
            m_ptrMultiWheel->stop(tmpVecWheel);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }

    try
    {
        if(tmpVecStageXY.size())
            m_ptrMultiStageXY->stop(tmpVecStageXY);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
    }
    U8_LLOG("");
    return status;
}


int motionCtrl::abort(const std::vector<std::string> &_vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    std::vector<std::string> tmpVecWheel;
    std::vector<std::string> tmpVecStageX;
    std::vector<std::string> tmpVecStageXY;
    int status = EXIT_SUCCESS;

    std::vector<std::string>::const_iterator it;
    for(it = _vecDevAlias.begin(); it != _vecDevAlias.end(); ++it)
    {
        if(m_ptrMultiWheel->isDevice(*it))
        {
            tmpVecWheel.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageX->isDevice(*it))
        {
            tmpVecStageX.push_back(*it);
            continue;
        }
        if(m_ptrMultiStageXY->isDevice(*it))
        {
            tmpVecStageXY.push_back(*it);
            continue;
        }
    }

    try
    {
        if(!tmpVecStageX.empty())
            m_ptrMultiStageX->abort(tmpVecStageX);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }

    try
    {
        if(!tmpVecWheel.empty())
            m_ptrMultiWheel->abort(tmpVecWheel);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }

    try
    {
        if(!tmpVecStageXY.empty())
            m_ptrMultiStageXY->abort(tmpVecStageXY);
    }
    catch(...)
    {
        E_LLOG(__FUNCTION__);
        U8_LLOG("");
        throw;
    }
    A_LLOG(__FUNCTION__ << " Called abort method on selected devices");

    U8_LLOG("");
    return status;
}


// The method returns a STL vector containing endpoint names as [ALIAS=aliasName].
// It scans the alias file looking for the [ALIAS= substring and extracts
// the substring between the '=' and the first ','
// an example of alias string is
// motion_ctrl-svr.nd_w.ALIAS:Endpoint=[ALIAS=nd_w, NAME=motion_ctrl, PATH=MOTION_CTRL.ND_W.SVC]
// string _strFileName: path to aliases file
// string _component:   sw componet, which can assume specific values: motion_ctrl, calunit_ctrl, track_ctrl
std::vector<std::string> motionCtrl::getAliasesFromFile(const std::string &_strFileName, const std::string &_component)
{
    U6_LLOG(__FUNCTION__ << " ~ " << _strFileName);
    std::ifstream ifsAlias(_strFileName.c_str(), std::ifstream::in);

    std::vector<std::string> vecAlias;
    std::vector<std::string> tmpVecSplit;
    std::vector<std::string>::const_iterator itFind;

    std::string strLine = "";
    std::string strAlias = "";
    std::string strTokens = "[+#,=:\"";
    std::string logMsg;

    while (std::getline(ifsAlias, strLine))
    {
        if(boost::algorithm::starts_with(strLine, _component) &&              //
            !(boost::algorithm::contains(strLine, "LABTEST")) &&              //
            (boost::algorithm::contains(strLine, "ALIAS="))) // excluding alias endpoint for server
        {
            tmpVecSplit = shins::util::split_with_boost(strLine, strTokens);
            itFind = std::find(tmpVecSplit.begin(), tmpVecSplit.end(), "ALIAS");
            if(itFind < tmpVecSplit.end())
            {
                ++itFind;
                vecAlias.push_back(*itFind);
                tmpVecSplit.clear();
                // U2_LLOG(vecAlias.back());
                continue;
            }
            else
            {
                logMsg = "Did not find alias in line: " + strLine;
                E_LLOG(logMsg);
                U8_LLOG("");
                throw std::runtime_error(logMsg);
            }
        }
    }
    ifsAlias.close();
    U8_LLOG(__FUNCTION__);
    return vecAlias;
}

void LogBoostDiagnosticInformationMotion(int line)
{
    std::string what = boost::current_exception_diagnostic_information();
    std::string toErase = "std::exception::what:";
    size_t pos = what.find(toErase);
    if(pos != std::string::npos)what.erase(pos, toErase.length());
    what.erase(std::remove(what.begin(), what.end(), '\n'), what.end());
    std::stringstream log_msg;
    log_msg << "Line: " << line << " ~ " << what;
    E_LLOG(log_msg.str());
}


int motionCtrl::getJsonStatus(std::string & jsonStatus)
{
    int status = EXIT_SUCCESS;

    std::map<std::string, StageXDevice>::const_iterator itStageX;
    const std::map<std::string, StageXDevice> & MapStageX = m_ptrMultiStageX->GetMultiStageX();
    int counter = 0;
    std::stringstream catch_msg;
    for(itStageX = MapStageX.begin(); itStageX != MapStageX.end(); itStageX++)
    {
        counter++;
        bool inLimit = false;
        bool isMoving = false;
        std::string service_name;
        try{service_name = itStageX->second.iPtr->readServiceName();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ readServiceName"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        try
        {
            isMoving = itStageX->second.iPtr->isMoving();
        }
        // catch(Nice::ChatException err)
        // {
        //     usleep(1000000);
        //     continue;
        // }
        catch(...)
        {
            catch_msg.str("");
            catch_msg << service_name << " ~ isMoving";
            MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());
            usleep(1000000);
            continue;
        }
        try{inLimit = itStageX->second.iPtr->isAtLimit();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isAtLimit"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        std::string named_pos = itStageX->second.iPtr->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
        bool isSimulated = false;
        Nice::SeqBitSet switch_status;
        // try{switch_status = itStageX->second.iPtr->getPositionSwitchStatus(); isSimulated = switch_status[0].getDepth() < 32;}catch(...){E_LLOG(service_name << " ~ getPositionSwitchStatus"); }
//        try{U9_LLOG("getPositionSwitchStatus : " << switch_status << " ~ " << switch_status[0].getDepth());}catch(...){E_LLOG("getPositionSwitchStatus");}
        bool isAtHome = false;
        try{isAtHome = itStageX->second.iPtr->isAtHome();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isAtHome"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//         if(itStageX->second.iPtr->isAtLimit())
//         {
//             try{U9_LLOG("isAtLimit2 : " << itStageX->second.iPtr->get<int>("VAR.LIMIT"));}catch(...){E_LLOG("isAtLimit2");}
//             std::string limit_result = itStageX->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             itStageX->second.iPtr->config()->setAtNode<std::string>("VAR.LIMIT", "-2");
//             limit_result = itStageX->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             limit_result = itStageX->second.iPtr->config()->leaf("VAR.LIMIT")->get<std::string>();
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
// //                val.limit = m_node->getAtNode<double> ("LIMIT")
//             double limit_double = itStageX->second.iPtr->config()->getAtNode<double>("LIMIT");
//             U9_LLOG("PropertyName : " << "LIMIT" << "; limit_double : " << limit_double);
//         }
//        try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
//        try{U9_LLOG("getNamedPosition : " << itStageX->second.iPtr->getNamedPosition(1));}catch(...){E_LLOG("getNamedPosition");}
//        try{U9_LLOG("readPosition : " << itStageX->second.iPtr->readPosition());}catch(...){E_LLOG("readPosition");}
//        try{U9_LLOG("readDeviceEncoderPosition : " << itStageX->second.iPtr->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
//        try{U9_LLOG("readAbsoluteEncoderPosition : " << itStageX->second.iPtr->readAbsoluteEncoderPosition());}catch(...){E_LLOG("readAbsoluteEncoderPosition");}
//        try{U9_LLOG("readIncrementalEncoderPosition : " << itStageX->second.iPtr->readIncrementalEncoderPosition());}catch(...){E_LLOG("readIncrementalEncoderPosition");}
        bool isReachable = false;
        try{isReachable = itStageX->second.iPtr->isReachable();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isReachable"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getPosition : " << itStageX->second.iPtr->getPosition());}catch(...){E_LLOG("getPosition");}
        int deviceEncoderPosition;
        try{deviceEncoderPosition = itStageX->second.iPtr->getDeviceEncoderPosition();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ getDeviceEncoderPosition"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getAbsoluteEncoderPosition : " << itStageX->second.iPtr->getAbsoluteEncoderPosition());}catch(...){E_LLOG("getAbsoluteEncoderPosition");}
//        try{U9_LLOG("getIncrementalEncoderPosition : " << itStageX->second.iPtr->getIncrementalEncoderPosition());}catch(...){E_LLOG("getIncrementalEncoderPosition");}
//        try{U9_LLOG("getVelocity : " << itStageX->second.iPtr->getVelocity());}catch(...){E_LLOG("getVelocity");}

        std::stringstream json_alias;
        json_alias << "\"" << itStageX->first << "\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        if(counter < MapStageX.size())jsonStatus += ", ";
    }

    std::map<std::string, WheelDevice>::const_iterator itWheel;
    const std::map<std::string, WheelDevice> & MapWheel = m_ptrMultiWheel->GetMultiWheel();
    if(!MapWheel.empty() && counter > 0)
    {
        jsonStatus += ", ";
        counter = 0;
    }
    for(itWheel = MapWheel.begin(); itWheel != MapWheel.end(); itWheel++)
    {
        counter++;
        std::string named_pos = itWheel->second.iPtr->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
        std::string service_name;
        try{service_name = itWheel->second.iPtr->readServiceName();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ readServiceName"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        bool isSimulated = false;
        Nice::SeqBitSet switch_status;
        // try{switch_status = itWheel->second.iPtr->getPositionSwitchStatus(); isSimulated = switch_status.size() < 32;}catch(...){E_LLOG(service_name << " ~ getPositionSwitchStatus"); LogBoostDiagnosticInformationMotion(__LINE__);}
//        try{U9_LLOG("getPositionSwitchStatus : " << switch_status);}catch(...){E_LLOG("getPositionSwitchStatus");}
        bool isAtHome = false;
        try{isAtHome = itWheel->second.iPtr->isAtHome();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isAtHome"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        bool inLimit = false;
//         try{U9_LLOG("isAtLimit : " << itWheel->second.iPtr->isAtLimit());}catch(...){E_LLOG("isAtLimit");}
//         if(itWheel->second.iPtr->isAtLimit())
//         {
//             std::string limit_result = itWheel->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             itWheel->second.iPtr->config()->setAtNode<std::string>("VAR.LIMIT", "-2");
//             limit_result = itWheel->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             limit_result = itWheel->second.iPtr->config()->leaf("VAR.LIMIT")->get<std::string>();
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
// //                val.limit = m_node->getAtNode<double> ("LIMIT")
//             double limit_double = itWheel->second.iPtr->config()->getAtNode<double>("LIMIT");
//             U9_LLOG("PropertyName : " << "LIMIT" << "; limit_double : " << limit_double);
//         }
        bool isMoving = false;
        try{isMoving = itWheel->second.iPtr->isMoving();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isMoving"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
//        try{U9_LLOG("getNamedPosition : " << itWheel->second.iPtr->getNamedPosition(1));}catch(...){E_LLOG("getNamedPosition");}
//        try{U9_LLOG("readPosition : " << itWheel->second.iPtr->readPosition());}catch(...){E_LLOG("readPosition");} // signals cast error
//        try{U9_LLOG("readDeviceEncoderPosition : " << itWheel->second.iPtr->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
//        try{U9_LLOG("readAbsoluteEncoderPosition : " << itWheel->second.iPtr->readAbsoluteEncoderPosition());}catch(...){E_LLOG("readAbsoluteEncoderPosition");}
//        try{U9_LLOG("readIncrementalEncoderPosition : " << itWheel->second.iPtr->readIncrementalEncoderPosition());}catch(...){E_LLOG("readIncrementalEncoderPosition");}
        bool isReachable = false;
        try{isReachable = itWheel->second.iPtr->isReachable();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isReachable"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getPosition : " << itWheel->second.iPtr->getPosition());}catch(...){E_LLOG("getPosition");}
        int deviceEncoderPosition;
        try{deviceEncoderPosition = itWheel->second.iPtr->getDeviceEncoderPosition();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ getDeviceEncoderPosition"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getAbsoluteEncoderPosition : " << itWheel->second.iPtr->getAbsoluteEncoderPosition());}catch(...){E_LLOG("getAbsoluteEncoderPosition");}
//        try{U9_LLOG("getIncrementalEncoderPosition : " << itWheel->second.iPtr->getIncrementalEncoderPosition());}catch(...){E_LLOG("getIncrementalEncoderPosition");}
//        try{U9_LLOG("getVelocity : " << itWheel->second.iPtr->getVelocity());}catch(...){E_LLOG("getVelocity");}

        std::stringstream json_alias;
        json_alias << "\"" << itWheel->first << "\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        if(counter < MapWheel.size())jsonStatus += ", ";
    }
   
    std::map<std::string, StageXYDevice>::const_iterator itStageXY;
    const std::map<std::string, StageXYDevice> & MapStageXY = m_ptrMultiStageXY->GetMultiStageXY();
    if(!MapStageXY.empty() && counter > 0)
    {
        jsonStatus += ", ";
        counter = 0;
    }
    for(itStageXY = MapStageXY.begin(); itStageXY != MapStageXY.end(); itStageXY++)
    {
        counter++;
        std::string named_pos = itStageXY->second.iPtr->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
        std::string service_name;
        try{service_name = itStageXY->second.iPtr->readServiceName();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ readServiceName"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        bool isSimulated = false;
        Nice::SeqBitSet switch_status;
        // try{switch_status = itStageXY->second.iPtr->getPositionSwitchStatus(); isSimulated = switch_status.size() < 32;}catch(...){E_LLOG(service_name << " ~ getPositionSwitchStatus");}
//        try{U9_LLOG("getPositionSwitchStatus : " << switch_status);}catch(...){E_LLOG("getPositionSwitchStatus");}
        bool isAtHome = false;
        try{isAtHome = itStageXY->second.iPtr->isAtHome();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isAtHome"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        bool inLimit = false;
        try{inLimit = itStageXY->second.iPtr->isAtLimit();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isAtLimit"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
        if(inLimit)
        {
//             std::string limit_result = itStageXY->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             itStageXY->second.iPtr->config()->setAtNode<std::string>("VAR.LIMIT", "-2");
//             limit_result = itStageXY->second.iPtr->config()->getAtNode<std::string>("VAR.LIMIT");
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
//             limit_result = itStageXY->second.iPtr->config()->leaf("VAR.LIMIT")->get<std::string>();
//             U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
// //                val.limit = m_node->getAtNode<double> ("LIMIT")
//             double limit_double = itStageXY->second.iPtr->config()->getAtNode<double>("LIMIT");
//             U9_LLOG("PropertyName : " << "LIMIT" << "; limit_double : " << limit_double);
        }
        bool isMoving = false;
        try{isMoving = itStageXY->second.iPtr->isMoving();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isMoving"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
//        try{U9_LLOG("getNamedPosition : " << itStageXY->second.iPtr->getNamedPosition(1));}catch(...){E_LLOG("getNamedPosition");}
//        try{U9_LLOG("readPosition : " << itStageXY->second.iPtr->readPosition());}catch(...){E_LLOG("readPosition");}
//        try{U9_LLOG("readDeviceEncoderPosition : " << itStageXY->second.iPtr->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
//        try{U9_LLOG("readAbsoluteEncoderPosition : " << itStageXY->second.iPtr->readAbsoluteEncoderPosition());}catch(...){E_LLOG("readAbsoluteEncoderPosition");}
//        try{U9_LLOG("readIncrementalEncoderPosition : " << itStageXY->second.iPtr->readIncrementalEncoderPosition());}catch(...){E_LLOG("readIncrementalEncoderPosition");}
        bool isReachable = false;
        try{isReachable = itStageXY->second.iPtr->isReachable();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ isReachable"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getPosition : " << itStageXY->second.iPtr->getPosition());}catch(...){E_LLOG("getPosition");}
        Nice::Point deviceEncoderPosition;
        try{deviceEncoderPosition = itStageXY->second.iPtr->getDeviceEncoderPosition();}catch(...){catch_msg.str(""); catch_msg << service_name << " ~ getDeviceEncoderPosition"; MotionLogBoostDiagnosticInformation(__LINE__, catch_msg.str());}
//        try{U9_LLOG("getAbsoluteEncoderPosition : " << itStageXY->second.iPtr->getAbsoluteEncoderPosition());}catch(...){E_LLOG("getAbsoluteEncoderPosition");}
//        try{U9_LLOG("getIncrementalEncoderPosition : " << itStageXY->second.iPtr->getIncrementalEncoderPosition());}catch(...){E_LLOG("getIncrementalEncoderPosition");}
//        try{U9_LLOG("getVelocity : " << itStageXY->second.iPtr->getVelocity());}catch(...){E_LLOG("getVelocity");}

        std::stringstream json_alias;
        json_alias << "\"" << itStageXY->first << "\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        if(counter < MapStageXY.size())jsonStatus += ", ";
    }

//    std::map<std::string, std::string> status_motion;
//    status_motion = getNamedPosition(vecAlias);


    //    struct MotorStatus{
    //               int steps;
    //               string namedPos;
    //               bool connected;
    //               bool moving;
    //               bool atHome;
    //               bool inSwNegativeLimit;
    //               bool inSwPositiveLimit;
    //               bool inHwNegativeLimit;
    //               bool inWhPositiveLimit;
    //    };

    return status;
}


////////////////////////////

// int test(int argc, char* argv[])
// {
//   //
//   std::string configDir = "/home/tan_unstable/SHARK-NIR/shins/branches/unstable/config/etc/";
//   // shins::util::mapPosType mapMainMapPos;
//   std::map<std::string, std::string> mapCfgFiles;
//   mapCfgFiles["APO_W"          ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.apodizer_w-dev.cfg";
//   mapCfgFiles["CORO_W"         ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.coro_slit_w-dev.cfg";
//   mapCfgFiles["DB_W"           ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.db_w-dev.cfg";
//   mapCfgFiles["LYOT_W"         ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.lyot_grism_w-dev.cfg";
//   mapCfgFiles["ND_W"           ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.nd_w-dev.cfg";
//   mapCfgFiles["SCI1_W"         ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.sci1_w-dev.cfg";
//   mapCfgFiles["SCI2_W"         ] = configDir+"motion_ctrl/wheels/motion_ctrl.wheels.sci2_w-dev.cfg";
//   mapCfgFiles["INBEAM_DEP"     ] = configDir+"motion_ctrl/beam/motion_ctrl.beam.inbeam_dep-dev.cfg";
//   mapCfgFiles["INBEAM_TT"      ] = configDir+"motion_ctrl/beam/motion_ctrl.beam.inbeam_tt-dev.cfg";
//   mapCfgFiles["LENS_DEP"       ] = configDir+"motion_ctrl/beam/motion_ctrl.beam.pupil_lens_dep-dev.cfg";
//   mapCfgFiles["SHUTTER"        ] = configDir+"motion_ctrl/beam/motion_ctrl.beam.shutter-dev.cfg";
//   mapCfgFiles["CAL_MIRROR_DEP" ] = configDir+"calunit_ctrl/calunit_ctrl.cal_dep-dev.cfg";
//   mapCfgFiles["FIBER_DEP"      ] = configDir+"calunit_ctrl/calunit_ctrl.fiber_dep-dev.cfg";
//   mapCfgFiles["derotator1"     ] = configDir+"track_ctrl/drot/track_ctrl.drot.derotator1-dev.cfg";
//   mapCfgFiles["derotator1"     ] = configDir+"track_ctrl/drot/track_ctrl.drot.derotator2-dev.cfg";
//   mapCfgFiles["adc"            ] = configDir+"track_ctrl/adc/track_ctrl.adc.prisms-dev.cfg";
//
//   std::string strAliasFile = "/home/tan_unstable/SHARK-NIR/shins/branches/unstable/config/alias.cfg";
//   osfsm::OSFsm sm;
//   boost::shared_ptr<osfsm::OSFsm> ptrSM = boost::make_shared<osfsm::OSFsm>(sm);
//   motionCtrl motion(argc, argv, strAliasFile, mapCfgFiles, ptrSM);
//   return 0;
// }
