#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import sys
import Ice
import sharknir

from pathlib import Path

import master_listener_davide
import numpy as np
from astropy.time import Time
from tpl.shinsapp import SHINSApp
from tpl.SHARKNIR_gen_base import SHARKNIR_gen_base
from util import functions as ufun


actuators = np.array([
                                                  [0.0, 0.3], [0.0, 0.4], [0.0, 0.5], [0.0, 0.6], [0.0, 0.7],
                                      [0.1, 0.2], [0.1, 0.3], [0.1, 0.4], [0.1, 0.5], [0.1, 0.6], [0.1, 0.7], [0.1, 0.8],
                          [0.2, 0.1], [0.2, 0.2], [0.2, 0.3], [0.2, 0.4], [0.2, 0.5], [0.2, 0.6], [0.2, 0.7], [0.2, 0.8], [0.2, 0.9],
              [0.3, 0.0], [0.3, 0.1], [0.3, 0.2], [0.3, 0.3], [0.3, 0.4], [0.3, 0.5], [0.3, 0.6], [0.3, 0.7], [0.3, 0.8], [0.3, 0.9], [0.3, 1.0 ],
              [0.4, 0.0], [0.4, 0.1], [0.4, 0.2], [0.4, 0.3], [0.4, 0.4], [0.4, 0.5], [0.4, 0.6], [0.4, 0.7], [0.4, 0.8], [0.4, 0.9], [0.4, 1.0 ],
              [0.5, 0.0], [0.5, 0.1], [0.5, 0.2], [0.5, 0.3], [0.5, 0.4], [0.5, 0.5], [0.5, 0.6], [0.5, 0.7], [0.5, 0.8], [0.5, 0.9], [0.5, 1.0 ],
              [0.6, 0.0], [0.6, 0.1], [0.6, 0.2], [0.6, 0.3], [0.6, 0.4], [0.6, 0.5], [0.6, 0.6], [0.6, 0.7], [0.6, 0.8], [0.6, 0.9], [0.6, 1.0 ],
              [0.7, 0.0], [0.7, 0.1], [0.7, 0.2], [0.7, 0.3], [0.7, 0.4], [0.7, 0.5], [0.7, 0.6], [0.7, 0.7], [0.7, 0.8], [0.7, 0.9], [0.7, 1.0 ],
                          [0.8, 0.1], [0.8, 0.2], [0.8, 0.3], [0.8, 0.4], [0.8, 0.5], [0.8, 0.6], [0.8, 0.7], [0.8, 0.8], [0.8, 0.9],
                                      [0.9, 0.2], [0.9, 0.3], [0.9, 0.4], [0.9, 0.5], [0.9, 0.6], [0.9, 0.7], [0.9, 0.8],
                                                  [1.0, 0.3], [1.0, 0.4], [1.0, 0.5], [1.0, 0.6], [1.0, 0.7]
             ])

dummy_image = np.zeros(shape=(64,64))

signs_list = []
with open(u'/opt/tan/etc/tiptilt_ctrl/rtc/BAX479-DM97-sign.dat', "r") as f:
    for line in f:
        for x in line.split():
            signs_list.append(int(x))
signs = np.array(signs_list)


class DevicesObject(object):
    def __init__(self):
        self.mini_os = SHINSApp()
        self.obsController = self.mini_os.os()

    def readall(self):
        status, json_string = self.obsController.getJsonStatus()
        unordered_json_dict = json.loads(json_string)
        json_dict = {}

        # if "derotator" in unordered_json_dict:
        #     unordered_json_dict["DROT_PosAng"] = unordered_json_dict["derotator"]

        # if "PRISM1" in unordered_json_dict:
        #     unordered_json_dict["ADC1_PosAng"] = unordered_json_dict["PRISM1"]
        #     unordered_json_dict["ADC1_PosAng"]["name"] = "ADC1_PosAng"

        # if "PRISM2" in unordered_json_dict:
        #     unordered_json_dict["ADC2_PosAng"] = unordered_json_dict["PRISM2"]
        #     unordered_json_dict["ADC2_PosAng"]["name"] = "ADC2_PosAng"

        # for key in sorted(unordered_json_dict):
        #     json_dict[key] = unordered_json_dict[key]
        
        json_dict = unordered_json_dict
        return json_dict

    def read(self, dev):
        # inDict = {dev:self.dummy}
        # unit = "STEPS"
        # status, outDict = self.obsController.getPosition(unit, inDict)
        outDict = self.readall()
        return outDict[dev]

    def move(self, ins_s):

        status, _, __ = SHARKNIR_gen_base().main(ins_s=ins_s)

        return ins_s

    def absolute(self, ins_s):
        status = self.obsController.setupInstrumentAbsolutePosition(ins_s)
        return status

    def relative(self, ins_s):
        status = self.obsController.setupInstrumentRelativePosition(ins_s)
        return status

    def home(self, ins_s):

        lamps_s = dict((k, "OFF") for k in  ["CAL_FF_LAMP", "CAL_FIBER_FOCUS_LAMP", "CAL_FIBER_DEFOCUS_LAMP"] if k in ins_s)
        if lamps_s:
            status, _, __ = SHARKNIR_gen_base().main(ins_s=lamps_s)

        status = self.obsController.initMotion(ins_s.keys())
        status = self.obsController.initCalibUnit([k for k in ["CAL_FILTER_DEP", "CAL_MIRROR_DEP", "CAL_FIBER_DEP"] if k in ins_s])

        return status

    def stop(self, ins_s):
        status = self.obsController.stopMotion(ins_s.keys())
        return status

    def abort(self):
        status = self.mini_os.abort()
        return status


class RtcObject(object):
    def __init__(self):
        self.counter = 0
        self.todos = []
        self.mini_os = SHINSApp()
        self.obsController = self.mini_os.os()

        self.thread = master_listener_davide.PLOT_THREAD()
        try:
            self.thread.start()
            self.thread.RUN = True
            print("plot thread started")
        except KeyboardInterrupt:
            print(" Intercepted Keyboard Interrupt")
            self.thread.CLOSE = True
            self.thread.RUN = False
            del self.obsController
            print("plot thread stopped")
            sys.exit(1)
        except Exception as e:
            print("here")
            print(e)
            self.thread.CLOSE = True
            self.thread.RUN = False
            del self.obsController
            print("plot thread stopped")
            sys.exit(1)

    def read(self, id):
        for todo in self.todos:
            if todo["id"] == id:
                return todo

    def create(self, data):
        todo = data
        #todo["id"] = self.counter = self.counter + 1
        self.todos.append(todo)
        return todo

    def update(self, id, data):
        todo = self.read(id)
        todo.update(data)
        return todo

    def delete(self, id):
        todo = self.read(id)
        self.todos.remove(todo)

    @property
    def centroid(self):
        status, [cx, cy] = self.obsController.getInternalSeqFloatVariable("RtcCentroidOrigin")
        return [cx, cy]

    @centroid.setter
    def centroid(self, cx_cy):
        cx, cy = cx_cy
        status = self.obsController.RTCTTSetCentroidPos(cx, cy)
        return status

    def flatten(self):
        modes = [0] * 39
        status = self.obsController.RTCTTModeUpload(modes, absolute=True)
        return status

    def close_loop(self):
        status = self.obsController.RTCTTCloseLoop()
        return status

    def open_loop(self):
        status = self.obsController.RTCTTOpenLoop()
        return status

    def frequency(self, freq):
        """Set the loop frequency"""

        rtc_s = {"TTFRAMERATE": freq}
        status = self.obsController.RTCTTExecuteSetup(rtc_s)

        return status

    ##############
    # Dark stuff #
    ##############

    def todos_dark(self):
        """List dark files"""
        dark_path = Path("/opt/tan/etc/tiptilt_ctrl/rtc/darks/")

        return [ str(dark.name).decode('utf-8') for dark in sorted(dark_path.glob("*.dat")) ]

    def new_dark(self):
        """Create a new dark"""

        rtc_s = {
            "TTDIAGENABLED" : "true", # or "1"
            "TTDIAGDECIMATION" : "0",
            "TTPIXELENABLED" : "1",
            "TTPIXELDECIMATION" : "0", # if 0 master diagnostic stops
            "TTLOOPENABLED" : "false",
            "TTSAVEASBIAS" : "true",
            "TTUPDATEBIAS" : "true",
            "TTTESTTIME" : "1000000",
        }
        status, filename = self.obsController.RTCTTTeccamExpose(rtc_s)

        return filename

    def load_dark(self, name):
        """Load a dark file"""
        rtc_s = {"TTBIASFILE": name}
        status = self.obsController.RTCTTExecuteSetup(rtc_s)

        return status

    ##############
    # Gain stuff #
    ##############

    def todos_gain(self):
        """List gain files"""
        gain_path = Path("/opt/tan/etc/tiptilt_ctrl/rtc/gains/")

        return [ str(gain.name).decode('utf-8') for gain in sorted(gain_path.glob("*.dat")) ]

    def new_gain(self, gain_type, centroid_size):
        """Create a new gain"""
        rtc_s = {
            "TTPIXELGAINMODE": gain_type,
            "TTPIXELGAINRADIUS": centroid_size,
        }
        status = self.obsController.RTCTTExecuteSetup(rtc_s)
        return status

    def load_gain(self, name):
        """Load a gain file"""
        rtc_s = {
            "TTPIXELGAINMODE": "FILEGAIN",
            "TTPIXELGAINFILE": name,
        }
        status = self.obsController.RTCTTExecuteSetup(rtc_s)
        return status

    ##############
    # Flat stuff #
    ##############

    def todos_flat(self):
        """List flat files"""
        flat_path = Path("/opt/tan/etc/tiptilt_ctrl/rtc/dm_flats/")
        return [ str(flat.name).decode('utf-8') for flat in sorted(flat_path.glob("*.txt")) ]

    def save_flat(self):
        """Save the current flat in a file"""


        flat_path = "/opt/tan/etc/tiptilt_ctrl/rtc/dm_flats/"

        # Time
        now = Time.now().datetime.strftime("%Y-%m-%d-%H-%M-%S")

        # T, seeing, el, dome
        try:
            params = ["LBTTemp", "DIMMSeeing", "ELPosition", "L_ShutterOpen", "L_SWA_Argos_Deployed"]
            dd_params = self.obsController.getDdParameters(params)[1][1:]
            temp, seeing, el, dome_open, swing_arm = [int(round(float(p))) for p in dd_params]
            el /= 3600

        except Exception as e:
            print("Cannot get some dd param")
            print(e)
            dd_params = range(5)
            temp, seeing, el, dome_open, swing_arm = ["err" for p in dd_params]

        # int ext
        try:
            all_devs = DevicesObject().readall()
            shutter_in = all_devs["SHUTTER"]["named"] == "IN"
            some_fiber_on = all_devs["CAL_FIBER_FOCUS_LAMP"]["named"] == "ON" or all_devs["CAL_FIBER_DEFOCUS_LAMP"]["named"] == "ON"

            if some_fiber_on:
                source = "int"
            elif (not some_fiber_on) and (not dome_open) and (not swing_arm):
                source = "bar"
            elif (not some_fiber_on) and swing_arm:
                source = "arg"
            elif (not some_fiber_on) and dome_open:
                source = "sky"
            else:
                source = "boh"
        except Exception as e:
            print("Cannot get JSON")
            print(e)
            source = "err"

        # b
        try:
            status, bearing = self.obsController.getPosition("DEGREE", {"DEROT":-9999})
            bearing = int(round(float(bearing.values()[0])))
        except Exception as e:
            print("Cannot get derot position")
            print(e)
            bearing = "err"

        # Example name: dm_shape_2022-11-24-19-22-33_arg_95b_90el_-4C_1as.txt
        flat_name = flat_path + "dm_shape_{}_{}_{}b_{}el_T{}C_{}as.txt".format(now, source, bearing, el, temp, seeing)
        with open(flat_name, "w") as flat_file:
            for i in range(len(self.thread.shape)):
                flat_file.write("{}\n".format(self.thread.shape[i]*signs[i]))

        return flat_name

    def load_flat(self, name):
        """Load a flat file"""
        rtc_s = {"TTDMFLATFILE": name}
        status = self.obsController.RTCTTExecuteSetup(rtc_s)
        return status

    def tiptilt(self, modes=[0,0,0,0], absolute=False):
        status = self.obsController.RTCTTModeUpload(modes=modes, absolute=absolute)
        return status

    def teccam_http_start(self):
        status, _ = self.obsController.RTCTTSendCameraCommand("exec httpserver start")
        return status

    def teccam_custom_command(self, command):
        status, res = self.obsController.RTCTTSendCameraCommand(command)
        return res

    def cloud_samples(self, s):
        status = self.thread.cloud_samples = s
        return status

    def save(self):
        data = self.data()
        now = Time.now().isot

        filename = "/data/rtc/rtc-panel-save-{}.json".format(now)
        with open(filename, 'w') as f:
            json.dump(data, f)
        return filename


    @property
    def rtc_image(self):

        # Diagostic stuff
        try:
            #print("trying to get the image")
            image = self.thread.img.T # transpose
            #print("there is an image. probably diagnostic is arriving")
            #print(image.shape)
        except AttributeError as e:
            print("no image")
            image = dummy_image
        except IndexError as e:
            print("too many indices")
            image = dummy_image

        return image


    def data_cred(self, image, centroid):

        try:
            cx,cy = centroid

            cred = {
                "data": image.tolist(),
                "min": image[3:, 3:].min(),
                "max": image[3:, 3:].max(),
                "cx": cx,
                "cy": cy,
                "image_shape": list(image.shape),
            }

        except IndexError as e:
            print("error building the cred response")
            print(e)
            cred = None

        return cred


    def data_gain(self, image):

        # OS
        status, pixgain = self.obsController.getInternalSeqFloatVariable("RtcPixelGain")

        try:

            gain = {
                "data": np.array(pixgain).reshape(image.shape).T.tolist(),
                "min": np.array(pixgain).min(),
                "max": np.array(pixgain).max(),
                "image_shape": list(image.shape),
            }

        except IndexError as e:
            print("error building the gain response")
            print(e)
            gain = None

        return gain

    def data_dm(self, image):

        # Diagnostics
        shape = self.thread.shape

        # OS
        status, dmflat = self.obsController.getInternalSeqFloatVariable("RtcDmFlat")

        try:

            dm = {
                "actuators": actuators.tolist(),
                "shape": (shape*signs).tolist(),
                "flat": (np.array(dmflat[0:97])).tolist(),
                "min": round(shape.min(),2),
                "max": round(shape.max(),2),
                "image_shape": list(image.shape),
            }

        except IndexError as e:
            print("error building the dm response")
            print(e)
            dm = None

        return dm


    def data_cloud(self, image, centroid):

        centroids_x = self.thread.centroid_x
        centroids_y = self.thread.centroid_y
        cloud = np.flip(self.thread.centroid_matrix.T, 1)
        cx,cy = centroid

        try:

            cld = {
                "data": cloud.tolist(),
                "min": cloud.min(),
                "max": cloud.max(),
                "samples": len(self.thread.centroid_x),
                "cx": cx - self.thread.cx,
                "cy": cy - self.thread.cy,
                "wx": round(np.max(centroids_x)-np.min(centroids_x), 5),
                "wy": round(np.max(centroids_y)-np.min(centroids_y), 5),
                "image_shape": list(image.shape),
            }

        except IndexError as e:
            print("error building the cloud response")
            print(e)
            cld = None

        return cld

    def data(self):

        image = self.rtc_image
        centroid = self.centroid

        response = {
            "cred": self.data_cred(image, centroid),
            "gain": self.data_gain(image),
            "dm": self.data_dm(image),
            "cloud": self.data_cloud(image, centroid),
        }

        return response

