
from util.shinsLogger import logger
from util import exception
from util import variables as uvars
from util import functions as ufun
from shinsapp import SHINSApp
import time

class SHARKNIR_gen_Preset(SHINSApp):

    @exception.handler_for_run
    def run(self, ins_s=None, det_s=None, tel_s=None, rtc_s=None, **kwargs):
        '''Try some tcs command
        '''
        
        obsController = self.obsController
        self.status = uvars.EXIT_SUCCESS            
        logger.info(__name__)
        
        ################################
        # TRANSFORM COORDINATES TO RAD #
        ################################

        logger.debug("Enumerating telescope setup")
        for element, value in tel_s.items():
            logger.debug("Element : % 15s  -  Value % 15s    " % (element, value))

        tel_s = ufun.to_radians(tel_s)

        logger.debug("Enumerating telescope setup after conversion")
        for element, value in tel_s.items():
            logger.debug("Element : % 15s  -  Value % 15s    " % (element, value))
            
        ##############
        ### SETUP  ###
        ##############        
        if self.paragraph("INSTRUMENT SETUP"):
        
            logger.info("Running: Moving drot")
            self.status = obsController.setupInstrument(ins_s)
            logger.info("Running: Moving drot... DONE")

        ###################################
        ### STOP ANY PREVIOUS TRACKING  ###
        ###################################
        if self.paragraph("STOP TRACKING"):
        
            logger.info("Running: trackingStop")
            status = obsController.trackingStop()
            logger.info("trackingStop returned : %d" % status)

        ##########################
        # OPEN REAL-TIME TT LOOP #
        ##########################
        if self.paragraph("OPEN REAL-TIME TT LOOP"):

            logger.info("TT open loop")
            self.status = obsController.RTCTTOpenLoop()
            ufun.error_handler(self.status, "RTCTTOpenLoop")
            logger.debug(self.status)
        
            logger.info("Tip tilt loop open on TECCAM")
            
        #########################
        ### PRESET TELESCOPE  ###
        #########################
        if self.paragraph("PRESET"):
            
            logger.info("Running: PresetTelescope")
            status = obsController.PresetTelescope(tel_s)
            logger.info("PresetTelescope returned : %d" % status)
        
        ##############
        ### TRACK  ###
        ##############        
        if self.paragraph("TRACK"):

            derot_duration = 60*60 # integer seconds
            obsController.setDerotationParameters(derot_duration, 1, 1, 2, 300, 100)

            logger.info("Running: trackingStartHere")
            async = obsController.begin_trackingStart()

            # status = obsController.end_trackingStart(async)
            # logger.info("trackingStartHere returned : %d" % status)
