
from shinsapp import SHINSApp
from util.shinsLogger import logger
from util import functions as ufun
from util import variables as uvars
from util import exception
from time import sleep

from SHARKNIR_gen_base import SHARKNIR_gen_base

class SHARKNIR_gen_init_and_park(SHINSApp):

    @exception.handler_for_run
    def run(self, ins_s=None, det_s=None, tel_s=None, rtc_s=None, **kwargs):
        '''Init or park devices'''

        logger.info(__name__)
        obsController = self.obsController
        self.status = uvars.EXIT_SUCCESS

        gen_base = SHARKNIR_gen_base()


        ##########################
        # SET TECCAM TEMPERATURE #
        ##########################
        if self.paragraph("SET TECCAM TEMPERATURE"):

            command = "temperatures sensor"
            temp_string = obsController.RTCTTSendCameraCommand(command)
            try: # T(C) sensor: -14.98
                temp = float(temp_string[1].split(":")[1])
                if temp > -5:
                    logger.info("Setting temperature to -15")
                    obsController.RTCTTSendCameraCommand("set temperatures sensor -15")
                else:
                    logger.info("Setting temperature to 20")
                    obsController.RTCTTSendCameraCommand("set temperatures sensor 20")
            except (IndexError,TypeError) as e:
                logger.error("Unable to get sensor temperature: won't set anything.")

        ###########################
        # INIT ALMOST ALL DEVICES #
        ###########################
        if self.paragraph("START INIT DEVICES"):

            logger.debug("start mot")
            async_mot = obsController.begin_initMotion([
                "INBEAM_DEP",
                "INBEAM_TT",
                "SHUTTER",
                "APODIZER_W",
                "CORO_SLIT_W",
                "LYOT_GRISM_W",
                "ND_FILT_W",
                "SCI_FILT_W1",
                "SCI_FILT_W2",
                "PUPIL_LENS_DEP",
                "DB_FILT_W",
                # "ADC1",
                # "ADC2",
            ])

            logger.debug("start cal")
            async_cal = obsController.begin_initCalibUnit(["CAL_FILTER_DEP",
                                                           "CAL_MIRROR_DEP",
                                                           "CAL_FIBER_DEP"])

        ###########################
        # WAIT ALMOST ALL DEVICES #
        ###########################
        if self.paragraph("WAIT INIT DEVICES"):

            logger.debug("end cal")
            status = obsController.end_initCalibUnit(async_cal)
            logger.debug(status)
            status = obsController.end_initMotion(async_mot)
            logger.debug(status)

        #############################################
        # START AND WAIT INSTRUMENT AND SASHA SETUP #
        #############################################

        gen_base.main(ins_s=ins_s, expose=False)
