# $Header: https://oberon.oapd.inaf.it/shins/trunk/shins/src/scripts/templates/tpl/SHARKNIR_di_mon_DetectorGain.py 3952 2022-05-19 08:30:53Z depa $

from shinsapp import SHINSApp
from util.shinsLogger import logger
from util import variables as uvars
from util import exception
from util import functions as ufun

from SHARKNIR_gen_base import SHARKNIR_gen_base

class SHARKNIR_gen_tiptilt_offset_and_scicam(SHINSApp):
    """calibration
    """

    @exception.handler_for_run
    def run(self, ins_s=None, det_s=None, tel_s=None, rtc_s=None, **kwargs):

        logger.info(__name__)
        obsController = self.obsController
        self.status = uvars.EXIT_SUCCESS

        gen_base = SHARKNIR_gen_base()
        
        ###########################
        # CHECK IF LOOP IS CLOSED #
        ###########################
        status, is_loop_closed = obsController.RTCTTRead(uvars.TTControlLoopEna, 1, "dsp")
        is_loop_closed = is_loop_closed[0]
        
        logger.info("Is loop closed? {}".format(is_loop_closed))
            
        ####################################
        # IF LOOP IS CLOSED: MOVE CENTROID #
        ####################################
        if self.paragraph("MOVE CENTROID"):
        
            ##################################
            # Retrieving RTC centroid origin #
            ##################################
        
            logger.info("Read centroid origin on RTC...")
            self.status, [cx, cy] = obsController.RTCTTReadSingle(uvars.CentroidOrigin, 2, 'dsp')
            logger.info("Read centroid origin on RTC... Done")
        
            # self.status, rtc_filenames, self.other = gen_tec_CloseLoop.main()
            # cx,cy = self.other
        
            cx_orig = cx
            cy_orig = cy
        
            logger.info("Retrieved center: {} {}".format(cx, cy))

        ##########################
        # OPEN REAL-TIME TT LOOP #
        ##########################
        if self.paragraph("OPEN REAL-TIME TT LOOP"):

            logger.info("TT open loop")
            self.status = obsController.RTCTTOpenLoop()
            ufun.error_handler(self.status, "RTCTTOpenLoop")
            logger.debug(self.status)
        
            logger.info("Tip tilt loop open on TECCAM")
        
            # ############################
            # # Pass centroid to the RTC #
            # ############################
        
            # cx = cx_orig
            # cy = cy_orig + float(rtc_s["STEP"])
            
            # logger.info("Change centroid origin on the technical camera")
            # self.status = obsController.RTCTTSetCentroidPos(cx, cy)
            # ufun.error_handler(self.status, "RTCTTSetCentroidPos")
        
            # logger.info("TECCAM CentroidOrigin is now {}, {}".format(cx, cy))
                
        ###########
        # MOVE DM #
        ###########
        if self.paragraph("MOVE DM"):
            
            ##################
            # SHOW RTC SETUP #
            ##################
                
            logger.debug("Enumerating detector setup")
            for element, value in rtc_s.items():
                logger.debug("Element : % 15s  -  Value % 15s    " % (element, value))
            
            ##############
            # SAVE MODES #
            ##############
            
            modes = [ float(rtc_s["VECTOR_{}".format(i)]) for i in range(1, 39) ]
            
            logger.warning("Modes: {}".format(modes))
            
            #############################
            # APPLY THE DM DISPLACEMENT #
            #############################
            
            if "ABSOLUTE" in rtc_s:
                logger.warning("ABSOLUTE keyword found")                
                if (rtc_s["ABSOLUTE"] == "true") : # or rtc_s["ABSOLUTE"]=="0"):
                    is_absolute = True       
                    logger.warning("Absolute movement")
                else:
                    is_absolute = False                
            else:
                is_absolute = False
            
            logger.info("Upload a mode vector to internal DM")
            self.status = obsController.RTCTTModeUpload(modes, absolute=is_absolute)
            logger.info("Waiting setup to apply displacement... DONE")
            
            
        #################
        # Setup + image #
        #################
        # gen base already have paragraphs
        
        gen_base.main(ins_s=ins_s, det_s=det_s)
        

        # if is_loop_closed:
        #     # Loop closed

        #     # ###############################################
        #     # # IF LOOP IS CLOSED: PUT BACK CENTROID ORIGIN #
        #     # ###############################################
        #     # if self.paragraph("PUT CENTROID BACK, PASS CENTROID TO RTC"): 
            
        #     #     logger.info("Putting back the DM to {}, {}".format(cx_orig, cy_orig))
            
        #     #     ############################
        #     #     # Pass centroid to the RTC #
        #     #     ############################
            
        #     #     logger.info("Change centroid origin on the technical camera")
        #     #     self.status = obsController.RTCTTSetCentroidPos(cx_orig, cy_orig)
        #     #     ufun.error_handler(self.status, "RTCTTSetCentroidPos")

                
        # else:
        #     # Loop open    
                        
        #################################
        # IF LOOP IS OPEN: MOVE DM BACK #
        #################################
        if self.paragraph("MOVE DM BACK"):

            modes = [-m for m in modes]
        
            logger.info("Upload a mode vector to internal DM")
            self.status = obsController.RTCTTModeUpload(modes, absolute=is_absolute)
            logger.info("Waiting setup to apply displacement... DONE")

            
        ###########################
        # CLOSE REAL-TIME TT LOOP #
        ###########################
        if self.paragraph("CLOSE REAL-TIME TT LOOP"):

            logger.info("TT close loop")
            self.status = obsController.RTCTTCloseLoop()
            ufun.error_handler(self.status, "RTCTTCloseLoop")
