# $Header: https://oberon.oapd.inaf.it/shins/branches/gigan-virgin_system/templates/util/variables.py 2132 2019-10-07 12:58:21Z depa $
import os

#########
# PATHS #
#########

pkgDataPath = os.environ["INSROOT"] + "/share/shins/templates"

refDir = pkgDataPath + "/ref"
defDir = pkgDataPath + "/ob"
schemaDir = pkgDataPath + "/schema"
combDir = pkgDataPath + "/comb"
fitsDir = "/data/fits/"

filenameSHINS = "FILENAME_SHINS"

#################
# EXIT STATUSES #
#################

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

#############
# Resources #
#############

# Name        IP                #port  #mapped #description
resources = {
    "cisco":      ("192.168.61.40",  80,   801,  "switch"),
    "mocon1":     ("192.168.61.50",  None, None, "Motor board"),
    "mocon2":     ("192.168.61.51",  None, None, "Motor board"),
    "mocon3":     ("192.168.61.52",  None, None, "Motor board"),
    "cred":       ("192.168.61.64",  8888, 8064, "Technical camera"),
    "dm":         ("192.168.61.65",  80,   8065, "Deformable Mirror"),
    "rtc":        ("192.168.61.66",  None, None, "Real Time computer"),
    "lakeshore":  ("192.168.61.67",  80,   8067, "Temperature controller"),
    "ipcam":      ("192.168.61.68",  80,   8068, "IP Camera of the lab"),
    "pdu1":       ("192.168.61.221", 80,   8221, "DM, RTC Mocon, lamps"),
    "pdu2":       ("192.168.61.222", 80,   8222, "MACIE; MediaA, MediaB"),
    "sashaws":    ("192.168.61.115", 8888, 8125, "scicam"),
    "monit":      ("192.168.61.113", 2812, 2812, "Monit (already on ws)"),
}

###########
# RTC DSP #
###########

TTControlLoopEna = 4399
CentroidOrigin = 4432

###########
# Devices #
###########

comment_char = "#"

base_path = "/opt/tan/etc/"

motion_ctrl = "motion_ctrl"
calunit_ctrl = "calunit_ctrl"
track_ctrl = "track_ctrl"
    
devices = {

    "APODIZER_W" : motion_ctrl,
    "ND_FILT_W" : motion_ctrl,
    "CORO_SLIT_W" : motion_ctrl,
    "LYOT_GRISM_W" : motion_ctrl,
    "SCI_FILT_W1" : motion_ctrl,
    "SCI_FILT_W2" : motion_ctrl,
    "DB_FILT_W" : motion_ctrl,
    "PUPIL_LENS_DEP" : motion_ctrl,
    "INBEAM_TT" : motion_ctrl,
    "INBEAM_TT_X" : motion_ctrl,
    "INBEAM_TT_Y" : motion_ctrl,
    "INBEAM_DEP" : motion_ctrl,
    "SHUTTER" : motion_ctrl,

    "CAL_MIRROR_DEP" : calunit_ctrl,
    "CAL_FIBER_DEP" : calunit_ctrl,
    "CAL_FILTER_DEP" : calunit_ctrl,

    "PRISM1" : track_ctrl,
    "PRISM2" : track_ctrl,
    "ADC_Mode" : track_ctrl,
    "ADC1_PosAng" : track_ctrl,
    "ADC2_PosAng" : track_ctrl,
    
    "DEROT" : track_ctrl,
    "DROT_PosAng" : track_ctrl,

}

########
# PDUs #
########

# PDU and lamps
ipPDU1 = resources["pdu1"][0]
ipPDU2 = resources["pdu2"][0]
pduUID = "shins"
snmpVer = 3

outletCmdDict = {"ON": 1,
                 "OFF": 2,
                 "REBOOT": 3,
                 "DELAYED_ON": 4,
                 "DELAYED_OFF": 5,
                 "DELAYED_REBOOT": 6}

# dictionary associating to device alias name its outlet on PDU #2
 
aliasDict = {"MOCON": 1,
             "CAL_FIBER_FOCUS_LAMP": 2,
             "CAL_FF_LAMP": 3,
             "CAL_FIBER_DEFOCUS_LAMP": 4,
             "MACIE": 5,
             "DM": 6,
             "RTC": 7,
             "CRED2": 8}

aliasDict2 = {
    "MEDIA1": 1,
    "MEDIA2": 2,
    "ARDUINO": 3,
    "ELECTROVALVE": 4,
    "LAKESHORE": 5,    
    "LAMP_TEL_SIM": 8,
}

lampStateDict = {1: "ON",
                 2: "OFF"}

COMMON_OID_PATH = ".1.3.6.1.4.1.3808.1.1.3.3"
COMMON_CONTROL_PATH = ".3.1.1"
COMMON_STATUS_PATH = ".5.1.1"

dictOutletControl = {
    "Index"          : COMMON_OID_PATH + COMMON_CONTROL_PATH + ".1.{}",
    "OutletName"     : COMMON_OID_PATH + COMMON_CONTROL_PATH + ".2.{}",
    "OutletPhase"    : COMMON_OID_PATH + COMMON_CONTROL_PATH + ".3.{}",
    "OutletCommand"  : COMMON_OID_PATH + COMMON_CONTROL_PATH + ".4.{}",
    "OutletBank"     : COMMON_OID_PATH + COMMON_CONTROL_PATH + ".5.{}",
    "OutletState"    : COMMON_OID_PATH + COMMON_STATUS_PATH + ".4.{}",  # (R) INTEGER, # (R) INTEGER,
}

dictOutletStatus = {
    "Index"          : COMMON_OID_PATH + COMMON_STATUS_PATH + ".1.{}",  # (R) INTEGER, # (R) DisplayString,
    "OutletName"     : COMMON_OID_PATH + COMMON_STATUS_PATH + ".2.{}",
    "OutletPhase"    : COMMON_OID_PATH + COMMON_STATUS_PATH + ".3.{}",  # (R) INTEGER,
    "OutletState"    : COMMON_OID_PATH + COMMON_STATUS_PATH + ".4.{}",  # (R) INTEGER, # (R) INTEGER,
    "CommandPending" : COMMON_OID_PATH + COMMON_STATUS_PATH + ".5.{}",
    "OutletBank"     : COMMON_OID_PATH + COMMON_STATUS_PATH + ".6.{}",  # (R) INTEGER,
    "Load"           : COMMON_OID_PATH + COMMON_STATUS_PATH + ".7.{}",  # (R) Gauge,
    "ActivePower"    : COMMON_OID_PATH + COMMON_STATUS_PATH + ".8.{}",  # (R) Gauge,
    "Alarm"          : COMMON_OID_PATH + COMMON_STATUS_PATH + ".9.{}",  # (R) INTEGER,
    "PeakPower"      : COMMON_OID_PATH + COMMON_STATUS_PATH + ".10.{}", # (R) Gauge, # (R) DisplayString,
    "PeakPowerTime"  : COMMON_OID_PATH + COMMON_STATUS_PATH + ".11.{}", # (R) DisplayString,
    "PeakPowerStart" : COMMON_OID_PATH + COMMON_STATUS_PATH + ".12.{}",
    "Energy"         : COMMON_OID_PATH + COMMON_STATUS_PATH + ".13.{}", # (R) Gauge, # (R) DisplayString
    "EnergyStartTime": COMMON_OID_PATH + COMMON_STATUS_PATH + ".14.{}"
}

