// $Header: https://oberon.oapd.inaf.it/shins/trunk/shins/src/track_ctrl/trackingCtrl.cpp 3969 2022-06-03 15:00:46Z depa $

#include <pthread.h>

#include "trackingCtrl.h"



void TrackLogBoostDiagnosticInformation(int line)
{
    std::string what = boost::current_exception_diagnostic_information();
    std::string toErase = "std::exception::what:";
    size_t pos = what.find(toErase);
    if(pos != std::string::npos)what.erase(pos, toErase.length());
    what.erase(std::remove(what.begin(), what.end(), '\n'), what.end());
    std::stringstream log_msg;
    log_msg << "Line: " << line << " ~ " << what;
    E_LLOG(log_msg.str());
}


void DrotDevice::GetDataFromConfigFile()
{
    units = su::get_main_units_one_axe(cfgFile);
    namedPositions = su::get_named_positions_one_axe(cfgFile);
    alias = su::get_alias_from_config(cfgFile);
}


void * trackingCtrl::at_drot_exec(void * arg)
{
    U6_LLOG(__FUNCTION__);
    trackingCtrl & me = *((trackingCtrl *)arg);

    const DrotDevice & derotator = me.m_MapDrots.begin()->second;//find(shins::util::constants::derotator)->second;
    U7_LLOG(__FUNCTION__ << " ~ " << derotator.alias << " ~ " << derotator.GetMoving() << " ~ " << me.m_MapDrots.size());
    std::vector<Nice::Date> tempi;
    std::vector<double> posizioni;
    int counter = 0;
    while(derotator.GetMoving())
    {
        if(counter % 500 == 0)
        {
            U7_LLOG(__FUNCTION__ << " ~ " << derotator.GetMoving())
        }
        counter++;
        try
        {
            Nice::Date t0 = Nice::Date::now();
            Nice::Date t1 = Nice::Date::now();
            double position = me.getAbsEncPosition();//m_drot->getPosition("ABSENC");
            Nice::Date t2 = Nice::Date::now();
            tempi.push_back(t0);
            tempi.push_back(t1);
            tempi.push_back(t2);
            posizioni.push_back(position);
        }
        catch(Basda::ServiceIsBusyException &e)
        {
            E_LLOG(e.what());
            W_LLOG("Could not acquire position from derotator");
        }
        catch(Basda::ServiceIsNotOnlineException &e)
        {
            E_LLOG(e.what());
        }
        catch(std::runtime_error & e)
        {
            E_LLOG(e.what());
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
        }
        usleep(10000);
    }

    std::vector<Nice::Date>::iterator itTempi;
    std::vector<double>::iterator itPosition;
    std::ofstream out("/opt/tan/var/log/drot_positions.txt");
    for(itTempi = tempi.begin(), itPosition = posizioni.begin(); itTempi < tempi.end(); itTempi++)
    {
        out << *itTempi << "\t";
        itTempi++;
        out << *itTempi << "\t";
        itTempi++;
        out << *itTempi << "\t" << (int)*itPosition << "\n";
        itPosition++;
    }
    out.close();

    U8_LLOG(__FUNCTION__ << " Completed ~ Saved : " << posizioni.size() << " positions");
    return NULL;
}

bool trackingCtrl::AdcCallback(double & pressure, double & temperature, double & elevation, double & drot_pos)
{
    trackingCtrl & me = *trackingCtrl::m_pTrackingCtrl;

    double smt_pressure, smt_temperature;
    const double OneAtm = 101500.;

    lbto::DDstruct dd;
    lbto::SeqDD ddentries;
    dd.DDname = "ELPosition";
    ddentries.push_back(dd);
    dd.DDname = "LBTPressure";
    ddentries.push_back(dd);
    dd.DDname = "LBTTemp";
    ddentries.push_back(dd);
    dd.DDname = "SMTPressure";
    ddentries.push_back(dd);
    dd.DDname = "SMTTemp";
    ddentries.push_back(dd);

    lbto::result res;
    bool get_parameters = true;
    int retry_counter = 0;
    while(get_parameters)
    {
        get_parameters = false;
        try
        {
            res = me.m_iif->iifGetParameter(ddentries);
        }
        catch(Basda::ServiceIsBusyException)
        {
            if(retry_counter && retry_counter % 10 == 0)
            {
                W_LLOG("Busy : retrying " << retry_counter);
            }
            retry_counter++;
            get_parameters = true;
            usleep(50000);
            continue;
        }
        catch(const std::exception& e)
        {
            W_LLOG(e.what());
            return false;
        }
        catch(...)
        {
            W_LLOG(__FUNCTION__);
            return false;
        }
    }
    if(res.rescode == EXIT_SUCCESS)
    {
        su::getValue<double>(std::string(res.resmsg[1].c_str()), elevation);
        su::getValue<double>(std::string(res.resmsg[2].c_str()), pressure);
        su::getValue<double>(std::string(res.resmsg[3].c_str()), temperature);
        su::getValue<double>(std::string(res.resmsg[4].c_str()), smt_pressure);
        su::getValue<double>(std::string(res.resmsg[5].c_str()), smt_temperature);
        // U7_LLOG(pressure << " ~ " << temperature << " ~ " << elevation << " ~ " <<  smt_pressure << " ~ " << smt_temperature );
    }
    else
    {
        U7_LLOG("Error getting parameters from TCS");
        for(unsigned int i = 1; i < res.resmsg.size(); i++)
        {
            U7_LLOG("resmsg[" << i << "] = " << res.resmsg[i]);
        }
        return false;
    }
    double elevation2 = (elevation/3600)*3.14159265358979/180.;
    double ZA = 90*(3.14159265358979/180.) - elevation2;
    double P = 1000*100*(pressure/OneAtm);
    double P2 = 1000*100*smt_pressure/OneAtm; // pressure in hPa

    double T = temperature; // m_iif->getDouble("LBTTemp", "env.weather.lbt.temperature");// temperature in C
    double T2 = smt_temperature;// temperature in C
    // U7_LLOG(P << " ~ " << T << " ~ " << elevation2 << " ~ " <<  P2 << " ~ " << T2 );
    drot_pos = me.getPosition(suc::DEROT, "DEGREE");
    return true;
}

void trackingCtrl::drot_callback(const Nice::Completion & _c)
{
    pthread_t tid1;
    switch (_c.state())
    {
    case Nice::OK:
        break;
    case Nice::EXEC:
        {
//            pthread_create(&tid1, NULL, at_drot_exec, trackingCtrl::m_pTrackingCtrl);
        }
        break;
    default:
        break;
    }
    U6_LLOG(__FUNCTION__ << " ~ DEROTATOR ~ " << _c.state());
    U8_LLOG("");
}

trackingCtrl * trackingCtrl::m_pTrackingCtrl = NULL;

trackingCtrl::trackingCtrl(int argc, char* argv[], const std::string& _strAliasFileName,
                           const shins::util::mapPosType & mainMapNamedPos,
                           const std::map<std::string, std::string> &_mapCfgFiles,
                           const Ltcs::IifServiceWorkerInterfacePtr& _iif,
                           boost::shared_ptr<osfsm::OSFsm>& _ptrStateMachine,
                           Ice::PropertiesPtr props)
:
    m_mainMapNamedPos(mainMapNamedPos),
    __OBJ_NAME__(typeid(this).name()),
    m_drot(NULL),
    m_adc1(NULL),
    m_adc2(NULL),
    m_iif(_iif),
    m_mapCfgFiles(_mapCfgFiles),
    m_props(props)
{
    U6_LLOG(__FUNCTION__ << " - Constructor");
    m_pTrackingCtrl = NULL;

    m_ptrAdcWorkerThread = NULL;

    // m_ptrLogger = _ptrLogger;

    // initializing Nice Application wrapper to control wheels
//    m_appWrapADC = new applicationWrapper(argc, argv);
    m_ptrStateMachine = _ptrStateMachine;

    std::vector<std::string> vecAlias;
    vecAlias = su::getAliasesFromFile(_strAliasFileName, "track_ctrl");
    if(vecAlias.size() == 0)
    {
        std::stringstream log_msg;
        log_msg << "Issue with aliases: no alias found in " << _strAliasFileName;
        E_LLOG(log_msg.str());
        throw std::runtime_error(log_msg.str());
    }
//    m_mainMapNamedPos = mainMapNamedPos;//su::get_main_named_pos_map_boost(_mapCfgFiles);

    std::map<std::string, std::map<std::string, std::tr1::tuple<int, int> > >::const_iterator itMaps;
    for(itMaps = m_mainMapNamedPos.begin(); itMaps != m_mainMapNamedPos.end(); itMaps++)
    {
        std::map<std::string, std::tr1::tuple<int, int> >::const_iterator itNamed;
        U9_LLOG(itMaps->first << "'s Named Positions");
        for(itNamed = itMaps->second.begin(); itNamed != itMaps->second.end(); itNamed++)
        {
            U9_LLOG("Keyword : " << std::setfill('_') << std::setw(16) << itNamed->first << "; Value : " << std::setw(40) << std::tr1::get<0>(itNamed->second) << " ~ " << std::tr1::get<1>(itNamed->second));
        }
//        su::log_dictionary(__FILE__, __LINE__, itMaps->second);
    }

    // Creating Nice endpoints from aliases
    std::vector<std::string>::iterator it;
    for(it = vecAlias.begin(); it < vecAlias.end(); ++it)
    {
        std::string upper_it = su::str_toupper(*it);
        std::string strEP = "[ALIAS=" + *it +"]"; // endpoint name
        Nice::Endpoint ep = Nice::Endpoint::fromString(strEP); // endpoint
        Nice::JointPtr nodeADC  = Nice::Application::config()->node("ADC");
        Nice::JointPtr nodeDrot = Nice::Application::config()->node("DROT");
        U9_LLOG("Track Vec Alias : " << *it);
        // Creating endpoints in local property tree using aliases
        if(boost::algorithm::starts_with(*it, "derot"))
        {
            nodeDrot->setAtNode<Nice::Endpoint>(upper_it + ".CON", ep);
            m_vecAliasDrot.push_back(upper_it);
            U1_LLOG("Set node with endpoint " << m_vecAliasDrot.back());
            continue;
        }
        if(boost::algorithm::starts_with(*it, "adc"))
        {
            nodeADC->setAtNode<Nice::Endpoint>(upper_it + ".CON", ep);
            m_vecAliasADC.push_back(upper_it);
            U1_LLOG("Set node with endpoint " << m_vecAliasADC.back());
            continue;
        }
    }

    if(m_iif)
    {
        m_iif->changeState(Basda::ONLINE);
        U9_LLOG("IIF->state = " << _iif->state());
    }

    m_pTrackingCtrl = this;
    m_connectionEstablished = false;

    pthread_t tid1;
    pthread_create(&tid1, NULL, connect_to_service, trackingCtrl::m_pTrackingCtrl);

    std::string jsonStatus;
    getJsonStatus(jsonStatus);

    m_pTrackingCtrl = this;

    U8_LLOG(__FUNCTION__ << " ~ DONE contructor");
}

trackingCtrl::~trackingCtrl()
{
    I_LLOG(__FUNCTION__);
    if(m_ptrAdcWorkerThread)
        delete m_ptrAdcWorkerThread;
}

void * trackingCtrl::connect_to_service(void * arg)
{
    trackingCtrl & Me = *((trackingCtrl *)arg);
    while(!Me.m_connectionEstablished)
    {
        Me.connectToService();
        usleep(120000000);
    }
    return NULL;
}

void trackingCtrl::connectToService()
{
    std::string strDevAlias = "";
    std::string strSvcNode = "";
    // Create device node at client side
    Nice::JointPtr tmpNodeADC1;
    Nice::JointPtr tmpNodeADC2;
    Nice::JointPtr tmpNodeDrot;
    Nice::AppConfigClientPtr con;

    Nice::Time timeout = Nice::Time::seconds(2);
    int change_state_result;
    if(m_vecAliasDrot.size() != 0)
    {
        U9_LLOG("connecting to DROT remote node : " << suc::DEROT);
        // tmpNodeDrot = Nice::Application::config()->node("DROT");
        strSvcNode = "DROT.DEROTATOR.DROT.SVC";
        tmpNodeDrot = Nice::Application::config()->node("DROT");
        // con = Nice::AppConfigClient::create(tmpNodeDrot->getAtNode<Nice::Endpoint>("DROT1.CON"),
        // tmpNodeDrot->node("DROT1"));
        if(!m_drot)
        {
            try
            {
                con = Nice::AppConfigClient::create(tmpNodeDrot->getAtNode<Nice::Endpoint>(suc::DEROT + ".CON"), tmpNodeDrot->node(suc::DEROT));
            }
            catch (const Nice::Exception & e)
            {
                A_LLOG(e.what());
            }
            try
            {
                m_drot = Sharknir::DrotServiceWorkerInterface::create(tmpNodeDrot->node(suc::DEROT));
            }
            catch (const IceUtil::NullHandleException & e)
            {
                A_LLOG(e.what());
            }
            if(m_drot)
            {
                setDerotationParameters(1200 /*numPolynomials*/, 1 /*samplingRate*/, 1 /*frequency*/, 2 /*polynomialOrder*/, 350 /*faultTolerance*/, 100 /*samplesPerSegment*/);

                std::map<std::string, std::string>::const_iterator itFind;
                std::string cfgFile;
                bool bDerotatorConfigurationFound = false;
                for(itFind = m_mapCfgFiles.begin(); itFind != m_mapCfgFiles.end(); itFind++)
                {
                    if(su::left(itFind->first, 5) == suc::DEROT)
                    {
                        cfgFile = itFind->second;
                        m_MapDrots.insert(std::pair<std::string, DrotDevice>(suc::DEROT, DrotDevice(cfgFile)));
                        m_MapDrots.find(suc::DEROT)->second.SetInterfacePointer(m_drot);
                        install_callbacks();
                        m_MapDrots.find(suc::DEROT)->second.SetSecondaryCallbacks(NULL, // void (* moveToHomeCB)(const Nice::Completion & c),
                                                                                  NULL, // void (* moveToNamedCB)(const Nice::Completion & c),
                                                                                  NULL, // void (* moveAbsoluteCB)(const Nice::Completion & c),
                                                                                  NULL, // void (* moveRelativeCB)(const Nice::Completion & c),
                                                                                  NULL, // void (* moveToLimitCB)(const Nice::Completion & c),
                                                                                  drot_callback, // void (* startDerotationCB)(const Nice::Completion &c),
                                                                                  NULL, // void (* startDerotationHereCB)(const Nice::Completion & c),
                                                                                  NULL); // void (* varAbsoluteEncoderCB)(const double & val));
                        bDerotatorConfigurationFound = true;
                    }
                }
                if(!bDerotatorConfigurationFound)
                {
                    E_LLOG("Error could not find DROT config file");
                }
                change_state_result = m_drot->changeState(Basda::STANDBY, timeout);
                if(change_state_result == 0){W_LLOG("Derotator change state to STANDBY timed out");}
                change_state_result = m_drot->changeState(Basda::ONLINE, timeout);
                if(change_state_result == 0){W_LLOG("Derotator change state to ONLINE timed out");}
                U9_LLOG("DROT->state = " << m_drot->state());
            }
            else
            {
                E_LLOG("Error creating DROT Worker Interface. Module is not active");
            }
        }
    }

    /* connecting to ADC remote node */
    if(m_vecAliasADC.size() != 0)
    {
//        strSvcNode = "ADC.PRISM.PRISM1.SVC";
        if(!m_adc1)
        {
            U9_LLOG("connecting to ADC remote node " << strSvcNode);
            tmpNodeADC1 = Nice::Application::config()->node("ADC");
            // don't know what next line does but is necesary
            try
            {
                con = Nice::AppConfigClient::create(tmpNodeADC1->getAtNode<Nice::Endpoint>(suc::ADC1 + ".CON"), tmpNodeADC1->node(suc::ADC1));
            }
            catch (const Nice::Exception & e)
            {
                A_LLOG(e.what());
            }
            try
            {
                m_adc1 = Basda::MoccaWheelServiceWorkerInterface::create(tmpNodeADC1->node(suc::ADC1));
            }
            catch (const IceUtil::NullHandleException & e)
            {
                A_LLOG(e.what());
            }
            if(m_adc1)
            {
                U9_LLOG("BREAK");
                change_state_result = m_adc1->changeState(Basda::STANDBY, timeout);
                if(change_state_result == 0){W_LLOG("ADC1 change state to STANDBY timed out");}
                change_state_result = m_adc1->changeState(Basda::ONLINE, timeout);
                if(change_state_result == 0){W_LLOG("ADC1 change state to ONLINE timed out");}
            }
            else
            {
                E_LLOG("Error creating ADC1 Worker Interface. Module is not active");
            }
        }

//        strSvcNode = "ADC.PRISM.PRISM2.SVC";
        if(!m_adc2)
        {
            U9_LLOG("connecting to ADC remote node " << strSvcNode);
            tmpNodeADC2 = Nice::Application::config()->node("ADC");
            try
            {
            con = Nice::AppConfigClient::create(tmpNodeADC2->getAtNode<Nice::Endpoint>(suc::ADC2 + ".CON"), tmpNodeADC2->node(suc::ADC2));
            }
            catch (const Nice::Exception & e)
            {
                A_LLOG(e.what());
            }
            try
            {
                m_adc2 = Basda::MoccaWheelServiceWorkerInterface::create(tmpNodeADC2->node(suc::ADC2));
            }
            catch (const IceUtil::NullHandleException & e)
            {
                A_LLOG(e.what());
            }

            if(m_adc2)
            {
                change_state_result = m_adc2->changeState(Basda::STANDBY, timeout);
                if(change_state_result == 0){W_LLOG("ADC2 change state to STANDBY timed out");}
                change_state_result = m_adc2->changeState(Basda::ONLINE, timeout);
                if(change_state_result == 0){W_LLOG("ADC2 change state to ONLINE timed out");}
            }
            else
            {
                E_LLOG("Error creating ADC2 Worker Interface. Module is not active");
            }
        }
    }

    if(!m_connectionEstablished && (m_adc1 && m_adc2))
    {
        m_ptrAdcWorkerThread = new AdcWorkerThread(m_adc1, m_adc2, AdcCallback, m_props);
        U9_LLOG("ADC1->state = " << m_adc1->state() << "; ADC2->state = " << m_adc2->state());
    }
    else
    {
        if(!m_adc1)
        {
            E_LLOG("Error could not create ADC1 Worker Interface. Module is not active.");
        }
        if(!m_adc2)
        {
            E_LLOG("Error could not create ADC2 Worker Interface. Module is not active.");
        }
    }

    m_connectionEstablished = (m_drot && m_adc1 && m_adc2);
}


int trackingCtrl::setDerotationParameters(int numPolynomials, double samplingRate, double frequency, int polynomialOrder, int faultTolerance, int samplesPerSegment)
{
    U6_LLOG(__FUNCTION__ << " ~ Num Polynomials " << numPolynomials << " ~ Sampling Rate " << samplingRate << " ~ Frequency " << frequency << " ~ Polynomial Order " << polynomialOrder << " ~ fault Tolerance " << faultTolerance << " ~ Samples per segment " << samplesPerSegment);
    if(m_drot)
    {
        m_drot->setDerotationParameters(numPolynomials, samplingRate, frequency, polynomialOrder, faultTolerance, samplesPerSegment);
    }
    else
    {
        std::string log_msg("Could not set Derotation Parameters to DROT : module is not active");
        E_LLOG(log_msg);
        U8_LLOG("");
        throw(std::runtime_error(log_msg));
    }
    U8_LLOG("");
    return EXIT_SUCCESS;
}


int trackingCtrl::setAdcCorrectionParameters(double co, double p0, double p1, double t0, double t1)
{
    if(m_ptrAdcWorkerThread)
    {
        m_ptrAdcWorkerThread->SetCorrectionParameters(co, p0, p1, t0, t1);
        return EXIT_SUCCESS;
    }
    else
    {
        W_LLOG("Could not set Adc Correction Parameters : module is not active");
        return EXIT_FAILURE;
    }
}


bool trackingCtrl::isDevice(const std::string& _strDevAlias)
{
    std::vector<std::string>::const_iterator itAlias;
    itAlias = std::find(m_vecAliasADC.begin(), m_vecAliasADC.end(), _strDevAlias);
    if(itAlias != m_vecAliasADC.end())return true;
    itAlias = std::find(m_vecAliasDrot.begin(), m_vecAliasDrot.end(), _strDevAlias);
    if(itAlias != m_vecAliasDrot.end())return true;
    return false;
}

void trackingCtrl::changeState(const Basda::ServiceState& _state, const std::vector<std::string>& _vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    switch (_state)
    {
    case Basda::ONLINE:
    case Basda::STANDBY:
    {
        if(std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::DEROT) < _vecDevAlias.end())
        {
            if(m_drot)
            {
                m_drot->changeState(_state);
            }
            else
            {
                W_LLOG("Could not change state to DROT : module is not active");
            }
        }

        if(std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC1) < _vecDevAlias.end())
        {
            U9_LLOG("calling changeState on ADC.PRISM.1")
            if(m_adc1)
            {
                m_adc1->changeState(_state);
            }
            else
            {
                std::string log_msg("Could not change state to ADC1 : module is not active");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        if(std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC2) < _vecDevAlias.end())
        {
            U9_LLOG("calling changeState on ADC.PRISM.2")
            if(m_adc2)
            {
                m_adc2->changeState(_state);
            }
            else
            {
                std::string log_msg("Could not change state to ADC2 : module is not active");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        break;
    }
    case Basda::SHUTDOWN:
    case Basda::OFFLINE:
    case Basda::WORKING:
    case Basda::ERROR:
    default:
    {
        W_LLOG("State " << _state << " not handled");
    }
    }
    U8_LLOG("");
}


void trackingCtrl::initialize(const std::vector<std::string>& _vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    try
    {
        initializeStart(_vecDevAlias);
        initializeWait(_vecDevAlias, Nice::Time::milliSeconds(suc::DROT_TIMEOUT*1000));  // time estimated to move by 190 deg
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}


void trackingCtrl::initializeStart(const std::vector<std::string>& _vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    double position;
    try
    {
        // DROT has absolute encoder and is not initialized
        if(m_adc1)
        {
            m_adc1->moveToHomeStart();
        }
        else
        {
            std::string log_msg("Could not initialize ADC1 : module is not active");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
        if(m_adc2)
        {
             m_adc2->moveToHomeStart();
        }
        else
        {
            std::string log_msg("Could not initialize ADC2 : module is not active");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }
    catch(std::exception e)
    {
        E_LLOG(__FUNCTION__ << ":" << e.what());
        U8_LLOG("");
        throw e;
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG("");
}

void trackingCtrl::initializeWait(const std::vector<std::string>& _vecDevAlias, const Nice::Time& _time)
{
    U6_LLOG(__FUNCTION__);
    su::Timer timer(_time);
    try
    {
        if(m_adc1)
        {
            m_adc1->moveToHomeWait(timer.residual());
        }
        if(m_adc2)
        {
            m_adc2->moveToHomeWait(timer.residual());
        }
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }

}


Ice::Int trackingCtrl::setupInstrumentStart(const sharknir::dictSetupNamedPos &_dSetupPos)
{
    U6_LLOG(__FUNCTION__);
    Ice::Int status = EXIT_SUCCESS;
    setMode(_dSetupPos);
    if(false)
    {
        setupPrepare(_dSetupPos, m_track_setup);
        try
        {
            if(!m_track_setup.AbsMapDrot.empty())
                m_drot->moveAbsoluteStart(m_track_setup.AbsMapDrot.begin()->second, m_track_setup.AbsMapDrotUnits.begin()->second);
            else if(!m_track_setup.RelMapDrot.empty())
                m_drot->moveRelativeStart(m_track_setup.RelMapDrot.begin()->second, m_track_setup.RelMapDrotUnits.begin()->second);
            else if(!m_track_setup.NamedMapDrot.empty())
                m_drot->moveToNamedPositionStart(m_track_setup.NamedMapDrot.begin()->second);
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }

        try
        {
            std::map< std::string, std::string>::const_iterator itUnit;
            std::map< std::string, double>::const_iterator itWheel;
            std::map< std::string, int>::const_iterator itNamed;
            if(!m_track_setup.AbsMapWheel.empty())
            {
                for(itWheel = m_track_setup.AbsMapWheel.begin(); itWheel != m_track_setup.AbsMapWheel.end(); itWheel++)
                {
                    itUnit = m_track_setup.AbsMapWheelUnits.find(itWheel->first);
                    std::string unit = "STEPS";
                    if(itUnit != m_track_setup.AbsMapWheelUnits.end())unit = itUnit->second;
                    if(itWheel->first == "ADC1")
                    {
                        m_adc1->moveAbsoluteStart(itWheel->second, unit);
                    }
                    else if(itWheel->first == "ADC2")
                    {
                        m_adc2->moveAbsoluteStart(itWheel->second, unit);
                    }
                }
            }
            if(!m_track_setup.RelMapWheel.empty())
            {
                for(itWheel = m_track_setup.RelMapWheel.begin(); itWheel != m_track_setup.RelMapWheel.end(); itWheel++)
                {
                    itUnit = m_track_setup.RelMapWheelUnits.find(itWheel->first);
                    std::string unit = "STEPS";
                    if(itUnit != m_track_setup.RelMapWheelUnits.end())unit = itUnit->second;
                    if(itWheel->first == "ADC1")
                    {
                        m_adc1->moveRelativeStart(itWheel->second, unit);
                    }
                    else if(itWheel->first == "ADC2")
                    {
                        m_adc2->moveRelativeStart(itWheel->second, unit);
                    }
                }
            }
            if(!m_track_setup.NamedMapWheel.empty())
            {
                for(itNamed = m_track_setup.NamedMapWheel.begin(); itNamed != m_track_setup.NamedMapWheel.end(); itNamed++)
                {
                    if(itNamed->first == "ADC1")
                    {
                        m_adc1->moveToNamedPositionStart(itNamed->second);
                    }
                    else if(itNamed->first == "ADC2")
                    {
                        m_adc2->moveToNamedPositionStart(itNamed->second);
                    }
                }
            }
        }
        catch(...)
        {
            E_LLOG(__FUNCTION__);
            U8_LLOG("");
            throw;
        }
    }

    double absolute = 0, relative = 0;
    std::string unit;
    std::map<std::string, std::string>::const_iterator itFind;
    std::map<std::string, double> absolutePos;
    std::map<std::string, std::string> namedPos, absoluteUnits;
    itFind = _dSetupPos.find("DROT_PosAng");
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
        {
            absolutePos[itFind->first] = absolute;
            absoluteUnits[itFind->first] = unit;
        }
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second);
        }
    }
    itFind = _dSetupPos.find("ADC1_PosAng");
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
        {
            absolutePos[itFind->first] = absolute;
            absoluteUnits[itFind->first] = unit;
        }
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second);
        }
    }
    itFind = _dSetupPos.find("ADC2_PosAng");
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
        {
            absolutePos[itFind->first] = absolute;
            absoluteUnits[itFind->first] = unit;
        }
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second);
        }
    }

    try
    {
        moveAbsoluteStart(absolutePos, absoluteUnits);
        moveToNamedPositionStart(namedPos);
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }

    U8_LLOG(__FUNCTION__);
    return status;
}

void trackingCtrl::setupPrepare(const std::map<std::string, std::string>& _mapNamedPos, trackSetup & setup)
{
    U6_LLOG(__FUNCTION__);
    std::map<std::string, std::tr1::tuple<int, int> > NamedPosMap;
    int idxPos;
    std::map<std::string, std::string>::const_iterator it;
    shins::util::EShinsMovements detected;
    double absolute, relative = 0;
    std::string unit;
    setup.clear();
    for (it = _mapNamedPos.begin(); it != _mapNamedPos.end(); ++it)
    {
        if(it->first == suc::ADC1_PosAng || it->first == suc::ADC2_PosAng)
        {
            detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            switch(detected)
            {
            case shins::util::SHINS_MOV_NAMED:
                NamedPosMap = m_mainMapNamedPos.find(it->first)->second;
                if(NamedPosMap.find(it->second) != NamedPosMap.end())
                    idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                else
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
                setup.NamedMapWheel.insert(std::pair<std::string, int>(it->first, idxPos));
                break;
            case shins::util::SHINS_MOV_ABSOLUTE:
                U9_LLOG("Detected Absolute Position : " << it->first << " -> " << it->second << "; value = " << absolute << "; unit = " << unit);
                setup.AbsMapWheel.insert(std::pair<std::string, double>(it->first, absolute));
                setup.AbsMapWheelUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_RELATIVE:
                U9_LLOG("Detected Relative Position : " << it->first << " -> " << it->second << "; value = " << relative << "; unit = " << unit);
                setup.RelMapWheel.insert(std::pair<std::string, double>(it->first, relative));
                setup.RelMapWheelUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_UNDEFINED:
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    U9_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
            }
            continue;
        }
        if(it->first == suc::DROT_PosAng)
        {
            detected = shins::util::detect_command_type(it->second, absolute, relative, unit);
            U9_LLOG("Element : " << std::setw(20) << it->first << "  <> Value : " << it->second);
            switch(detected)
            {
            case shins::util::SHINS_MOV_NAMED:
                NamedPosMap = m_mainMapNamedPos.find(it->first)->second;
                if(NamedPosMap.find(it->second) != NamedPosMap.end())
                    idxPos = std::tr1::get<0>(NamedPosMap[it->second]);
                else
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
                setup.NamedMapDrot.insert(std::pair<std::string, int>(it->first, idxPos));
                NamedPosMap.clear();
                break;
            case shins::util::SHINS_MOV_ABSOLUTE:
                U9_LLOG("Detected Absolute Position : " << it->first << "-> " << it->second << "; value = " << absolute);
                setup.AbsMapDrot.insert(std::pair<std::string, double>(it->first, absolute));
                setup.AbsMapDrotUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_RELATIVE:
                U9_LLOG("Detected Relative Position : " << it->first << "-> " << it->second << "; value = " << relative);
                setup.RelMapDrot.insert(std::pair<std::string, double>(it->first, relative));
                setup.RelMapDrotUnits.insert(std::pair<std::string, std::string>(it->first, unit));
                break;
            case shins::util::SHINS_MOV_UNDEFINED:
                {
                    std::stringstream msg;
                    msg << "Invalid Argument : [" << it->first << ", " << it->second << "]";
                    E_LLOG(msg.str());
                    U8_LLOG("");
                    throw(std::invalid_argument(msg.str()));
                }
            }
            continue;
        }
    }
    U8_LLOG("");
}


Ice::Int trackingCtrl::setupInstrumentWait(const sharknir::dictSetupNamedPos &_dSetupPos, const Nice::Time& _time)
{
    U6_LLOG(__FUNCTION__);
    Ice::Int status = EXIT_SUCCESS;
    double absolute = 0, relative = 0;
    std::string unit;
    std::map<std::string, std::string>::const_iterator itFind;
    itFind = _dSetupPos.find("DROT_PosAng");
    std::map<std::string, double> absolutePos;
    std::map<std::string, std::string> namedPos;
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
            absolutePos[itFind->first] = absolute;
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second)
        }
    }
    itFind = _dSetupPos.find("ADC1_PosAng");
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
            absolutePos[itFind->first] = absolute;
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second)
        }
    }
    itFind = _dSetupPos.find("ADC2_PosAng");
    if(itFind != _dSetupPos.end())
    {
        su::EShinsMovements movement = su::detect_command_type(itFind->second, absolute, relative, unit);
        if(movement == su::SHINS_MOV_ABSOLUTE)
            absolutePos[itFind->first] = absolute;
        else if(movement == su::SHINS_MOV_NAMED)
            namedPos[itFind->first] = itFind->second;
        else
        {
            W_LLOG("Unrecognized setup command " << itFind->first << " ~ " << itFind->second)
        }
    }
    try
    {
        moveAbsoluteWait(absolutePos, _time);
        moveToNamedPositionWait(namedPos, _time);
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG(__FUNCTION__);
    return status;
}


void trackingCtrl::moveToNamedPosition(const std::map<std::string, std::string>& _mapNamedPos)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");
    try
    {
        moveToNamedPositionStart(_mapNamedPos);
        moveToNamedPositionWait(_mapNamedPos, Nice::Time::seconds(100));
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}


void trackingCtrl::moveToNamedPositionStart(const std::map<std::string, std::string>& _mapNamedPos)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");

    int idxPos;
    std::map<std::string, std::string>::const_iterator itPos = _mapNamedPos.find("DROT_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        U9_LLOG("DROT_PosAng :" << itPos->second);
        if(m_drot)
        {
            try
            {
                I_LLOG("Moving to named position " << itPos->second);
                const std::map<std::string, std::tr1::tuple<int, int> > & NamedPosMap = m_mainMapNamedPos.find(suc::DEROT)->second;
                if(NamedPosMap.find(itPos->second) != NamedPosMap.end())
                {
                    m_drot->moveToNamedPositionStart(std::tr1::get<0>(NamedPosMap.find(itPos->second)->second));
                }
                else
                {
                    std::stringstream log_msg;
                    log_msg << "Error :cannot find named position " << itPos->second;
                    E_LLOG(log_msg.str());
                    throw(std::runtime_error(log_msg.str()));
                }
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested derotation to position: derotator is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }

    itPos = _mapNamedPos.find("ADC1_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        if(m_ptrAdcWorkerThread && m_ptrAdcWorkerThread->GetMode() == "OFF")
        {
            if(m_adc1)
            {
                try
                {
                    const std::map<std::string, std::tr1::tuple<int, int> > & NamedPosMap = m_mainMapNamedPos.find(suc::ADC1)->second;
                    if(NamedPosMap.find(itPos->second) != NamedPosMap.end())
                    {
                        m_adc1->moveToNamedPositionStart(std::tr1::get<0>(NamedPosMap.find(itPos->second)->second));
                    }
                    else
                    {
                        std::stringstream log_msg;
                        log_msg << "Error :cannot find named position " << itPos->second;
                        E_LLOG(log_msg.str());
                        throw(std::runtime_error(log_msg.str()));
                    }
                }
                catch(...)
                {
                    TrackLogBoostDiagnosticInformation(__LINE__);
                    U8_LLOG(__FUNCTION__);
                    throw;
                }
            }
            else
            {
                std::string log_msg("Error : ADC1 is not connected");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        else
        {
            std::stringstream log_msg;
            if(m_ptrAdcWorkerThread)
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC automatic correction mode is ON";
            }
            else
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC is not available";
            }
            E_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    itPos = _mapNamedPos.find("ADC2_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        if(m_ptrAdcWorkerThread && m_ptrAdcWorkerThread->GetMode() == "OFF")
        {
            if(m_adc2)
            {
                try
                {
                    const std::map<std::string, std::tr1::tuple<int, int> > & NamedPosMap = m_mainMapNamedPos.find(suc::ADC2)->second;
                    if(NamedPosMap.find(itPos->second) != NamedPosMap.end())
                    {
                        m_adc2->moveToNamedPositionStart(std::tr1::get<0>(NamedPosMap.find(itPos->second)->second));
                    }
                    else
                    {
                        std::stringstream log_msg;
                        log_msg << "Error :cannot find named position " << itPos->second;
                        E_LLOG(log_msg.str());
                        throw(std::runtime_error(log_msg.str()));
                    }
                }
                catch(...)
                {
                    TrackLogBoostDiagnosticInformation(__LINE__);
                    U8_LLOG(__FUNCTION__);
                    throw;
                }
            }
            else
            {
                std::string log_msg("Requested positioning to angle: ADC2 is not connected");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        else
        {
            std::stringstream log_msg;
            if(m_ptrAdcWorkerThread)
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC automatic correction mode is ON";
            }
            else
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC is not available";
            }
            E_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}


void trackingCtrl::moveToNamedPositionWait(const std::map<std::string, std::string>& _mapNamedPos, const Nice::Time & _time)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");

    ////// DROT //////
    std::map<std::string, std::string>::const_iterator itPos = _mapNamedPos.find("DROT_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        if(m_drot)
        {
            try
            {
                U9_LLOG(__FUNCTION__ << " ~ Start Waiting " << _time << " seconds");
                m_drot->moveToNamedPositionWait(_time);
                U9_LLOG(__FUNCTION__ << " ~ Waiting finished");
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested derotation to position: derotator is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }

    ////// ADC //////
    itPos = _mapNamedPos.find("ADC1_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        if(m_adc1)
        {
            try
            {
                int result = m_adc1->moveToNamedPositionWait(_time);
                U9_LLOG("m_adc1->moveAbsoluteWait returned -> " << result);
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested positioning to angle: ADC is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }
    itPos = _mapNamedPos.find("ADC2_PosAng");
    if(itPos != _mapNamedPos.end())
    {
        if(m_adc2)
        {
            try
            {
                m_adc2->moveToNamedPositionWait(_time);
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested positioning to angle: ADC is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }
    U8_LLOG("");
}


void trackingCtrl::install_callbacks()
{
    U6_LLOG("trackingCtrl::"  << __FUNCTION__ );
    Nice::PropertyPtr prop;
    std::string bridgeName;
    std::vector<std::string> vecCommands;
    vecCommands.push_back("MOVETOHOME");
    vecCommands.push_back("MOVETONAMEDPOSITION");
    vecCommands.push_back("MOVEABSOLUTE");
    vecCommands.push_back("MOVERELATIVE");
    vecCommands.push_back("MOVETOLIMIT");
    vecCommands.push_back("STARTDEROTATION");
    vecCommands.push_back("STARTDEROTATIONHERE");
    std::vector<std::string>::iterator itCmd;
    if(m_drot)
    {
    for(itCmd = vecCommands.begin(); itCmd < vecCommands.end(); itCmd++)
    {
        std::string strPropName = "CMD." + *itCmd + ".EXEC";
        bridgeName = "derotator-" + su::str_tolower(*itCmd);
        prop = m_drot->config()->leaf(strPropName);
        if(prop->getBridge(bridgeName))
        {
            U9_LLOG("Deleting bridge : " << bridgeName);
            prop->removeBridge(bridgeName);
        }
        void (DrotDevice::* Bridge)(const Nice::Completion &)  = NULL;
        if(strPropName == "CMD.MOVETOHOME.EXEC")
        {
            Bridge = &DrotDevice::MoveToHomeBridge;
        }
        else if(strPropName == "CMD.MOVETONAMEDPOSITION.EXEC")
        {
            Bridge = &DrotDevice::MoveToNamedBridge;
        }
        else if(strPropName == "CMD.MOVEABSOLUTE.EXEC")
        {
            Bridge = &DrotDevice::MoveAbsoluteBridge;
        }
        else if(strPropName == "CMD.MOVERELATIVE.EXEC")
        {
            Bridge = &DrotDevice::MoveRelativeBridge;
        }
        else if(strPropName == "CMD.MOVETOLIMIT.EXEC")
        {
            Bridge = &DrotDevice::MoveToLimitBridge;
        }
        else if(strPropName == "CMD.STARTDEROTATION.EXEC")
        {
            Bridge = &DrotDevice::StartDerotationBridge;
        }
        else if(strPropName == "CMD.STARTDEROTATIONHERE.EXEC")
        {
            Bridge = &DrotDevice::StartDerotationHereBridge;
        }
        if(m_MapDrots.begin()->second.iPtr)
        {
            prop->applyBridge(
                new Nice::PropertyBridgeMemFn<Nice::Completion, DrotDevice, void (DrotDevice::*)(const Nice::Completion &)>(&m_MapDrots.begin()->second, Bridge, prop->id()), bridgeName);
            U9_LLOG( m_drot << " -- " << bridgeName << " ~ " << Bridge);
        }
        else
        {
            U9_LLOG("Could  not create Bridge due to NULL Object");
        }
    }
    }

    // std::string strPropName = "VAR.POSITION.ABSOLUTE_ENCODER";
    // void (DrotDevice::* Bridge)(const double & val)  = NULL;
    // bridgeName = "derotator-absolute-encoder";
    // prop = m_drot->config()->leaf(strPropName);
    // Bridge = &DrotDevice::VarAbsoluteEncoderBridge;

    // // Non funzionano
    // // prop->applyBridge(
    // //     new Nice::PropertyBridgeMemFn<double, DrotDevice, void (DrotDevice::*)(double)>(&m_MapDrots.begin()->second, Bridge, prop->id()), bridgeName);
    // //prop->applyBridge(
    // //    new Nice::PropertyBridgeMemFn<double, DrotDevice>(&m_MapDrots.begin()->second, &DrotDevice::VarAbsoluteEncoderBridge), strPropName);

    // // Funziona ma non prosegue OS
    // prop->applyBridge(
    //    new Nice::PropertyBridgeMemFn<double, niceCallback>(&m_callbackObject, &niceCallback::callback_absolute_encoder_position), strPropName);
    // U9_LLOG( m_drot << " -- " << bridgeName << " ~ " << Bridge);

    U8_LLOG("");
}

std::string trackingCtrl::getMode(const std::string & _device)
{
//    U6_LLOG(__FUNCTION__);
    std::string dev = su::left(_device, 3);
    std::string result;
    if(dev == "DRO")
    {
        if(m_drot)
            result = m_drot->getMode();
        else {W_LLOG("Asked DROT mode but DROT is unavailable");}
    }
    else if(dev == "ADC")
    {
        if(m_ptrAdcWorkerThread)
        {
            result = m_ptrAdcWorkerThread->GetMode();
        }
    }
    else
    {
        std::stringstream log_msg;
        log_msg << "Error : " << _device << " non recognized";
        U6_LLOG(__FUNCTION__);
        E_LLOG(log_msg.str());
        U8_LLOG("");
        throw(std::invalid_argument(log_msg.str()));
    }
//    U8_LLOG("");
    return result;
}


void trackingCtrl::moveAbsolute(const std::map<std::string, double> & _mapPos, const std::map<std::string, std::string> & _mapUnit)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");
    try
    {
        moveAbsoluteStart(_mapPos, _mapUnit);
        moveAbsoluteWait(_mapPos, Nice::Time::seconds(100));
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        U8_LLOG("");
        throw;
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}


void trackingCtrl::moveAbsoluteStart(const std::map<std::string, double> & _mapPos, const std::map<std::string, std::string> & _mapUnit)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");

    std::map<std::string, double>::const_iterator itPos = _mapPos.find("DROT_PosAng");
    if(itPos != _mapPos.end())
    {
        U9_LLOG("DROT_PosAng :" << itPos->second);
        if(m_drot)
        {
            try
            {
                std::string unit = "DEGREE";
                if(_mapUnit.find("DROT_PosAng") != _mapUnit.end() && _mapUnit.find("DROT_PosAng")->second != "DEFAULT")
                    unit = _mapUnit.find("DROT_PosAng")->second;
                m_drot->moveAbsoluteStart(itPos->second, unit);
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested derotation to position: derotator is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }

    itPos = _mapPos.find("ADC1_PosAng");
    if(itPos != _mapPos.end())
    {
        if(m_ptrAdcWorkerThread && m_ptrAdcWorkerThread->GetMode() == "OFF")
        {
            if(m_adc1)
            {
                try
                {
                    std::string unit = "DEGREE";
                    if(_mapUnit.find("ADC1_PosAng") != _mapUnit.end() && _mapUnit.find("ADC1_PosAng")->second != "DEFAULT")
                        unit = _mapUnit.find("ADC1_PosAng")->second;
                    m_adc1->moveAbsoluteStart(itPos->second, unit);
                }
                catch(...)
                {
                    TrackLogBoostDiagnosticInformation(__LINE__);
                    U8_LLOG(__FUNCTION__);
                    throw;
                }
            }
            else
            {
                std::string log_msg("Error : ADC1 is not connected");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        else
        {
            std::stringstream log_msg;
            if(m_ptrAdcWorkerThread)
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC automatic correction mode is ON";
            }
            else
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC is not available";
            }
            E_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    itPos = _mapPos.find("ADC2_PosAng");
    if(itPos != _mapPos.end())
    {
        if(m_ptrAdcWorkerThread && m_ptrAdcWorkerThread->GetMode() == "OFF")
        {
            if(m_adc2)
            {
                try
                {
                    std::string unit = "DEGREE";
                    if(_mapUnit.find("ADC2_PosAng") != _mapUnit.end() && _mapUnit.find("ADC2_PosAng")->second != "DEFAULT")
                        unit = _mapUnit.find("ADC2_PosAng")->second;
                    m_adc2->moveAbsoluteStart(itPos->second, unit);
                }
                catch(...)
                {
                    TrackLogBoostDiagnosticInformation(__LINE__);
                    U8_LLOG(__FUNCTION__);
                    throw;
                }
            }
            else
            {
                std::string log_msg("Requested positioning to angle: ADC2 is not connected");
                E_LLOG(log_msg);
                U8_LLOG("");
                throw(std::runtime_error(log_msg));
            }
        }
        else
        {
            std::stringstream log_msg;
            if(m_ptrAdcWorkerThread)
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC automatic correction mode is ON";
            }
            else
            {
                log_msg << __FUNCTION__ << " ~ Error : ADC is not available";
            }
            E_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}

/** Method to wait end of  motion to absolute position for tracking devices */
void trackingCtrl::moveAbsoluteWait(const std::map<std::string, double>& _mapPos, const Nice::Time& _time)
{
    U6_LLOG(__FUNCTION__ << " ~ entering");
    su::Timer timer(_time);

    const int WAIT_RESULT_TIMED_OUT = 0;
    ////// DROT //////
    std::map<std::string, double>::const_iterator itPos = _mapPos.find("DROT_PosAng");
    if(itPos != _mapPos.end())
    {
        if(m_drot)
        {
            try
            {
                U9_LLOG(__FUNCTION__ << " ~ Start Waiting for " << timer.residual());
                if(m_drot->moveAbsoluteWait(timer.residual()) == WAIT_RESULT_TIMED_OUT)
                {
                    std::stringstream log_msg;
                    log_msg << __FUNCTION__ << suc::tilde << "DROT_PosAng" << suc::tilde << "Error : wait timed out";
                    throw(std::runtime_error(log_msg.str()));
                }
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested derotation to position: derotator is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }

    ////// ADC //////
    itPos = _mapPos.find("ADC1_PosAng");
    if(itPos != _mapPos.end())
    {
        if(m_adc1)
        {
            try
            {
                if(m_adc1->moveAbsoluteWait(timer.residual()) == WAIT_RESULT_TIMED_OUT)
                {
                    std::stringstream log_msg;
                    log_msg << __FUNCTION__ << suc::tilde << "ADC1_PosAng" << suc::tilde << "Error : wait timed out";
                    throw(std::runtime_error(log_msg.str()));
                }
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested positioning to angle: ADC is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }
    itPos = _mapPos.find("ADC2_PosAng");
    if(itPos != _mapPos.end())
    {
        if(m_adc2)
        {
            try
            {
                if(m_adc2->moveAbsoluteWait(timer.residual()) == WAIT_RESULT_TIMED_OUT)
                {
                    std::stringstream log_msg;
                    log_msg << __FUNCTION__ << suc::tilde << "ADC2_PosAng" << suc::tilde << "Error : wait timed out";
                    throw(std::runtime_error(log_msg.str()));
                }
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::string log_msg("Requested positioning to angle: ADC is not connected");
            E_LLOG(log_msg);
            U8_LLOG("");
            throw(std::runtime_error(log_msg));
        }
    }
    U8_LLOG("");
}


void trackingCtrl::moveRelative(const std::map<std::string, double>& _mapPos, const std::map<std::string, std::string> & _mapUnit)
{
    std::stringstream log_msg;
    log_msg << __FUNCTION__ << " ~ Not yet implemented";
    W_LLOG(log_msg.str());
    throw(std::runtime_error(log_msg.str()));
}


void trackingCtrl::moveRelativeStart(const std::map<std::string, double>& _mapPos, const std::map<std::string, std::string> & _mapUnit)
{
    std::stringstream log_msg;
    log_msg << __FUNCTION__ << " ~ Not yet implemented";
    W_LLOG(log_msg.str());
    throw(std::runtime_error(log_msg.str()));
}


void trackingCtrl::moveRelativeWait(const std::map<std::string, double>& _mapPos, const Nice::Time& _time)
{
    std::stringstream log_msg;
    log_msg << __FUNCTION__ << " ~ Not yet implemented";
    W_LLOG(log_msg.str());
    throw(std::runtime_error(log_msg.str()));
}

/** @brief      Starts tracking test. Starts derotation for DROT at constant speed for the specified range. OsStartDerotationTest is blocking (synchronous)
  * @param[in] startPosDegree  The start position of derotation
  * @param[in] endPosDegree  End position of derotation
  * @param[in] speedDegreeSec Desired speed for derotation. Absolute value is taken into account */
void trackingCtrl::startTrackingTest(double startPosDegree, double endPosDegree, double speedDegreeSec)
{
    U6_LLOG(__FUNCTION__);
    if(m_drotMode == DROT_MODE_FIELD_FREE || m_drotMode == DROT_MODE_FIELD_FIXED)
    {
        try
        {
            I_LLOG("Starting Tracking test ... ");
            m_drot->OsStartDerotationTest(startPosDegree, endPosDegree, speedDegreeSec);
            I_LLOG("tracking test Completed");
        }
        catch(Basda::Dev::MoccaUploadExternalProfileException & err)
        {
            U7_LLOG(err.what())
            U8_LLOG("");
            throw err;
        }
        catch(Basda::DevException & err)
        {
            U7_LLOG(err.what())
            U8_LLOG("");
            throw err;
        }
        catch(...)
        {
            TrackLogBoostDiagnosticInformation(__LINE__);
            U8_LLOG("");
            throw;
        }
    }
    else
    {
        std::string log_msg = "Cannot start tracking when DROT is in PUPIL mode";
        E_LLOG(log_msg);
        throw(std::runtime_error(log_msg));
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}



/** @brief      Starts tracking. Starts derotation and tracking for DROT. OsStartDerotation is blocking (synchronous)
  * @param[in] _startDate  The start date
  * @param[in] _position   The start position of the derotator
  * @param[in] _unit       The unit for the start position 
  * @param[in] _autoStart  Flag indicating the system can choose automatically the start position. If true _position is ignored */
void trackingCtrl::startTracking(const Nice::Date& _startDate, double _position, std::string _unit, bool _autoStart)
{
    U6_LLOG(__FUNCTION__);
    if(m_drotMode == DROT_MODE_FIELD_FREE || m_drotMode == DROT_MODE_FIELD_FIXED)
    {
        if(m_drot)
        {
            try
            {
                I_LLOG("Starting tracking ... ");
//                m_drot->OsStartDerotation(_startDate, _position, _unit, _autoStart);
                I_LLOG("tracking Complted");
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::stringstream log_msg;
            log_msg << "Cannot start Derotation: DROT is unavailable";
            W_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    else
    {
        std::string log_msg = "Cannot start tracking when DROT is in PUPIL mode";
        E_LLOG(log_msg);
        throw(std::runtime_error(log_msg));
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}

void trackingCtrl::startTracking()
{
    U6_LLOG(__FUNCTION__);
    if(m_drotMode == DROT_MODE_FIELD_FREE || m_drotMode == DROT_MODE_FIELD_FIXED)
    {
        if(m_drot)
        {
            try
            {
                m_drot->OsStartDerotation(m_drotMode == DROT_MODE_FIELD_FREE);
                I_LLOG("tracking started");
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                U8_LLOG("");
                throw;
            }
        }
        else
        {
            std::stringstream log_msg;
            log_msg << "Cannot start Derotation: DROT is unavailable";
            W_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }
    else
    {
        std::string log_msg = "Cannot start tracking when DROT is in PUPIL mode";
        E_LLOG(log_msg);
        U8_LLOG("");
        throw(std::runtime_error(log_msg));
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
}

int trackingCtrl::waitTracking(Ice::Long seconds)
{
    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    if(m_drot)
    {
        try
        {
            status = m_drot->OsWaitDerotation(seconds);
        }
        catch(...)
        {
            TrackLogBoostDiagnosticInformation(__LINE__);
            U8_LLOG("");
            throw;
        }
    }
    else
    {
        std::stringstream log_msg;
        log_msg << "Cannot start Derotation: DROT is unavailable";
        W_LLOG(log_msg.str());
        U8_LLOG("");
        throw(std::runtime_error(log_msg.str()));
    }
    U8_LLOG(__FUNCTION__ << " ~ exiting");
    return status;
}


/** @brief set devices mode, will be used as discriminant when starting tracking
  * @param[in] _mapSetup: the map as read from the OB */
int trackingCtrl::setMode(const std::map<std::string, std::string> & _mapSetup)
{
    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    std::map<std::string, std::string>::const_iterator itFindMode;

    itFindMode = _mapSetup.find("DROT_Mode");
    if(itFindMode != _mapSetup.end())
    {
        EDrotModes previousMode = m_drotMode;
        if(itFindMode->second == "PUPIL")
        {
            m_drotMode = DROT_MODE_PUPIL;
        }
        else if(itFindMode->second == "FIELD_FIXED")
        {
            m_drotMode = DROT_MODE_FIELD_FIXED;
        }
        else if(itFindMode->second == "FIELD_FREE")
        {
            m_drotMode = DROT_MODE_FIELD_FREE;
        }
        else
        {
            std::stringstream log_msg;
            log_msg << "Error : admitted modes for DROT are {\"PUPIL\", \"FIELD_FIXED\", \"FIELD_FREE\"}";
            E_LLOG(log_msg.str());
            throw(std::invalid_argument(log_msg.str()));
        }
        if(m_drot)
        {
            if(itFindMode->second == "PUPIL" && m_drot->isMoving())
            {
                m_drot->stop();
                I_LLOG("DROT moving when switching to PUPIL mode: motion stopped ");
            }
            else
            {
    //            startTracking
            }
        }
        else
        {
            std::stringstream log_msg;
            log_msg << "Cannot change mode : DROT is unavailable";
            W_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }

    itFindMode = _mapSetup.find("ADC_Mode");
    if(itFindMode != _mapSetup.end())
    {
        if(m_ptrAdcWorkerThread)
            m_ptrAdcWorkerThread->SetMode(itFindMode->second);
        else
        {
            std::stringstream log_msg;
            log_msg << __FUNCTION__ << " ~ Error : ADC is not available";
            E_LLOG(log_msg.str());
            U8_LLOG("");
            throw(std::runtime_error(log_msg.str()));
        }
    }

    U8_LLOG("");
    return status;
}


// get the number of devices connected
int trackingCtrl::getNdevices()
{
    int nDevices = 0;
    U6_LLOG(__FUNCTION__ << " ~ Getting number of tracking devices");
    if(m_adc1)
    {
        nDevices++;
    }
    if(m_adc2)
    {
        nDevices++;
    }
    if(m_drot)
    {
        nDevices++;
    }
    U8_LLOG("");
    return nDevices;
}


Ice::Int trackingCtrl::getPositionDict(std::map<std::string, double>& _mapPosition, const std::string& _unit)
{
    U6_LLOG(__FUNCTION__);
    Ice::Int status = EXIT_SUCCESS;

    std::map<std::string, double>::const_iterator itFindDerot = _mapPosition.find(suc::DEROT);
    std::map<std::string, double>::const_iterator itFindADC1 = _mapPosition.find(suc::ADC1);
    std::map<std::string, double>::const_iterator itFindADC2 = _mapPosition.find(suc::ADC2);

    if(itFindDerot != _mapPosition.end())
    {
        _mapPosition[itFindDerot->first] = getPosition(itFindDerot->first, _unit);
    }
    if(itFindADC1 != _mapPosition.end())
    {
        _mapPosition[itFindADC1->first] = getPosition(itFindADC1->first, _unit);
    }
    if(itFindADC2 != _mapPosition.end())
    {
        _mapPosition[itFindADC2->first] = getPosition(itFindADC2->first, _unit);
    }
    U8_LLOG("");
    return status;
}

/** input
 *  @param[in] _deviceName : equal the TAN service alias name
 *  @param[in] _unit       :  */
double trackingCtrl::getPosition(const std::string& _deviceName, const std::string &_unit)
{
//    U6_LLOG(__FUNCTION__);

    double position = -9e12;
    if(_deviceName == suc::DEROT)
    {
        if(m_drot)
        {
            position = m_drot->getPosition(_unit);
        }
        else
        {
            W_LLOG("Could not get DROT position : module is not active");
        }
    }

    if(_deviceName == suc::ADC1)
    {
        if(m_adc1)
        {
            position = m_adc1->getPosition(_unit);
        }
        else
        {
            W_LLOG("Could not get ADC1 position : module is not active");
        }
    }

    if(_deviceName == suc::ADC2)
    {
        if(m_adc2)
        {
            position = m_adc2->getPosition(_unit);
        }
        else
        {
            W_LLOG("Could not get DROT position : module is not active");
        }
    }
//    U8_LLOG("");
    return position;
}

// only derotator is equipped with absolute encoder
double trackingCtrl::getAbsEncPosition()
{
    double absEncPos = 0;
    U6_LLOG(__FUNCTION__ << " ~ Getting drot absolute encoder postion ...")
    if(m_drot)
    {
        absEncPos = m_drot->getAbsoluteEncoderPosition();
    }
    else
    {
        std::stringstream log_msg;
        log_msg << "Cannot get absolute encoder position : DROT is unavailable";
        W_LLOG(log_msg.str());
        U8_LLOG("");
        throw(std::runtime_error(log_msg.str()));
    }
    U8_LLOG("");
    return absEncPos;
}

int trackingCtrl::isMoving(std::map<std::string, bool>& _mapDev)
{
    U6_LLOG(__FUNCTION__ << " ~ Checking ifDROT is moving ... ");
    W_LLOG(__FUNCTION__ << " ~ NOT YET IMPLEMENTED")
    U8_LLOG("");
    throw(std::runtime_error("isMoving : function not yet implemented"));
    return EXIT_FAILURE;
}

int trackingCtrl::emergencyAbort()
{
    U6_LLOG(__FUNCTION__ << " ~ ABORT motion for derotator");
    int status = EXIT_SUCCESS;
    try
    {
        if(m_drot)
        {
            m_drot->abort();
        }
        else
        {
            W_LLOG("Could not abort DROT : module is not active");
        }
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        W_LLOG("Caught exception aborting DROT, not retrown to assure emergency abort");
        status = EXIT_FAILURE;
    }

    try
    {
        if(m_adc1)
        {
            m_adc1->abort();
        }
        else
        {
            W_LLOG("Could not abort ADC1 : module is not active");
        }
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        W_LLOG("Caught exception aborting ADC1, not retrown to assure emergency abort");
        status = EXIT_FAILURE;
    }
    try
    {
        if(m_adc2)
        {
            m_adc2->abort();
        }
        else
        {
            W_LLOG("Could not stop ADC2 : module is not active");
        }
    }
    catch(...)
    {
        TrackLogBoostDiagnosticInformation(__LINE__);
        W_LLOG("Caught exception aborting ADC2, not retrown to assure emergency abort");
        status = EXIT_FAILURE;
    }
    U8_LLOG("");
    return status;
}


int trackingCtrl::stopAll()
{
    U6_LLOG(__FUNCTION__);
    W_LLOG("Stopping all tracking devices");
    int status = EXIT_SUCCESS;
    if(m_drot)
    {
        try
        {
            I_LLOG("Stopping derotator ...");
            m_drot->stop();
        }
        catch(...)
        {
            TrackLogBoostDiagnosticInformation(__LINE__);
            W_LLOG("Caught exception stopping DROT, not retrown to assure emergency stop");
            status = EXIT_FAILURE;
        }
    }
    else
    {
        W_LLOG("Could not stop DROT : module is not active");
    }
    if(m_adc1)
    {
        try
        {
            m_adc1->stop();
        }
        catch(...)
        {
            TrackLogBoostDiagnosticInformation(__LINE__);
            W_LLOG("Caught exception stopping ADC1, not retrown to assure emergency stop");
            status = EXIT_FAILURE;
        }
    }
    else
    {
        W_LLOG("Could not stop ADC1 : module is not active");
    }
    if(m_adc2)
    {
        try
        {
            m_adc2->stop();
            I_LLOG("Stopped");
        }
        catch(...)
        {
            TrackLogBoostDiagnosticInformation(__LINE__);
            W_LLOG("Caught exception stopping ADC2, not retrown to assure emergency stop");
            status = EXIT_FAILURE;
        }
    }
    else
    {
        W_LLOG("Could not stop ADC2 : module is not active");
    }
    U8_LLOG("");
    return status;
}


int trackingCtrl::stop(const std::vector<std::string>& _vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    std::vector<std::string>::const_iterator itFind;
    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::DROT_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_drot)
        {
            try
            {
                m_drot->stop();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception stopping DROT, not retrown to assure emergency stop");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not stop DROT : module is not active");
        }
    }

    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC1_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_adc1)
        {
            try
            {
                m_adc1->stop();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception stopping ADC1, not retrown to assure emergency stop");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not stop ADC1 : module is not active");
        }
    }

    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC2_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_adc2)
        {
            try
            {
                m_adc2->stop();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception stopping ADC2, not retrown to assure emergency stop");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not stop ADC2 : module is not active");
        }
    }
    U8_LLOG("");
    return status;
}

int trackingCtrl::abort(const std::vector<std::string>& _vecDevAlias)
{
    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    std::vector<std::string>::const_iterator itFind;
    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::DROT_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_drot)
        {
            try
            {
                m_drot->abort();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception aborting DROT, not retrown to assure emergency abort");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not abort DROT : module is not active");
        }
    }
    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC1_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_adc1)
        {
            try
            {
                m_adc1->abort();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception aborting ADC2, not retrown to assure emergency abort");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not abort ADC1 : module is not active");
        }
    }

    itFind = std::find(_vecDevAlias.begin(), _vecDevAlias.end(), suc::ADC2_PosAng);
    if(itFind != _vecDevAlias.end())
    {
        if(m_adc2)
        {
            try
            {
                m_adc2->abort();
            }
            catch(...)
            {
                TrackLogBoostDiagnosticInformation(__LINE__);
                W_LLOG("Caught exception aborting ADC2, not retrown to assure emergency abort");
                status = EXIT_FAILURE;
            }
        }
        else
        {
            W_LLOG("Could not abort ADC2 : module is not active");
        }
    }
    U8_LLOG("");
    return status;
}


// Nice::SeqDouble trackingCtrl::getSinusCurve(Nice::Time _duration,
//                                             Nice::Time _period,
//                                             Nice::Time _timeResolution,
//                                             double _amplitude,
//                                             double _phase,
//                                             double _startPoint)
// {
//     U6_LLOG(__FUNCTION__);
//     Nice::SeqDouble positionSequence;
//     int numberOfSamples = (int)(_duration.toSecondsDouble() / _timeResolution.toSecondsDouble());
//     for(int i = 0; i < numberOfSamples; i++)
//     {
//         positionSequence.push_back(_amplitude * sin(2 * M_PI * i * _timeResolution.toSecondsDouble() / _period.toSecondsDouble() + _phase) + _startPoint);
//     }
//     U8_LLOG("");
//     return positionSequence;
// }


int trackingCtrl::getJsonStatus(std::string & jsonStatus)
{
//    U6_LLOG(__FUNCTION__);
    int status = EXIT_SUCCESS;
    bool bJsonEmpty = jsonStatus.empty();
    bool bDrotReachable = true;

    //    struct MotorStatus{
    //               int steps;
    //               string namedPos;
    //               bool connected;
    //               bool moving;
    //               bool atHome;
    //               bool inSwNegativeLimit;
    //               bool inSwPositiveLimit;
    //               bool inHwNegativeLimit;
    //               bool inWhPositiveLimit;
    //    };
    if(m_drot)
    {
        bool isMoving = false;
//        try{isMoving = m_drot->isMoving();}catch(...){E_LLOG("isMoving");}
        static int DrotCounter = 0;
//        if((DrotCounter++ % 60) == 0){U9_LLOG("Drot Reachable -> " << m_drot->isReachable());}
        try
        {
            isMoving = m_drot->isMoving();
        }
        // catch(Nice::ChatException err)
        // {
        //     usleep(1000000);
        //     bDrotReachable = false;
        // }
        catch(...)
        {
            bDrotReachable = false;
            E_LLOG("isMoving failed");
            usleep(1000000);
        }
        if(bDrotReachable)
        {
            std::string named_pos = m_drot->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
            std::string service_name;
            try{service_name = m_drot->readServiceName();}catch(...){E_LLOG("readServiceName");}// service_name = "derotator"
            bool isSimulated = false;
            Nice::SeqBitSet switch_status;
        //    try{switch_status = m_drot->getPositionSwitchStatus(); isSimulated = switch_status[0].getDepth() < 32;}catch(...){E_LLOG("getPositionSwitchStatus");}
        //    try{U9_LLOG("getPositionSwitchStatus : " << switch_status << " ~ " << switch_status[0].getDepth());}catch(...){E_LLOG("getPositionSwitchStatus");}
            bool isAtHome = false;
            try{isAtHome = m_drot->isAtHome();}catch(...){E_LLOG("isAtHome");}
            bool inLimit = false;
            try{inLimit = m_drot->isAtLimit();}catch(...){E_LLOG("isAtLimit");}
            if(inLimit)
            {
        //         try{U9_LLOG("isAtLimit2 : " << m_drot->get<int>("VAR.LIMIT"));}catch(...){E_LLOG("isAtLimit2");}
        //         std::string limit_result = m_drot->config()->getAtNode<std::string>("VAR.LIMIT");
        //         U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
        //         m_drot->config()->setAtNode<std::string>("VAR.LIMIT", "-2");
        //         limit_result = m_drot->config()->getAtNode<std::string>("VAR.LIMIT");
        //         U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
        //         limit_result = m_drot->config()->leaf("VAR.LIMIT")->get<std::string>();
        //         U9_LLOG("PropertyName : " << "VAR.LIMIT" << "; limit_result : " << limit_result);
        // //                val.limit = m_node->getAtNode<double> ("LIMIT")
        //         double limit_double = m_drot->config()->getAtNode<double>("LIMIT");
        //         U9_LLOG("PropertyName : " << "LIMIT" << "; limit_double : " << limit_double);
            }
        //    try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
        //    try{U9_LLOG("getNamedPosition : " << m_drot->getNamedPosition(1));}catch(...){E_LLOG("getNamedPosition");}
        //    try{U9_LLOG("readPosition : " << m_drot->readPosition());}catch(...){E_LLOG("readPosition");}
        //    try{U9_LLOG("readDeviceEncoderPosition : " << m_drot->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
        //    try{U9_LLOG("readAbsoluteEncoderPosition : " << m_drot->readAbsoluteEncoderPosition());}catch(...){E_LLOG("readAbsoluteEncoderPosition");}
        //    try{U9_LLOG("readIncrementalEncoderPosition : " << m_drot->readIncrementalEncoderPosition());}catch(...){E_LLOG("readIncrementalEncoderPosition");}
            bool isReachable = false;
            try{isReachable = m_drot->isReachable();}catch(...){E_LLOG("isReachable");}
        //    try{U9_LLOG("getPosition : " << m_drot->getPosition());}catch(...){E_LLOG("getPosition");}
            int deviceEncoderPosition;
            try{ deviceEncoderPosition = m_drot->getDeviceEncoderPosition();}catch(...){E_LLOG("getDeviceEncoderPosition");}
        //    try{U9_LLOG("getDeviceEncoderPosition : " << m_drot->getDeviceEncoderPosition());}catch(...){E_LLOG("getDeviceEncoderPosition");}
        //    try{U9_LLOG("getAbsoluteEncoderPosition : " << m_drot->getAbsoluteEncoderPosition());}catch(...){E_LLOG("getAbsoluteEncoderPosition");}
        //    try{U9_LLOG("getIncrementalEncoderPosition : " << m_drot->getIncrementalEncoderPosition());}catch(...){E_LLOG("getIncrementalEncoderPosition");}
        //    try{U9_LLOG("getVelocity : " << m_drot->getVelocity());}catch(...){E_LLOG("getVelocity");}

            std::stringstream json_alias;
            json_alias << "\"" << "DROT_PosAng" << "\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
            json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
            jsonStatus += json_alias.str();
            jsonStatus += ", ";
        }
    }
    if(!m_drot || !bDrotReachable)
    {
        std::stringstream json_alias;
        std::string drot_status = "UNAVAILABLE";
        std::string unavailable = drot_status;
        bDrotReachable = false;
        bool isMoving = false, inLimit = false, isSimulated = false;
        json_alias << "\"" << "DROT_PosAng" << "\": { \"name\": \"" << "derotator" << "\", \"named\": \"" << drot_status << "\", \"encoder\": " << 0 << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << bDrotReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        jsonStatus += ", ";
    }
    // DROT_Mode
    {
        std::stringstream json_alias;
        std::string drot_status = "UNAVAILABLE";
        std::string unavailable = drot_status;
        bool isReachable = false;
        bool isMoving = false, inLimit = false, isSimulated = false;
        if(m_drot)
        {
            drot_status = m_drot->getMode();
            isReachable = true;
        }
        json_alias << "\"" << "DROT_Mode" << "\": { \"name\": \"" << "DROT_Mode" << "\", \"named\": \"" << drot_status << "\", \"encoder\": " << 0 << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
        jsonStatus += ", ";
    }

    Basda::MoccaWheelServiceWorkerInterfacePtr adcs[2] = {m_adc1, m_adc2};
    for(int i = 0; i < 2; i++)
    {
        if(m_adc1 && m_adc2)
        {
            bool isMoving = false;
            std::string service_name;
            try{service_name = adcs[i]->readServiceName();}catch(...){E_LLOG("readServiceName");}
//            try{isMoving = adcs[i]->isMoving();}catch(...){E_LLOG(service_name << " ~ isMoving");}
            try
            {
                isMoving = adcs[i]->isMoving();
            }
            // catch(const Nice::ChatException & err)
            // {
            //     continue;
            //     usleep(1000000);
            // }
            catch(...)
            {
                E_LLOG(service_name << " ~ isMoving");
                usleep(1000000);
                continue;
            }
            std::string named_pos = adcs[i]->config()->leaf("VAR.NAMED_POSITION")->get<std::string>();
            bool isSimulated = false;
            Nice::SeqBitSet switch_status;
    //        try{switch_status = adcs[i]->getPositionSwitchStatus(); isSimulated = switch_status.size() < 32;}catch(...){E_LLOG(service_name << " ~ getPositionSwitchStatus");} lots of errors
    //        try{U9_LLOG("getPositionSwitchStatus : " << switch_status);}catch(...){E_LLOG("getPositionSwitchStatus");}
            bool isAtHome = false;
            try{isAtHome = adcs[i]->isAtHome();}catch(...){E_LLOG(service_name << " ~ isAtHome");}
            bool inLimit = false;
    //        try{U9_LLOG("NamedPosition : " << named_pos);}catch(...){E_LLOG("Named Position");}
    //        try{U9_LLOG("readDeviceEncoderPosition : " << adcs[i]->readDeviceEncoderPosition());}catch(...){E_LLOG("readDeviceEncoderPosition");}
            bool isReachable = false;
            try{isReachable = adcs[i]->isReachable();}catch(...){E_LLOG(service_name << " ~ isReachable");}
    //        try{U9_LLOG("getPosition : " << adcs[i]->getPosition());}catch(...){E_LLOG("getPosition");}
            int deviceEncoderPosition;
            try{deviceEncoderPosition = adcs[i]->getDeviceEncoderPosition();}catch(...){E_LLOG(service_name << " ~ getDeviceEncoderPosition");}
    //        try{U9_LLOG("getVelocity : " << adcs[i]->getVelocity());}catch(...){E_LLOG("getVelocity");}
            std::stringstream json_alias;
            json_alias << "\"" << "ADC" << i + 1 << "_PosAng\": { \"name\": \"" << service_name << "\", \"named\": \"" << named_pos << "\", \"encoder\": " << deviceEncoderPosition << ", ";
            json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
            jsonStatus += json_alias.str();
            jsonStatus += ", ";
        }
        else
        {
            std::stringstream json_alias;
            std::string adc_status = "UNAVAILABLE";
            std::string unavailable = adc_status;
            bool isReachable = false;
            bool isMoving = false, inLimit = false, isSimulated = false;
            json_alias << "\"" << "ADC" << i + 1 << "_PosAng\": { \"name\": \"" << "PRISM" << i + 1 << "\", \"named\": \"" << unavailable << "\", \"encoder\": " << 0 << ", ";
            json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
            jsonStatus += json_alias.str();
            jsonStatus += ", ";
        }
    }
    // ADC_Mode
    {
        std::stringstream json_alias;
        std::string adc_status = "UNAVAILABLE";
        std::string unavailable = adc_status;
        bool isReachable = false;
        bool isMoving = false, inLimit = false, isSimulated = false;
        if(m_ptrAdcWorkerThread)
        {
            adc_status = m_ptrAdcWorkerThread->GetMode();
            isReachable = true;
        }
        json_alias << "\"" << "ADC_Mode" << "\": { \"name\": \"" << "ADC_Mode" << "\", \"named\": \"" << adc_status << "\", \"encoder\": " << 0 << ", ";
        json_alias << "\"is_moving\": " << isMoving << ", \"in_limit\": " << inLimit << ", \"available\": " << isReachable << ", \"simulated\": " << isSimulated << "}";
        jsonStatus += json_alias.str();
    }
    if(bJsonEmpty)
    {
//        U9_LLOG("jsonStatus : " << jsonStatus);
    }
//    U8_LLOG("");
    return status;
}


void trackingCtrl::setFlags(uint64_t flags, uint64_t more_flags)
{
    if(flags & shins::flags::TT_SAVE_RAW_DIAG_DATA)
    {
//        m_bSaveRawDiagData = true;
//        U9_LLOG("Save raw diag data : ON");
    }
    if(m_ptrAdcWorkerThread)
        m_ptrAdcWorkerThread->setFlags(flags, more_flags);
};
void trackingCtrl::unsetFlags(uint64_t flags, uint64_t more_flags)
{
    if(flags & shins::flags::TT_SAVE_RAW_DIAG_DATA)
    {
//        m_bSaveRawDiagData = false;
//        U9_LLOG("Save raw diag data : OFF");
    }
    if(m_ptrAdcWorkerThread)
        m_ptrAdcWorkerThread->unsetFlags(flags, more_flags);
};
void trackingCtrl::toggleFlags(uint64_t flags, uint64_t more_flags)
{
    if(flags & shins::flags::TT_SAVE_RAW_DIAG_DATA)
    {
//        m_bSaveRawDiagData = !m_bSaveRawDiagData;
//        U9_LLOG("Save raw diag data : " << (m_bSaveRawDiagData ? "ON" : "OFF"));
    }
    if(m_ptrAdcWorkerThread)
        m_ptrAdcWorkerThread->toggleFlags(flags, more_flags);
};
